/*************************************************************************

This software module was originally developed by 

	Ming-Chieh Lee (mingcl@microsoft.com), Microsoft Corporation
	Wei-ge Chen (wchen@microsoft.com), Microsoft Corporation
	Bruce Lin (blin@microsoft.com), Microsoft Corporation
	Chuang Gu (chuanggu@microsoft.com), Microsoft Corporation
	(date: March, 1996)

and edited by

	Hiroyuki Katata (katata@imgsl.mkhar.sharp.co.jp), Sharp Corporation
	Norio Ito (norio@imgsl.mkhar.sharp.co.jp), Sharp Corporation
	Shuichi Watanabe (watanabe@imgsl.mkhar.sharp.co.jp), Sharp Corporation
	(date: October, 1997)

in the course of development of the MPEG-4 Video (ISO/IEC 14496-2). 
This software module is an implementation of a part of one or more MPEG-4 Video tools 
as specified by the MPEG-4 Video. 
ISO/IEC gives users of the MPEG-4 Video free license to this software module or modifications 
thereof for use in hardware or software products claiming conformance to the MPEG-4 Video. 
Those intending to use this software module in hardware or software products are advised that its use may infringe existing patents. 
The original developer of this software module and his/her company, 
the subsequent editors and their companies, 
and ISO/IEC have no liability for use of this software module or modifications thereof in an implementation. 
Copyright is not released for non MPEG-4 Video conforming products. 
Sharp retains full right to use the code for his/her own purpose, 
assign or donate the code to a third party and to inhibit third parties from using the code for non <MPEG standard> conforming products. 
This copyright notice must be included in all copies or derivative works. 

Copyright (c) 1997.

Module Name:

	tps_decoder.cpp

Abstract:

	caller for decoder for temporal scalability

Revision History:

*************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <fstream.h>

#ifdef __PC_COMPILER_
#include <windows.h>
#include <mmsystem.h>
#endif // __PC_COMPILER_

#include "typeapi.h"
#include "mode.hpp"
#include "vopses.hpp"
#include "entropy/bitstrm.hpp"
#include "tps_enhcbuf.hpp"
#include "decoder/enhcbufdec.hpp"
#include "decoder/vopsedec.hpp"
//#include "decoder/tps_vopsedec.hpp" // deleted by Sharp (98/2/12)
#ifndef __GLOBAL_VAR_
#define __GLOBAL_VAR_
#endif

#include "global.hpp"

#ifdef __MFC_
#ifdef _DEBUG
#undef THIS_FILE
static char BASED_CODE THIS_FILE[] = __FILE__;
#endif

#define new DEBUG_NEW
#endif // __MFC_

#define HELP_MESSAGE "Usage: DECODER bitstream_file [enhn_layer_bitstream_file] output_file width height"
/* NBIT: add parameter nBits
Void dumpFrame (const CVOPU8YUVBA* pvopcQuant, FILE* pfYUV, FILE* pfSeg, AlphaUsage, CRct& rct);
*/
Void dumpFrame (const CVOPU8YUVBA* pvopcQuant, FILE* pfYUV, FILE* pfSeg, AlphaUsage, CRct& rct, UInt nBits, Int iAlphaScale);


#define MAX_TIME	99999			// for bitstreams stopped at BVOP


// Main Routin
void main(int argc, char* argv[])
{
	Bool bScalability = FALSE;

	if (argc < 5 || argc > 6) {
		printf ("%s\n", HELP_MESSAGE);
		exit(1); //return;
	}
	else if (argc == 6)
		bScalability = TRUE;

	CRct rctDisplay (0, 0, atoi (argv [3 + bScalability]), atoi (argv [4 + bScalability]));
	FILE *pfBits;
	if ((pfBits = fopen (argv [1], "rb")) == NULL ){
		printf ("Bitstream File Not Found\n");
		exit (1);
	}
	fclose (pfBits);
	if (bScalability == TRUE) {
		if ((pfBits = fopen (argv [2], "rb")) == NULL )	{
			printf ("Bitstream File for Enhancement Layer Not Found\n");
			exit (1);
		}
	}
	fclose (pfBits);

// FIX 
	Bool bTemporalScalability;

	if (bScalability){
		CVideoObjectDecoder* pvodec_tps[2];
		cout << "Checking scalability type...\n";
		pvodec_tps[BASE_LAYER] = new CVideoObjectDecoder (argv[1], rctDisplay.width, rctDisplay.height (), NULL);
		pvodec_tps[ENHN_LAYER] = new CVideoObjectDecoder (argv[2], rctDisplay.width, rctDisplay.height (), NULL);
// begin: deleted by Sharp (98/2/12)
//		CVideoObjectDecoderTPS* pvodec_tps[2];
//		pvodec_tps[BASE_LAYER] = new CVideoObjectDecoderTPS (argv[1], rctDisplay.width, rctDisplay.height ());
//		pvodec_tps[ENHN_LAYER] = new CVideoObjectDecoderTPS (argv[2], rctDisplay.width, rctDisplay.height ());
// end: deleted by Sharp (98/2/12)
		Time FirstBase = pvodec_tps[BASE_LAYER] -> senseTime ();
		Time FirstEnhn = pvodec_tps[ENHN_LAYER] -> senseTime ();

		if ( FirstBase != FirstEnhn ){
			bTemporalScalability = 1;
			cout << "Starting temporal scalability decoding...\n\n";
		}
		else{
			bTemporalScalability = 0;
			cout << "Starting spatial scalability decoding...\n\n";
		}
		
		delete pvodec_tps[BASE_LAYER];
		delete pvodec_tps[ENHN_LAYER];
	}

	if ( !bScalability || !bTemporalScalability ){
		CVideoObjectDecoder* pvodec [2];
		Bool bSpatialScalable = FALSE;
		CRct rctDisplayBackup = rctDisplay; // for recovery from sprite decoding
		pvodec [BASE_LAYER] = new CVideoObjectDecoder (argv [1], rctDisplay.width, rctDisplay.height(), &bSpatialScalable);

	// NBIT: get nBits information
		UInt nBits = pvodec [BASE_LAYER]->volmd ().nBits;

		if (bScalability)
			pvodec [ENHN_LAYER] = new CVideoObjectDecoder (argv [2], rctDisplay.width, rctDisplay.height(), &bSpatialScalable);

		if (pvodec [BASE_LAYER] -> fSptUsage () == 1) {
				// decode the initial sprite: //wchen: instead of calling decode ()
			pvodec [BASE_LAYER] -> decodeInitSprite ();
				//if (pvodec [BASE_LAYER]->iNumOfPnts () > 0) {
					// change m_pvopcRefQ1 to m_pvopcSptQ for warping
					//pvodec [BASE_LAYER] -> swapRefQ1toSpt ();
					// restore m_pvopcCurrQ size to the normal one
					//pvodec [BASE_LAYER] ->  changeSizeofCurrQ (rctDisplayBackup);
				//}
		}

		FILE* pfReconYUV [2];
		Char pchTmp [100];
		sprintf (pchTmp, "%s.yuv", argv [2 + bScalability]);
		pfReconYUV [BASE_LAYER] = fopen (pchTmp, "wb");
		assert (pfReconYUV [BASE_LAYER] != NULL);
		if (bScalability)	{
			sprintf (pchTmp, "%s_e.yuv", argv [2 + bScalability]);
			pfReconYUV [ENHN_LAYER]= fopen (pchTmp, "wb");
			assert (pfReconYUV [ENHN_LAYER] != NULL);
		}
		sprintf (pchTmp, "%s.seg", argv [2 + bScalability]);
		FILE* pfReconSeg = NULL;
		if (pvodec [BASE_LAYER]-> volmd ().fAUsage != RECTANGLE)	{
			pfReconSeg = fopen (pchTmp, "wb");
			assert (pfReconSeg != NULL);
		}

		Int iEof = 1;
		Int nFrames = 0;
		Int NextVopPredType = 0;
		CVOPU8YUVBA* pvopcBaseQuant;
		const CVOPU8YUVBA* pvopcQuant;
#ifdef __PC_COMPILER_
		Int tickBegin = ::GetTickCount ();
#endif // __PC_COMPILER_
		while (iEof != EOF && NextVopPredType != -1) {
			iEof = pvodec [BASE_LAYER] -> decode ();
			if(iEof==FALSE)
				dumpNonCodedFrame(pfReconYUV [BASE_LAYER], pfReconSeg, rctDisplay, nBits);
			if (pvodec[BASE_LAYER]->vopmd().bInterlace) {
				// Re-order VOPs for display.  Write the previous anchor VOP (I or P),
				// skipping the first I-VOP, or the current B-VOP.  This has nothing 
				// to do with interlaced video.  The else clause of this 'if' should be
				// deleted but I don't want to break the old test cases.
				//                                                 (Bob Eifrig)
				if (iEof != EOF) {
					if (nFrames == 0) {
						nFrames++;
						continue;
					}
					pvopcQuant = (pvodec[BASE_LAYER]->vopmd().vopPredType == BVOP) ?
						pvodec [BASE_LAYER]->pvopcReconCurr () :
						pvodec [BASE_LAYER]->pvopcRefQPrev();
				} else
					pvopcQuant = pvodec [BASE_LAYER]->pvopcRefQLater();
	/* NBIT: add parameter nBits
				dumpFrame (pvopcQuant, pfReconYUV [BASE_LAYER], pfReconSeg, pvodec[BASE_LAYER]->volmd().fAUsage, rctDisplay);
	*/
				dumpFrame (pvopcQuant, pfReconYUV [BASE_LAYER], pfReconSeg, pvodec[BASE_LAYER]->volmd().fAUsage, rctDisplay,
					nBits, pvodec [BASE_LAYER]->vopmd().iVopConstantAlphaValue);
			} else {
				pvopcQuant = pvodec [BASE_LAYER]->pvopcReconCurr ();
				if (iEof != EOF && iEof!=FALSE)
	/* NBIT: add parameter nBits
					dumpFrame (pvopcQuant, pfReconYUV [BASE_LAYER], pfReconSeg, pvodec[BASE_LAYER]->volmd().fAUsage, rctDisplay);
	*/
					dumpFrame (pvopcQuant, pfReconYUV [BASE_LAYER], pfReconSeg, pvodec[BASE_LAYER]->volmd().fAUsage, rctDisplay,
						nBits, pvodec [BASE_LAYER]->vopmd().iVopConstantAlphaValue);
			}
			if (bSpatialScalable == TRUE) {		
				pvopcBaseQuant = new CVOPU8YUVBA(*pvopcQuant, pvopcQuant->whereY());	
				NextVopPredType = pvodec[BASE_LAYER]->ReadNextVopPredType();
			}
			while( NextVopPredType == BVOP){		
				iEof = pvodec [BASE_LAYER] -> decode ();
				if(iEof==FALSE)
					dumpNonCodedFrame(pfReconYUV [BASE_LAYER], pfReconSeg, rctDisplay, nBits);

				pvopcQuant = pvodec [BASE_LAYER]->pvopcReconCurr ();
				if (iEof != EOF && iEof!=FALSE)
	/* NBIT: add parameter nBits
					dumpFrame (pvopcQuant, pfReconYUV [BASE_LAYER], pfReconSeg, pvodec[BASE_LAYER]->volmd().fAUsage, rctDisplay);
	*/
					dumpFrame (pvopcQuant, pfReconYUV [BASE_LAYER], pfReconSeg, pvodec[BASE_LAYER]->volmd().fAUsage, rctDisplay,
						nBits, pvodec [BASE_LAYER]->vopmd().iVopConstantAlphaValue);
				NextVopPredType = pvodec[BASE_LAYER]->ReadNextVopPredType();
				if (bSpatialScalable == TRUE) {		
					iEof = pvodec [ENHN_LAYER] -> decode (pvopcQuant);
					if(iEof==FALSE)
						dumpNonCodedFrame(pfReconYUV [ENHN_LAYER], NULL, rctDisplay * 2, nBits);

					if (iEof != EOF && iEof!=FALSE)
	/* NBIT: add parameter nBits
						dumpFrame (pvodec [ENHN_LAYER]->pvopcReconCurr (), pfReconYUV [ENHN_LAYER], NULL, RECTANGLE, rctDisplay * 2);
	*/
						dumpFrame (pvodec [ENHN_LAYER]->pvopcReconCurr (), pfReconYUV [ENHN_LAYER], NULL, RECTANGLE, rctDisplay * 2,
							nBits, pvodec [BASE_LAYER]->vopmd().iVopConstantAlphaValue); // still base layer
				}
			}
			if (bSpatialScalable == TRUE)	{
				iEof = pvodec [ENHN_LAYER] -> decode (pvopcBaseQuant);
				delete pvopcBaseQuant;
				if(iEof==FALSE)
					dumpNonCodedFrame(pfReconYUV [ENHN_LAYER], NULL, rctDisplay * 2, nBits);

				if (iEof != EOF && iEof!=FALSE)
	/* NBIT: add parameter nBits
					dumpFrame (pvodec [ENHN_LAYER]->pvopcReconCurr (), pfReconYUV [ENHN_LAYER], NULL, RECTANGLE, rctDisplay * 2);
	*/
					dumpFrame (pvodec [ENHN_LAYER]->pvopcReconCurr (), pfReconYUV [ENHN_LAYER], NULL, RECTANGLE, rctDisplay * 2,
						nBits, pvodec [BASE_LAYER]->vopmd().iVopConstantAlphaValue);
			}
			nFrames++;
		}
#ifdef __PC_COMPILER_
		Int tickAfter = ::GetTickCount ();
		printf ("Total time: %d\n", tickAfter - tickBegin);
		Double dAverage = (Double) (tickAfter - tickBegin) / (Double) (nFrames -1 );
		printf ("Total frame %d\tAverage time: %.6lf\n", nFrames, dAverage);
		printf ("FPS %.6lf\n", 1000.0 / dAverage);
#endif // __PC_COMPILER_

		fclose (pfReconYUV [BASE_LAYER]);
		if (pvodec [BASE_LAYER]->volmd ().fAUsage != RECTANGLE)
			fclose (pfReconSeg);
		delete pvodec [BASE_LAYER];
		if (bScalability)	{
			fclose (pfReconYUV [ENHN_LAYER]);
			delete pvodec [ENHN_LAYER];
		}

	}
	else { // loop for temporal scalability
		CVideoObjectDecoder* pvodec[2]; // modified by Sharp (98/2/12)
//		puts("Entering TPS loop"); // deleted by Sharp (98/2/12)
//	CRct rctDisplayBackup = rctDisplay;									// for recovery from sprite decoding
																// for *S*P*R*I*T*E*
//	printf("\n **** Start decoding of <        Base_Layer data > **** \n");  fflush(stdout); // deleted by Sharp (98/2/12)
		pvodec[BASE_LAYER] = new CVideoObjectDecoder (argv[1], rctDisplay.width, rctDisplay.height (), NULL); // modified by Sharp (98/2/12)
	//	if (pvodec [BASE_LAYER] -> fSptUsage () == 1) {										// for *S*P*R*I*T*E*
	//		//pvodec [BASE_LAYER] -> setRctDisp (rctDisplayBackup);								//
	//		// decode the initial sprite											//
	//		pvodec [BASE_LAYER] -> decode ();										//
	//		if (pvodec [BASE_LAYER]->volmd ().fAUsage != RECTANGLE) {							//
	//			// change m_pvopcRefQ1 to m_pvopcSptQ for warping							//
	//			pvodec [BASE_LAYER] -> swapRefQ1toSpt ();								//
	//			// restore m_pvopcCurrQ size to the normal one								//
	//			pvodec [BASE_LAYER] ->  changeSizeofCurrQ (rctDisplayBackup);						//
	//		}														//
	//	}															//

		if(bScalability){
	//		printf(" ****                   < Enhancement_Layer data > **** \n");  fflush(stdout); // deleted by Sharp (98/2/12)
			pvodec[ENHN_LAYER] = new CVideoObjectDecoder (argv[2], rctDisplay.width, rctDisplay.height (), NULL);
			// for back/forward shape
			pvodec[ENHN_LAYER]->rgpbfShape[0] = new CVideoObjectDecoder (rctDisplay.width, rctDisplay.height ());
			pvodec[ENHN_LAYER]->rgpbfShape[1] = new CVideoObjectDecoder (rctDisplay.width, rctDisplay.height ());
	// begin: deleted by Sharp (98/2/12)
	//		pvodec[ENHN_LAYER] = new CVideoObjectDecoderTPS (argv[2], rctDisplay.width, rctDisplay.height ());
	//		// for back/forward shape
	//		pvodec[ENHN_LAYER]->rgpbfShape[0] = new CVideoObjectDecoderTPS (rctDisplay.width, rctDisplay.height ());
	//		pvodec[ENHN_LAYER]->rgpbfShape[1] = new CVideoObjectDecoderTPS (rctDisplay.width, rctDisplay.height ());
	// end: deleted by Sharp (98/2/12)
			// copy pointers
			pvodec[ENHN_LAYER]->copyTobfShape ();
		}

		Char pchTmp [100];
#ifndef __OUT_ONE_FRAME_
		FILE* pfReconYUV [2];
		sprintf (pchTmp, "%s.yuv", argv [2 + bScalability]);
		pfReconYUV [BASE_LAYER] = fopen (pchTmp, "wb");
		assert (pfReconYUV [BASE_LAYER] != NULL);
		if (bScalability) {
			sprintf (pchTmp, "%s_e.yuv", argv [2 + bScalability]);
			pfReconYUV [ENHN_LAYER]= fopen (pchTmp, "wb");
			assert (pfReconYUV [ENHN_LAYER] != NULL);
		}
		sprintf (pchTmp, "%s.seg", argv [2 + bScalability]);
		FILE* pfReconSeg [2];
		if (pvodec [BASE_LAYER]-> volmd ().fAUsage != RECTANGLE) {
			pfReconSeg [BASE_LAYER] = fopen (pchTmp, "wb");
			assert (pfReconSeg [BASE_LAYER] != NULL);
		}
		if (bScalability) {
			if (pvodec [ENHN_LAYER]-> volmd ().fAUsage != RECTANGLE) {
				sprintf (pchTmp, "%s_e.seg", argv [2 + bScalability]);
				pfReconSeg [ENHN_LAYER]= fopen (pchTmp, "wb");
				assert (pfReconSeg [ENHN_LAYER] != NULL);
			}
		}

		if (bScalability)
			if (pvodec [ENHN_LAYER]-> volmd ().iEnhnType == 1) {
				FILE *pfTemp;
				sprintf (pchTmp, "%s_bgc.yuv", argv [2 + bScalability]);
				pfTemp = fopen (pchTmp, "wb");			 // clear file pointer for background composition
				fclose(pfTemp);
			}
#endif
		// for back/forward shape output 
		FILE* pfTestSeg;
		if (bScalability)
			if (pvodec [ENHN_LAYER]-> volmd ().iEnhnType == 1) {
				sprintf (pchTmp, "%s.bfseg", argv [2 + bScalability]);
				pfTestSeg = fopen (pchTmp, "wb"); //reconstructed Seg file(bfShape)
			}

		int iEof = 1;
#ifdef __PC_COMPILER_
		Int tickBegin = ::GetTickCount ();
#endif // __PC_COMPILER_
		Time tPvopBase, tNextBase, tNextEnhc;
		int nBaseFrames = 0;
		int nEnhcFrames = 0;

		while (iEof != EOF) {
	//		if (pvodec [BASE_LAYER]->fSptUsage () == 1) {									// for *S*P*R*I*T*E*
	//			iEof = pvodec [BASE_LAYER] -> decodeSpt ();								//
	//			continue;												//
	//		}														//

			if (pvodec[BASE_LAYER] -> getPredType () != BVOP) {			// for bitstreams stopped at BVOP
				iEof = pvodec[BASE_LAYER] -> decode ();  if(iEof == EOF) break;
	//			if (!(pvodec[BASE_LAYER] -> m_coded)) continue;
				if(bScalability)
					pvodec[BASE_LAYER] -> updateBuffVOPsBase (pvodec[ENHN_LAYER]);
				nBaseFrames ++; //printf(" BaseFrame update I/P1 -> %d\n", nBaseFrames);

				// Output BASE_LAYER
				if (iEof != EOF) {
#ifndef __OUT_ONE_FRAME_
					pvodec [BASE_LAYER]->dumpDataAllFrame (pfReconYUV [BASE_LAYER], pfReconSeg [BASE_LAYER], rctDisplay);
#else
					pvodec [BASE_LAYER]->dumpDataOneFrame(argv, bScalability, rctDisplay);
#endif
				}

	//			if (pvodec[BASE_LAYER] -> getPredType () == IVOP)  continue;
				tPvopBase = pvodec[BASE_LAYER] -> getTime ();
			}
			else {									// for bitstreams stopped at BVOP
				tPvopBase = MAX_TIME;
				pvodec[BASE_LAYER] -> copyRefQ1ToQ0 ();
			}

			tNextBase = pvodec[BASE_LAYER] -> senseTime ();
			//printf("          tPvopBase = %d,  tNextBase = %d\n", tPvopBase, tNextBase);
			while((tPvopBase > tNextBase) && (tNextBase != EOF)) {
					iEof = pvodec[BASE_LAYER] -> decode ();
					if(bScalability)
						pvodec[BASE_LAYER] -> updateBuffVOPsBase (pvodec[ENHN_LAYER]);
					nBaseFrames ++; //printf(" BaseFrame update B -> %d\n", nBaseFrames);

					// Output BASE_LAYER
					if (iEof != EOF) {
#ifndef __OUT_ONE_FRAME_
						pvodec [BASE_LAYER]->dumpDataAllFrame (pfReconYUV [BASE_LAYER], pfReconSeg [BASE_LAYER], rctDisplay);
#else
						pvodec [BASE_LAYER]->dumpDataOneFrame(argv, bScalability, rctDisplay);
#endif
					}

				if(bScalability && tNextEnhc != EOF) {
					tNextEnhc = pvodec[ENHN_LAYER] -> senseTime ();
					//printf("          tPvopBase(B) = %d,  tNextEnhc = %d\n", tPvopBase, tNextEnhc);
					while((tNextBase > tNextEnhc) && (tNextEnhc != EOF)) {
						iEof = pvodec[ENHN_LAYER] -> decode ();
	//					if (!(pvodec[ENHN_LAYER] -> m_coded)) continue;
						nEnhcFrames ++; //printf(" EnhcFrame update -> %d\n", nEnhcFrames);

					if (pvodec [ENHN_LAYER] -> volmd ().fAUsage != RECTANGLE) {
						// for background composition
						if(pvodec [ENHN_LAYER] -> volmd ().iEnhnType == 1){ // should be changed to background_composition flag later.
							printf("============== background composition (1)\n");
							pvodec [ENHN_LAYER] -> BackgroundComposition(argv, bScalability, rctDisplay.width, rctDisplay.height ());
						}
						// for back/forward shape
						if(pvodec [ENHN_LAYER] -> vopmd ().iLoadBakShape){
							printf("---------- output backward shape\n");
							const CVOPU8YUVBA* pvopcQuant = pvodec [ENHN_LAYER]->rgpbfShape[0]->pvopcReconCurr ();
							pvopcQuant->getPlane (BY_PLANE)->dump (pfTestSeg, rctDisplay);
						}
						if(pvodec [ENHN_LAYER] -> vopmd ().iLoadForShape){
							printf("---------- output forward shape\n");
							const CVOPU8YUVBA* pvopcQuant = pvodec [ENHN_LAYER]->rgpbfShape[1]->pvopcReconCurr ();
							pvopcQuant->getPlane (BY_PLANE)->dump (pfTestSeg, rctDisplay);
						}
					}

						// Output ENHN_LAYER
						if (iEof != EOF) {
#ifndef __OUT_ONE_FRAME_
							pvodec [ENHN_LAYER]->dumpDataAllFrame (pfReconYUV [ENHN_LAYER], pfReconSeg [ENHN_LAYER], rctDisplay);
#else
							pvodec [ENHN_LAYER]->dumpDataOneFrame(argv, bScalability, rctDisplay);
#endif
						}

					tNextEnhc = pvodec[ENHN_LAYER] -> senseTime ();
					//printf("          tPvopBase(B) = %d,  tNextEnhc = %d\n", tPvopBase, tNextEnhc);
					}
				}
				tNextBase = pvodec[BASE_LAYER] -> senseTime ();
				//printf("          tPvopBase = %d,  tNextBase = %d\n", tPvopBase, tNextBase);
			}
			if(bScalability && tNextEnhc != EOF) {
				pvodec[ENHN_LAYER] -> bufferB2flush ();

				tNextEnhc = pvodec[ENHN_LAYER] -> senseTime ();
				//printf("          tPvopBase(P) = %d,  tNextEnhc = %d\n", tPvopBase, tNextEnhc);
				while((tPvopBase > tNextEnhc) && (tNextEnhc != EOF)) {
					iEof = pvodec[ENHN_LAYER] -> decode ();
	//				if (!(pvodec[ENHN_LAYER] -> m_coded)) continue;
					nEnhcFrames ++; //printf(" EnhcFrame update -> %d\n", nEnhcFrames);

				if (pvodec [ENHN_LAYER] -> volmd ().fAUsage != RECTANGLE) {
					// for background composition
					if(pvodec [ENHN_LAYER] -> volmd ().iEnhnType == 1){ // should be changed to background_composition flag later.
						printf("============== background composition (2)\n");
						pvodec [ENHN_LAYER] -> BackgroundComposition(argv, bScalability, rctDisplay.width, rctDisplay.height ());
					}
					// for back/forward shape
					if(pvodec [ENHN_LAYER] -> vopmd ().iLoadBakShape){
						printf("---------- output backward shape\n");
						const CVOPU8YUVBA* pvopcQuant = pvodec [ENHN_LAYER]->rgpbfShape[0]->pvopcReconCurr ();
						pvopcQuant->getPlane (BY_PLANE)->dump (pfTestSeg, rctDisplay);
					}
					if(pvodec [ENHN_LAYER] -> vopmd ().iLoadForShape){
						printf("---------- output forward shape\n");
						const CVOPU8YUVBA* pvopcQuant = pvodec [ENHN_LAYER]->rgpbfShape[1]->pvopcReconCurr ();
						pvopcQuant->getPlane (BY_PLANE)->dump (pfTestSeg, rctDisplay);
					}
				}

					// Output ENHN_LAYER
					if (iEof != EOF) {
#ifndef __OUT_ONE_FRAME_
						pvodec [ENHN_LAYER]->dumpDataAllFrame (pfReconYUV [ENHN_LAYER], pfReconSeg [ENHN_LAYER], rctDisplay);
#else
						pvodec [ENHN_LAYER]->dumpDataOneFrame(argv, bScalability, rctDisplay);
#endif
					}
					tNextEnhc = pvodec[ENHN_LAYER] -> senseTime ();
					//printf("          tPvopBase(P) = %d,  tNextEnhc = %d\n", tPvopBase, tNextEnhc);
				}

				if((tPvopBase == tNextEnhc) && (tNextEnhc != EOF)) {						// for Spatial Scalability
					pvodec[ENHN_LAYER] -> copyBufP2ToB1 ();							//
					iEof = pvodec[ENHN_LAYER] -> decode ();							//
					nEnhcFrames ++; //printf(" EnhcFrame update -> %d\n", nEnhcFrames);			//
				}												//

				pvodec[ENHN_LAYER] -> bufferB1flush ();

				// Enhancement Layer after Base BVOP
				if (tNextBase == EOF && tNextEnhc != EOF) {
					pvodec[ENHN_LAYER] -> copyBufP2ToB1 ();
					while(tNextEnhc != EOF) {
						iEof = pvodec[ENHN_LAYER] -> decode ();
						nEnhcFrames ++; //printf(" EnhcFrame update -> %d\n", nEnhcFrames);

						// Output ENHN_LAYER
						if (iEof != EOF) {
#ifndef __OUT_ONE_FRAME_
							pvodec [ENHN_LAYER]->dumpDataAllFrame (pfReconYUV [ENHN_LAYER], pfReconSeg [ENHN_LAYER], rctDisplay);
#else
							pvodec [ENHN_LAYER]->dumpDataOneFrame(argv, bScalability, rctDisplay);
#endif
						}
						tNextEnhc = pvodec[ENHN_LAYER] -> senseTime ();
						//printf("          tPvopBase(P) = %d,  tNextEnhc = %d\n", tPvopBase, tNextEnhc);
					}
					pvodec[ENHN_LAYER] -> bufferB1flush ();
				}
			}

	//		pvodec[BASE_LAYER] -> bufferP1flush ();
			if (bScalability)
				pvodec[ENHN_LAYER] -> bufferP1flush ();

			iEof = tNextBase;
		}

#ifdef __PC_COMPILER_
		Int tickAfter = ::GetTickCount ();
		printf ("Total time: %d\n", tickAfter - tickBegin);
		Double dAverage = (Double) (tickAfter - tickBegin) / (Double) (nBaseFrames + nEnhcFrames);
		printf ("Total frame %d\tAverage time: %.6lf\n", nBaseFrames + nEnhcFrames, dAverage);
		printf ("FPS %.6lf\n", 1000.0 / dAverage);
#endif // __PC_COMPILER_

#ifndef __OUT_ONE_FRAME_
		fclose (pfReconYUV[BASE_LAYER]);
		if (pvodec[BASE_LAYER]->volmd ().fAUsage != RECTANGLE)
			fclose (pfReconSeg [BASE_LAYER]);
#endif
		delete pvodec[BASE_LAYER];
		if (bScalability) {
#ifndef __OUT_ONE_FRAME_
			fclose (pfReconYUV[ENHN_LAYER]);
			if (pvodec[ENHN_LAYER]->volmd ().fAUsage != RECTANGLE)
				fclose (pfReconSeg [ENHN_LAYER]);
#endif
			delete pvodec[ENHN_LAYER];
		}

	}

	return;
}

/* NBIT: add parameter nBits
Void dumpFrame (const CVOPU8YUVBA* pvopcQuant, FILE* pfYUV, FILE* pfSeg, AlphaUsage fAUsage, CRct& rct)
*/
Void dumpFrame (const CVOPU8YUVBA* pvopcQuant, FILE* pfYUV, FILE* pfSeg, AlphaUsage fAUsage, CRct& rct, UInt nBits, Int iAlphaScale)
{
	pvopcQuant->getPlane (Y_PLANE)->dump (pfYUV, rct);
	pvopcQuant->getPlane (U_PLANE)->dump (pfYUV, rct / 2);
	pvopcQuant->getPlane (V_PLANE)->dump (pfYUV, rct / 2);
	if(pfSeg!=NULL)
	{
		if (fAUsage == ONE_BIT)
			pvopcQuant->getPlane (BY_PLANE)->dump (pfSeg, rct, iAlphaScale);
		else if (fAUsage == EIGHT_BIT)
			pvopcQuant->getPlane (A_PLANE)->dumpWithMask (pfSeg, pvopcQuant->getPlane (BY_PLANE),
				rct, iAlphaScale);
	}
	return;
}
