/* $Id: dataStruct.h,v 1.19 1998/02/20 21:14:05 hjlee Exp $ */
/****************************************************************************/
/*   MPEG4 Visual Texture Coding (VTC) Mode Software                        */
/*                                                                          */
/*   This software was developed by                                         */
/*   Sarnoff Coporation                   and    Texas Instruments          */
/*   Iraj Sodagar   (iraj@sarnoff.com)           Jie Liang (liang@ti.com)   */
/*   Hung-Ju Lee    (hjlee@sarnoff.com)                                     */
/*   Paul Hatrack   (hatrack@sarnoff.com)                                   */
/*   Shipeng Li     (shipeng@sarnoff.com)                                   */
/*   Bing-Bing Chai (bchai@sarnoff.com)                                     */
/*                                                                          */
/* In the course of development of the MPEG-4 standard. This software       */
/* module is an implementation of a part of one or more MPEG-4 tools as     */
/* specified by the MPEG-4 standard.                                        */
/*                                                                          */
/* The copyright of this software belongs to ISO/IEC. ISO/IEC gives use     */
/* of the MPEG-4 standard free license to use this  software module or      */
/* modifications thereof for hardware or software products claiming         */
/* conformance to the MPEG-4 standard.                                      */
/*                                                                          */
/* Those intending to use this software module in hardware or software      */
/* products are advised that use may infringe existing  patents. The        */
/* original developers of this software module and their companies, the     */
/* subsequent editors and their companies, and ISO/IEC have no liability    */
/* and ISO/IEC have no liability for use of this software module or         */
/* modification thereof in an implementation.                               */
/*                                                                          */
/* Permission is granted to MPEG memebers to use, copy, modify,             */
/* and distribute the software modules ( or portions thereof )              */
/* for standardization activity within ISO/IEC JTC1/SC29/WG11.              */
/*                                                                          */
/* Copyright (C) 1998  Sarnoff Coporation and Texas Instruments             */ 
/****************************************************************************/

/************************************************************/
/*     Sarnoff Very Low Bit Rate Still Image Coder          */
/*     Copyright 1995, 1996, 1997, 1998 Sarnoff Corporation */
/************************************************************/

#ifndef DATA_STRUCT_H
#define DATA_STRUCT_H

#ifndef QUANT_H
#include "quant.h"
#endif QUANT_H

#define UNIT32      unsigned int
#define UNIT16      unsigned short int
#define UNIT8       unsigned char

#define LINT        long int
#define INT         int
#define SINT        short int

#define DATA        SINT

#define MAXLEV      12
#define NCOLOR      3

#define SINGLE_Q    1
#define MULTIPLE_Q  2
#define BILEVEL_Q   3

/* zero tree symbols - if changed please change MapTypeToText variable */
#define IZ       0  /* Isolated Zero */
#define VAL      1  /* Value */
#define ZTR      2  /* Zero-Tree Root */
#define VZTR     3  /* Valued Zero-Tree Root */
#define ZTR_D    4  /* Parent has type of ZTR, VZTR, or ZTR_D (not coded) */
#define VLEAF    5  /* Leaf coefficient with non-zero value (not coded) */
#define ZLEAF    6  /* Leaf coefficient with zero value (not coded) */
#define UNTYPED  7  /* so far only for clearing of ZTR_Ds */

#define MAXDECOMPLEV 10
#define INF_RES       1024


#define MONO 1

#define FULLSIZE 1
#define PROGRESSIVE 0
#define MASK_VAL (0xff)



/* QValArithModel field types - if changed please change mapArithModelToText
   variable */
#define ACM_NONE  0 /* When there's no value to code */
#define ACM_SKIP  1
#define ACM_ROOT  2
#define ACM_VALZ  3
#define ACM_VALNZ 4
#define ACM_RESID 5 /* should have one for each set of coeffs arising in 
		       different initial spatial layers */
#define ACM_DC    6

typedef struct 
{
  /* not updated */
  DATA           wvt_coeff;        /* Original value in encoding.*/
  DATA           rec_coeff;        /* Reconstructed value in decoding 
				      Put here for comparing with originals
				      at decoder. Reconstructed values can
				      be put in original when memory is an
				      issue and we don't want stats to be
				      computed at decoder. */

  /* updated by quantization */
  DATA           quantized_value; /* quantized value                    */
  quantState     qState;          /* state of quantizer for coefficient */

  /* updated by marking */
  UNIT8          state;          /* state of coefficient                   */
  UNIT8          type;           /* MZTE tree types: ZTR, IZ, VZTR, or VAL */
  UNIT8          skip;  /* Skip coding of coefficient value (not type)     */

  /* updated by Shipeng */
  UNIT8          mask;

} COEFFINFO;


typedef struct {
  int   num_ZTR;
  int   num_VZTR;
  int   num_VAL;
} STATINFO;


typedef struct  {
  int    height;
  int    width;
  UNIT8  *mask;
  void   *data; 
} PICTURE;

typedef struct {
   SINT         quant;
   UNIT8        allzero;
   int          root_max;   /* three maximum values for AC magnitude coding */
   int          valz_max;
   int          valnz_max;
   int          residual_max;
   STATINFO     stat;
} SNR_IMAGE;

typedef struct {
   SNR_IMAGE  snr_image; 
} SNR_LAYER;

typedef struct {
   SINT         height;
   SINT         width;
   SINT         SNR_scalability_levels;
   COEFFINFO    **coeffinfo;
   SNR_LAYER    SNRlayer;
} SPATIAL_LAYER;


typedef struct snr_param {
   int SNR_scalability_levels;
   int *Quant;
} SNR_PARAM;


typedef struct wvt_codec 
{
  int   bit_depth; /* number bits per pixel (spatial) */
  int   colors;    /* number of color components: 0 = mono, 3=yuv */
  int   color_format;  /* 4:4:4, 4:2:2, or 4:2:0 ???? */
  int   sa_dwt;
  PICTURE         *Image; /* spatial source */
  PICTURE         *SegImage; /* spatial source */
  
  int             wvtType;      /* Type of filter */
  int             wvtDecompLev; 

  int             mean[NCOLOR]; /* mean of wvt coeffs in DC band ???? */
  int             Qdc[NCOLOR];  
  int             offset_dc;
  int             max_dc; /* max quantized DC coeff - pre-shifting */
  int             dcWidth, dcHeight;
  
  int             width, height; 
  int             spatial_scalability_levels;
  unsigned char   quantization_type;
  unsigned char   SNR_start_code_enable;
  SPATIAL_LAYER   SPlayer[NCOLOR];
  
  SNR_PARAM       *Qinfo[NCOLOR];
  
  int             target_spatial_levels;
  int             target_snr_levels;
  int             target_bitrate;
  int             deringWindowSize;
  int             deringThreshold;
  
  int             origin_x;
  int             origin_y;
  int             real_width;
  int             real_height;

  /*-----------------------*/
  int curSpatialLev;
  int curSNRLev;
  int curColor;
  
} WVT_CODEC;


/*-----------   SOL parameter -------------------*/


typedef struct sol_parameters {
  int            bit_depth;
  int            colors;
  int            sa_dwt;
  PICTURE       *Image;
  PICTURE       *SegImage;
  char          *ImagePath;
  char          *SegPath;
  int            wvtType;
  int            wvtDecompLev;
  int            QdcY;
  int            QdcUV;
  int            spatial_scalability_levels;
  int            quantization_type;
  int            SNR_start_code_enable;
  SNR_PARAM     *SPlayer[NCOLOR];
  unsigned char *buffer;
  int            target_bitrate;
  int            target_spatial_levels;
  int            target_snr_levels;
  int            deringWindowSize;
  int            deringThreshold;
  /* for shape coding */
  int            alphaTH;
  int            change_CR_disable;
  int            STO_const_alpha;
  unsigned char  STO_const_alpha_value;
  int            singleBitFile;
  char          *bitFile;
  char          *bitFileAC;
} SOL_PARAMETERS;



/*-----------   SOL parameter for decoder -------------------*/


typedef struct sol_parameters_dec {
  int            bit_depth;
  int            colors;
  int            sa_dwt;
  PICTURE       *SegImage;
  char          *SegPath;
  int            width;
  int            height;
  int            wvtType;
  int            wvtDecompLev;
  int            target_bitrate;
  int            target_spatial_levels;
  int            target_snr_levels;
  int            deringWindowSize;
  int            deringThreshold;
  int            singleBitFile;
  char          *bitFile;
  char          *bitFileAC;

} SOL_PARAMETERS_DEC;

#endif DATA_STRUCT_H

