/* $Id: main.c,v 1.42 1998/02/20 21:14:14 hjlee Exp $ */
/****************************************************************************/
/*   MPEG4 Visual Texture Coding (VTC) Mode Software                        */
/*                                                                          */
/*   This software was developed by                                         */
/*   Sarnoff Coporation                   and    Texas Instruments          */
/*   Iraj Sodagar   (iraj@sarnoff.com)           Jie Liang (liang@ti.com)   */
/*   Hung-Ju Lee    (hjlee@sarnoff.com)                                     */
/*   Paul Hatrack   (hatrack@sarnoff.com)                                   */
/*   Shipeng Li     (shipeng@sarnoff.com)                                   */
/*   Bing-Bing Chai (bchai@sarnoff.com)                                     */
/*                                                                          */
/* In the course of development of the MPEG-4 standard. This software       */
/* module is an implementation of a part of one or more MPEG-4 tools as     */
/* specified by the MPEG-4 standard.                                        */
/*                                                                          */
/* The copyright of this software belongs to ISO/IEC. ISO/IEC gives use     */
/* of the MPEG-4 standard free license to use this  software module or      */
/* modifications thereof for hardware or software products claiming         */
/* conformance to the MPEG-4 standard.                                      */
/*                                                                          */
/* Those intending to use this software module in hardware or software      */
/* products are advised that use may infringe existing  patents. The        */
/* original developers of this software module and their companies, the     */
/* subsequent editors and their companies, and ISO/IEC have no liability    */
/* and ISO/IEC have no liability for use of this software module or         */
/* modification thereof in an implementation.                               */
/*                                                                          */
/* Permission is granted to MPEG memebers to use, copy, modify,             */
/* and distribute the software modules ( or portions thereof )              */
/* for standardization activity within ISO/IEC JTC1/SC29/WG11.              */
/*                                                                          */
/* Copyright (C) 1998  Sarnoff Coporation and Texas Instruments             */ 
/****************************************************************************/

/************************************************************/
/*     Sarnoff Very Low Bit Rate Still Image Coder          */
/*     Copyright 1995, 1996, 1997, 1998 Sarnoff Corporation */
/************************************************************/

/****************************************************************************/
/*     Texas Instruments Predictive Embedded Zerotree (PEZW) Image Codec    */
/*     Copyright 1996, 1997, 1998 Texas Instruments	      		    */
/****************************************************************************/

#include <stdio.h>
#include <math.h>
#include "dataStruct.h"

#define DEFINE_GLOBALS
#include "globals.h"

#include "Utils.h"
#include "QM.h"
#include "errorHandler.h"
#include "startcode.h"

#include "bitpack.h"
#include "ac.h"
#include "ztscan.h"

#include "dwt.h"
#include "default.h"
#include "download_filter.h"

#include "main.h"
#include "ShapeCodec.h"

/* added for bilevel quantization mode */
#include "wvtPEZW.h"


#ifdef _CHECK_BITS_
FILE *fpe;
#endif

#ifdef _CHECK_HI_LO_
extern FILE *acfp;
#endif

extern int STO_const_alpha;
extern unsigned char STO_const_alpha_value;

  
/* Argument variables - set defaults */
static char *parmFileDef      = "VTC_param.ctl";
static char *statsFileDef     = "stat";
static char *decRecImgFileDef = "decRec.yuv";
static char *encRecImgFileDef = "encRec.yuv";

static char *parmFile;
static char *statsFile;
static char *decRecImgFile;
static char *encRecImgFile;
static char *recImgFile;

static char *bitFile;
static char *bitFileAC;
static int singleBitFile;


/*-----------------------------------------------------------------------*/
/********************************************************************/
/* Select a wavelet filter banks from the ones defined in default.h */
/* There are currently 10 filter banks available                    */
/********************************************************************/
static void choose_wavelet_filter(FILTER **anafilter,FILTER **synfilter,
				  int type)
{
  switch(type){
    case 0:  /* default int filter, odd symmetric */
      *anafilter=&DefaultAnalysisFilterInt;
      *synfilter=&DefaultSynthesisFilterInt;
      return;
    case 1: /* default double filter, odd symmetric */
      *anafilter=&DefaultAnalysisFilterDbl;
      *synfilter=&DefaultSynthesisFilterDbl;
      return;
    case 2:  
      *anafilter=&DefaultEvenAnalysisFilterInt;
      *synfilter=&DefaultEvenSynthesisFilterInt;
      return;
    case 3:
      *anafilter=&DefaultEvenAnalysisFilterDbl;
      *synfilter=&DefaultEvenSynthesisFilterDbl;
      return;
     case 4:
      *anafilter=&HaarAna;
      *synfilter=&HaarSyn;
      return;
    case 5:
      *anafilter=&qmf9Ana;
      *synfilter=&qmf9Syn;
      return;
    case 6:
      *anafilter=&qmf9aAna;
      *synfilter=&qmf9aSyn;
      return;
    case 7:
      *anafilter=&fpr53Ana;
      *synfilter=&fpr53Syn;
      return;
    case 8:
      *anafilter=&fpr53aAna;
      *synfilter=&fpr53aSyn;
      return;
    case 9:
      *anafilter=&asd93Ana;
      *synfilter=&asd93Syn;
      return;
    default:
      errorHandler("Filter type %d is not available.",type);
  }
}

/*-----------------------------------------------------------------------*/
static void perform_DWT(FILTER *wvtfilter)
{
  int wordsize = (mzte_codec.bit_depth>8)?2:1;
  int col,j,k,x,y;
  
  unsigned char *inimage[3];
  unsigned char *inmask[3], *outmask[3];
  DATA *outcoeff[3];
  int Nx[3], Ny[3];
  int width[3], height[3];
  int useint=1, usemask=0;
  int nLevels[3], ret;
  
  /* for 4:2:0 YUV image only */
  Nx[0] = Ny[0]=2;
  for(col=1; col<mzte_codec.colors; col++) Nx[col]=Ny[col]=1;


  nLevels[0] = mzte_codec.wvtDecompLev ;
  nLevels[1] = nLevels[2] = nLevels[0]-1;
  width[0] = mzte_codec.width;
  width[1] = width[2] = (width[0] >> 1);
  height[0] = mzte_codec.height;
  height[1] = height[2] = (height[0] >> 1);

  useint  = 1;
  usemask = mzte_codec.sa_dwt;

  
  Nx[0] = Ny[0]= 2;
  for(col=1;col<3;col++)
    Nx[col]=Ny[col]=1;
  
  for (col=0; col<mzte_codec.colors; col++) {

    inimage[col] = (unsigned char *)mzte_codec.Image[col].data;
    inmask[col] = mzte_codec.Image[col].mask;
    
    outcoeff[col] = 
      (DATA *)malloc(sizeof(DATA)*width[col]*height[col]);
    outmask[col] = 
      (unsigned char *)malloc(sizeof(char)*width[col]*height[col]);
    ret = do_DWT(inimage[col], inmask[col], width[col], height[col], 
		 nLevels[col], 0,wvtfilter, outcoeff[col], outmask[col]);
    
    if (ret!=DWT_OK) 
      errorHandler("DWT Error Code %d\n", ret);
    
    mzte_codec.mean[col] = RemoveDCMean(outcoeff[col], outmask[col], 
					width[col], height[col], nLevels[col]);
    
    for (j=0; j<width[col]*height[col]; j++)
      if (outmask[col][j] != IN)
	outcoeff[col][j]=0;
    
    for (k=0, y = 0; y < height[col]; ++y)
      for (x = 0; x < width[col]; ++x,++k) {
	COEFF_ORGVAL(x,y,col) = outcoeff[col][k];
	COEFF_MASK(x,y,col) = outmask[col][k];
      }
  }
  
  for (col=0; col<mzte_codec.colors; col++) {
    if (outmask[col]) free(outmask[col]);
    if (outcoeff[col]) free(outcoeff[col]);
  }
}

void write_image(char *recImgFile, int colors,
		 int width, int height,
		 int real_width, int real_height,
		 int rorigin_x, int rorigin_y,
		 unsigned char *outimage[3], unsigned char *outmask[3],
		 int usemask, int fullsize, int MinLevel)

     /* write outimage */
{
  FILE *outfptr, *maskfptr;
  char recSegFile[200];
  unsigned char *ptr;
  unsigned char *recmask[3];
  unsigned char *recimage[3];
  int status,i,col,ret;
  int w,h;
  int rwidth[3],rheight[3];
  int Width[3], Height[3];
  int origin_x[3],origin_y[3];
  
  Width[0] = width;
  Width[1] = Width[2] = (Width[0]+1)>>1;
  Height[0] = height;
  Height[1] = Height[2] = (Height[0]+1)>>1;
  rwidth[0] = real_width;
  rheight[0] = real_height;
  rwidth[1] = rwidth[2] = rwidth[0] >> 1; 
  rheight[1] = rheight[2] = rheight[0] >> 1; 
  origin_x[0] = rorigin_x;
  origin_x[1] = origin_x[2] = origin_x[0]>>1;
  origin_y[0] = rorigin_y;
  origin_y[1] = origin_y[2] = origin_y[0]>>1;
  outfptr = fopen(recImgFile,"wb");
  if(usemask) {
    sprintf(recSegFile, "%s.seg",recImgFile);
    maskfptr = fopen(recSegFile,"wb");
  }
  noteProgress("Writing the reconstruction image: '%s'",recImgFile);
  for (col=0; col<colors; col++) {
    
    recmask[col]  = (unsigned char *)malloc(sizeof(unsigned char)*
					    rwidth[col]*rheight[col]);
    recimage[col] = (unsigned char *)malloc(sizeof(unsigned char)*
					    rwidth[col]*rheight[col]);
    
    
    ret=PutBox(outimage[col], outmask[col], recimage[col], recmask[col], 
	       (rwidth[col]), (rheight[col]), 
	       (Width[col]), (Height[col]),
	       origin_x[col], origin_y[col], 0, 
	       (usemask)?(STO_const_alpha?STO_const_alpha_value:MASK_VAL):RECTANGULAR);
    if(ret!= DWT_OK)
      errorHandler("DWT Error code %d", ret);

    ptr = recimage[col];
    w = rwidth[col] >> (fullsize ? 0 : MinLevel);
    h = rheight[col] >> (fullsize ? 0 : MinLevel);
    for (i=0; i<h; i++) {
      if ((status = fwrite((unsigned char *)ptr, sizeof(char), 
			   w, outfptr)) != w)
	errorHandler("Error in writing image file.");
      
      ptr += w;
    } 

    if(usemask && col==0) {
      ptr = recmask[col];
      w = rwidth[col] >> (fullsize ? 0 : MinLevel);
      h = rheight[col] >> (fullsize ? 0 : MinLevel);
      for (i=0; i<h; i++) {
	if ((status = fwrite((unsigned char *)ptr, sizeof(char), 
			     w, maskfptr)) != w)
	  errorHandler("Error in writing image file.");
	ptr += w;
      } 
    }
    free(recmask[col]);
    free(recimage[col]);
  }  /* col */
  fclose(outfptr);
  if(usemask) fclose(maskfptr);
}


void ComputePSNR(unsigned char *orgY, unsigned char *recY, 
		 unsigned char *maskY,
		 unsigned char *orgU, unsigned char *recU, 
		 unsigned char *maskU,
		 unsigned char *orgV, unsigned char *recV, 
		 unsigned char *maskV,
		 int width, int height)
{
    int i;
    double quad, quad_Cr, quad_Cb, diff;
    double SNR_l, SNR_Cb, SNR_Cr;
    int  num;
    int peakvalue = 255;
    int infSNR_l, infSNR_Cb, infSNR_Cr;

    quad  = quad_Cr = quad_Cb = 0.0;
    
    /* Calculate luminance PSNR */
    num = 0;
    for (i=0; i<width*height; i++) {
      if (maskY[i] == IN) {
	diff = (int)((unsigned char *)orgY)[i] -
	  (int)((unsigned char *)recY)[i];
	quad += diff * diff;
	num ++;
      }
    }
    
    /* SNR_l = quad/(pels*lines); */
    SNR_l = quad/num;
    if (SNR_l) 
    {
      SNR_l = (peakvalue*peakvalue) / SNR_l;
      SNR_l = 10 * log10(SNR_l);
      infSNR_l = 0;
    }
    else
    {
      infSNR_l = 1;
    }

    /* Calculate chroma PSNRs */
    num = 0;
    for (i=0; i<width*height/4; i++) {
      if (maskU[i] == IN) {
	diff = (int)((unsigned char *)orgU)[i] -
	  (int)((unsigned char *)recU)[i]; 
	quad_Cb += diff * diff;
	num ++;
      }
    }
    SNR_Cb = quad_Cb/num;
    
    /* SNR_Cb = quad_Cb/(pels*lines/4.0); */
    SNR_Cb = quad_Cb/num;
    if (SNR_Cb) 
    {
      SNR_Cb = (peakvalue*peakvalue) / SNR_Cb;
      SNR_Cb = 10 * log10(SNR_Cb);
      infSNR_Cb = 0;
    }
    else 
    {
      infSNR_Cb = 1;
    }
    
    num = 0;
    for (i=0; i<width*height/4; i++) {
      if (maskV[i] == IN) {
	diff = (int)((unsigned char *)orgV)[i] -
	  (int)((unsigned char *)recV)[i];
	quad_Cr += diff * diff;
	num ++;
      }
    }
    
    SNR_Cr = quad_Cr/num;
    
    if (SNR_Cr) 
    {
      SNR_Cr = (peakvalue*peakvalue) / SNR_Cr;
      SNR_Cr = 10 * log10(SNR_Cr);
      infSNR_Cr = 0;
    }
    else
    {
      infSNR_Cr = 1;
    }
    
    
    /* Display PSNRs */
    if (infSNR_l==0)
      noteProgress("\nPSNR_Y: %.4f dB",SNR_l);
    else
      noteProgress("\nPSNR_Y: +INF dB");
    
    if (infSNR_Cb==0)
      noteProgress("PSNR_U: %.4f dB",SNR_Cb);
    else
      noteProgress("PSNR_U: +INF dB");
    
    if (infSNR_Cr==0)
      noteProgress("PSNR_V: %.4f dB\n",SNR_Cr);
    else
      noteProgress("PSNR_V: +INF dB\n");
}
  

static void perform_layer_IDWT(
        int spa_lev, int snr_lev, int usemask,
	int colors,
	int wvtDecompLev,  FILTER *wvtfilter, int fullsize)
{
  int j,k,x,y,col;
  unsigned char *outimage[3];
  unsigned char *inmask[3], *outmask[3];
  DATA *incoeff[3];
  char layer_recImgFile[200];
  int Mean[3];
  int Nx[3], Ny[3];
  int Width[3], Height[3];
  int useint=1;
  int nLevels[3], destLevel[3], ret;
 
  Width[0] = mzte_codec.width;
  Width[1] = Width[2] = (Width[0]+1)>>1;

  Height[0] = mzte_codec.height;
  Height[1] = Height[2] = (Height[0]+1)>>1;

   
  nLevels[0] = wvtDecompLev ;
  nLevels[1] = nLevels[2] = nLevels[0]-1;

  destLevel[0] = /* mzte_codec.wvtDecompLev -  */
                 mzte_codec.spatial_scalability_levels - 1 - spa_lev;
  destLevel[1] = destLevel[2] = destLevel[0];

  Mean[0] = mzte_codec.mean[0];
  Mean[1] = mzte_codec.mean[1];
  Mean[2] = mzte_codec.mean[2];

  useint  = 1;

  Nx[0] = Ny[0]= 2;
  for(col=1;col<3;col++)
    Nx[col]=Ny[col]=1;


  for (col=0; col<colors; col++) {

    if ((inmask[col]=(unsigned char *)malloc(sizeof(unsigned char)*
					     Width[col]*Height[col]))==NULL)
      errorHandler("Memory Failed\n");
    
    incoeff[col] = (DATA *)malloc(sizeof(DATA)*
				  Width[col]*Height[col]);
    
    /* copy dequantized coefficients to incoeff */
    for (k=0, y=0; y<Height[col]; y++) 
      for (x=0; x<Width[col]; x++,k++) {
	incoeff[col][k] = COEFF_RECVAL(x,y,col);
	inmask[col][k]  = COEFF_MASK(x,y,col);
      }
    for (j=0; j<Width[col]*Height[col];j++) 
      if (inmask[col][j]!=IN) 
	incoeff[col][j]=0;
    
    AddDCMean(incoeff[col], inmask[col], 
	      Width[col], Height[col], 
	      nLevels[col], Mean[col]);
    
    
    outmask[col]  = (unsigned char *)malloc(sizeof(unsigned char)*
					    Width[col]*Height[col]);
    outimage[col] = (unsigned char *)malloc(sizeof(unsigned char)*
					    Width[col]*Height[col]);

    
    ret = do_iDWT(incoeff[col], inmask[col], Width[col], Height[col], 
		  nLevels[col], destLevel[col], 0 /* byte */, 
		  wvtfilter, outimage[col], outmask[col], 0, fullsize);
    if (ret!=DWT_OK) 
      errorHandler("DWT Error Code %d\n", ret);

  }  /* col */
#ifndef _DECODER_
  sprintf(layer_recImgFile,"encRec_%02d_%02d.yuv",spa_lev,snr_lev);
#else
  sprintf(layer_recImgFile,"decRec_%02d_%02d.yuv",spa_lev,snr_lev);
#endif
  write_image(layer_recImgFile, mzte_codec.colors,
	      mzte_codec.width, mzte_codec.height,
	      mzte_codec.real_width, mzte_codec.real_height,
	      mzte_codec.origin_x, mzte_codec.origin_y,
	      outimage, outmask,
	      usemask, fullsize, destLevel[0]);

#ifndef _DECODER_
  /* compute PSNR */
  
  ComputePSNR(mzte_codec.Image[0].data, outimage[0], outmask[0],
	      mzte_codec.Image[1].data, outimage[1], outmask[1],
	      mzte_codec.Image[2].data, outimage[2], outmask[2],
	      mzte_codec.width, mzte_codec.height);
	      
#endif

  for (col=0; col<colors; col++) {
    if (incoeff[col]) free(incoeff[col]);
    if (inmask[col]) free(inmask[col]);
    if (outmask[col]) free(outmask[col]);
    if (outimage[col]) free(outimage[col]);
  }
}



static void perform_IDWT(SOL_PARAMETERS_DEC *vm_param, FILTER *wvtfilter)
{
  int j,k,x,y,col;
  unsigned char *outimage[3];
  unsigned char *inmask[3], *outmask[3];
  DATA *incoeff[3];
  int Mean[3];
  int Nx[3], Ny[3];
  int Width[3], Height[3];
  int useint=1, usemask=0;
  int nLevels[3], ret, MinLevel;
  INT fullsize = 0;
 
  fullsize = FULLSIZE;
  Width[0] = mzte_codec.width;
  Width[1] = Width[2] = (Width[0]+1)>>1;

  Height[0] = mzte_codec.height;
  Height[1] = Height[2] = (Height[0]+1)>>1;

  nLevels[0] = mzte_codec.wvtDecompLev ;
  nLevels[1] = nLevels[2] = nLevels[0]-1;

  Mean[0] = mzte_codec.mean[0];
  Mean[1] = mzte_codec.mean[1];
  Mean[2] = mzte_codec.mean[2];

  useint  = 1;
  usemask = mzte_codec.sa_dwt;

  Nx[0] = Ny[0]= 2;
  for(col=1;col<3;col++)
    Nx[col]=Ny[col]=1;


  for (col=0; col<mzte_codec.colors; col++) {

    if ((inmask[col]=(unsigned char *)malloc(sizeof(unsigned char)*
					     Width[col]*Height[col]))==NULL)
      errorHandler("Memory Failed\n");
    
    incoeff[col] = (DATA *)malloc(sizeof(DATA)*
				  Width[col]*Height[col]);
    
    /* copy dequantized coefficients to incoeff */
    for (k=0, y=0; y<Height[col]; y++) 
      for (x=0; x<Width[col]; x++,k++) {
	incoeff[col][k] = COEFF_RECVAL(x,y,col);
	inmask[col][k]  = COEFF_MASK(x,y,col);
      }
    for (j=0; j<Width[col]*Height[col];j++) 
      if (inmask[col][j]!=IN) 
	incoeff[col][j]=0;
    
    AddDCMean(incoeff[col], inmask[col], 
	      Width[col], Height[col], 
	      nLevels[col], Mean[col]);
    
    
    outmask[col]  = (unsigned char *)malloc(sizeof(unsigned char)*
					    Width[col]*Height[col]);
    outimage[col] = (unsigned char *)malloc(sizeof(unsigned char)*
					    Width[col]*Height[col]);
    
#if  PROGRESSIVE
    MinLevel = (nLevels[0] > nLevels[1])? nLevels[1]:nLevels[0];
    MinLevel = (MinLevel > nLevels[2])? nLevels[2]: MinLevel;
#else /*PROGRESSIVE*/
    MinLevel = mzte_codec.spatial_scalability_levels -
      mzte_codec.target_spatial_levels;
#endif /*PROGRESSIVE*/
    
    ret = do_iDWT(incoeff[col], inmask[col], Width[col], Height[col], 
		  nLevels[col], MinLevel, 0 /* byte */, 
		  wvtfilter, outimage[col], outmask[col], 2, fullsize);
    if (ret!=DWT_OK) 
      errorHandler("DWT Error Code %d\n", ret);
    free(incoeff[col]);
    free(inmask[col]);
  }  /* col */
  
  write_image(recImgFile, mzte_codec.colors,
	      mzte_codec.width, mzte_codec.height,
	      mzte_codec.real_width, mzte_codec.real_height,
	      mzte_codec.origin_x, mzte_codec.origin_y,
	      outimage,outmask,
	      usemask, fullsize, MinLevel);

#ifndef _DECODER_
  /* compute PSNR */
  ComputePSNR(mzte_codec.Image[0].data, outimage[0], outmask[0],
	      mzte_codec.Image[1].data, outimage[1], outmask[1],
	      mzte_codec.Image[2].data, outimage[2], outmask[2],
	      mzte_codec.width, mzte_codec.height);
#endif

  for(col=0; col< mzte_codec.colors; col++) {
    free(outmask[col]);
    free(outimage[col]);
  }
}



/*******************************************************/
/*  Flush buffer and close file. Only and must be used */
/*   by encoder to close bitstream file                */
/*******************************************************/
static void flush_buffer_file()
{
  static int totByte;
  int levByte;

  flush_bits();
  flush_bytes();

  levByte  = get_bit_rate();
  totByte += levByte;

  if (totByte==0)
    noteStat("  DC  : Bitstream Length = %d bytes, "\
	     "Total So Far = %d bytes.\n",
	     mzte_codec.curSpatialLev,mzte_codec.curSNRLev,
	     levByte,totByte);
  else
    noteStat("%2d, %2d: Bitstream Length = %d bytes, "\
	     "Total So Far = %d bytes.\n",
	     mzte_codec.curSpatialLev,mzte_codec.curSNRLev,
	     levByte,totByte);
}

static void close_buffer_file(FILE *fp)
{
  flush_buffer_file();
  fclose(fp);
}



static void zeroACBand(int c)
{
  int x, y;

  noteDetail("Zeroing out AC coefficients....");
  for (x = 0; x < mzte_codec.SPlayer[c].width; ++x)
    for (y = 0; y < mzte_codec.SPlayer[c].height; ++y)
      if (x >= mzte_codec.dcWidth || y >= mzte_codec.dcHeight)
	COEFF_VAL(x,y,c) = 0;
  noteDetail("Completed zeroing out AC coefficients.");

}


/*************************************/
/* Encode SNR layer in seperate file */
/*************************************/  
static void TextureSNRLayerMQ_encode(int spa_lev, int snr_lev, FILE *fp)
{
  SNR_IMAGE *snr_image;
  int col;
  int texture_snr_layer_id=0;

#ifdef _CHECK_HI_LO_
  acfp=fopen("encode_hl.dat","a");
#endif

  if(snrStartCode){
    noteProgress("Encoding Multi-Quant Mode Layer with SNR start code....");
    /* header */  
    emit_bits(texture_snr_layer_start_code>>16,16);
    emit_bits(texture_snr_layer_start_code,16);
    emit_bits(texture_snr_layer_id,5);
  }
  else
    noteProgress("Encoding Multi-Quant Mode Layer without SNR start code....");

  for(col=0;col<mzte_codec.colors;col++)
  {
    noteProgress("Multi-Quant Mode - Spatial %d, SNR %d, Color %d",
                 spa_lev,snr_lev,col);
    noteDetail("width=%d  height=%d",mzte_codec.SPlayer[col].width,
	      mzte_codec.SPlayer[col].height);    
    /* Set global color variable */
    mzte_codec.curColor = col;
    
    /* Set quant value and write to bitstream */
    snr_image=&(mzte_codec.SPlayer[col].SNRlayer.snr_image);
    snr_image->quant = mzte_codec.Qinfo[col][spa_lev].Quant[snr_lev];
    noteDebug("AC quant=%d", 
               mzte_codec.Qinfo[col][spa_lev].Quant[snr_lev]);
    
    /* initialization of spatial dimensions for each color component */
    if (snr_lev==0) 
      setSpatialLevelAndDimensions(spa_lev, col);
    
#ifdef _WRITE_ORV_MAP_
    dumpACORVMap(mapFileEnc, col, 0);
#endif _WRITE_ORV_MAP_

    /* get maximum residual value - this one is derived from user Q inputs not 
       actual values. Also assign skip modes. */
    updateResidMaxAndAssignSkips(col);
    
    /* quantize and mark zerotree structure for AC coefficients */
    if (encQuantizeAndMarkAC(col))
      errorHandler("encQuantizeAndMarkAC");
    noteDebug("max_root=%d max_valz=%d max_valnz=%d max_resi=%d",
               ROOT_MAX(col),VALZ_MAX(col),VALNZ_MAX(col),
               RESID_MAX(col));

    /* Write quant value and all zero flag to bitstream */
    /* emit_bits(snr_image->quant,8); */
    put_param(snr_image->quant,7);
    emit_bits(snr_image->allzero,1);

    if(snr_image->allzero==0)
      wavelet_higher_bands_encode_MQ(snr_image,col);
    
#ifdef _WRITE_COEFF_MAP_
    /* Write out coeffinfo stuff to file. */
    dumpACMap(mapFileEnc, col);
#endif _WRITE_COEFF_MAP_

    /* Update states of AC coefficients */
    if (encUpdateStateAC(mzte_codec.curColor))
      errorHandler("encUpdateStateAC");
  }
  
#ifdef _CHECK_HI_LO_
  fclose(acfp);
#endif
}



static void TextureObjectLayer_enc(FILTER *wvtfilter,SOL_PARAMETERS *vm_param)
{
  FILE *bitfile;
  int  texture_object_id=0;
  int  texture_object_layer_shape=vm_param->sa_dwt;
  int  wavelet_stuffing = 0x0f;
  int  wavelet_upload;
  int  texture_spatial_layer_id=0;
  int  spa_lev,snr_lev,col;
  char fname[100];
  int  err;
  int snr_scalability_levels;

  /*-------------------------------------------------*/
  /*--------- DC (and overall header info) ----------*/
  /*-------------------------------------------------*/

  /*------- DC: Open and initialize bitstream file -------*/
  if ((bitfile=fopen(bitFile,"wb"))==NULL)
    errorHandler("Can't open file '%s' for writing.",bitFile);

  /* initialize variables */
  init_bit_packing_fp(bitfile,1);


  /*------- DC: Write header info to bitstream file -------*/
  noteProgress("Encoding DC coefficients....");


  emit_bits(STILL_TEXTURE_OBJECT_START_CODE>>16, 16);
  emit_bits(STILL_TEXTURE_OBJECT_START_CODE, 16);
  emit_bits(texture_object_id, 16);
  emit_bits(MARKER_BIT, 1);
  emit_bits(wvtfilter->Type, 1);
  wavelet_upload=vm_param->wvtType>1?1:0;
  emit_bits(wavelet_upload, 1);
  if (wavelet_upload == 1) {
    upload_wavelet_filters(wvtfilter);
  }
  emit_bits(mzte_codec.wvtDecompLev, 8);
  emit_bits(texture_object_layer_shape, 2);
  emit_bits(wavelet_stuffing, 3);
  if (texture_object_layer_shape == 0x00) {
    emit_bits(mzte_codec.width, 15);
    emit_bits(MARKER_BIT, 1);
    emit_bits(mzte_codec.height, 15);
    emit_bits(MARKER_BIT, 1);
  }
  else { /* Arbitrary shape info, SL */
    /*horizontal_ref */
    emit_bits(mzte_codec.origin_x, 15);
    /* marker_bit */
    emit_bits(MARKER_BIT, 1);
    /*vertical_ref */
    emit_bits(mzte_codec.origin_y, 15);
    /* marker_bit */
    emit_bits(MARKER_BIT, 1);
    /* object_width */
    emit_bits(mzte_codec.width, 15);
     /* marker_bit */
    emit_bits(MARKER_BIT, 1);
    /* object_height */
    emit_bits(mzte_codec.height, 15);
    /* marker_bit */
    emit_bits(MARKER_BIT, 1);
    /* merge shape bitstream into the main bitstream */
    noteProgress("Merge Shape Bitstream ....");
    MergeShapeBitstream();
  }
  emit_bits(mzte_codec.spatial_scalability_levels, 5);
  emit_bits(mzte_codec.quantization_type, 2);
  emit_bits(mzte_codec.SNR_start_code_enable, 1);

  /*------- DC: Quantize and encode all color components -------*/
  for (col=0; col<mzte_codec.colors; col++) 
  {
    /* Set global color variable */
    mzte_codec.curColor=col;

    /* initialize DC coefficient info */
    if (err=ztqInitDC(0, col))
      errorHandler("ztqInitDC");

#ifdef _WRITE_ORV_MAP_
    dumpDCORVMap(mapFileEnc, col, 0);
#endif /*_WRITE_ORV_MAP_*/

    /* quantize DC coefficients */
    if (err=encQuantizeDC(col))
      errorHandler("encQuantizeDC");

#ifdef _WRITE_COEFF_MAP_
    /* Write out coeffinfo stuff to file. */
    dumpDCMap(mapFileEnc, col);
#endif /*_WRITE_COEFF_MAP_*/

    /* losslessly encoding DC coefficients */
    wavelet_dc_encode(col);
  }  


  /*------- DC: Close bitstream file -------*/
  if (singleBitFile) 
    flush_buffer_file(); 
  else  
    close_buffer_file(bitfile); 

  noteProgress("Completed encoding DC coefficients.");

  /*-------------------------------------------------*/
  /*--------------------- AC ------------------------*/
  /*-------------------------------------------------*/

  /*------- AC: Initialize QList Structure -------*/
  if (err=ztqQListInit())
    errorHandler("Allocating memory for QList information.");


  /*------- AC: SINGLE-QUANT MODE -------*/
  if (mzte_codec.quantization_type == SINGLE_Q)
  {
    /*------- AC: Set spatial and SNR levels to zero -------*/
    mzte_codec.curSpatialLev = 0;
    mzte_codec.curSNRLev = 0;


    /*------- AC: Open and initialize bitstream file -------*/
    if (singleBitFile==0)
    {
      sprintf(fname,bitFileAC,
	      mzte_codec.curSNRLev,mzte_codec.curSpatialLev);
      if ((bitfile=fopen(fname,"wb"))==NULL)
	errorHandler("Can't open file '%s' for writing.",fname);
    }

    /* initialize the buffer */
    init_bit_packing_fp(bitfile,1);
    

    /*------- AC: Write header info to bitstream file -------*/
    emit_bits(TEXTURE_SPATIAL_LAYER_START_CODE >> 16, 16);
    emit_bits(TEXTURE_SPATIAL_LAYER_START_CODE, 16);
    emit_bits(texture_spatial_layer_id, 5);
    

    /*------- AC: Quantize and encode all color components -------*/
    for (col=0; col<mzte_codec.colors; col++)
    {       
      noteProgress("Single-Quant Mode - Color %d",col);
      
      /* Set global color variable */
      mzte_codec.curColor = col;

      /* Set quant value and write to bitstream */
      if (texture_spatial_layer_id == 0) {
        mzte_codec.SPlayer[col].SNRlayer.snr_image.quant =
          mzte_codec.Qinfo[col][0].Quant[0];
        /* emit_bits(mzte_codec.Qinfo[col][0].Quant[0], 8); */
        put_param(mzte_codec.Qinfo[col][0].Quant[0], 7);
        noteDebug("AC quant=%d", mzte_codec.Qinfo[col][0].Quant[0]);
      }

      /* initialization of spatial dimensions for each color component */
      setSpatialLevelAndDimensions(mzte_codec.curSpatialLev, col);
      
      /* initialize AC coefficient info for each color component */
      if (err=ztqInitAC(0, col))
        errorHandler("ztqInitAC");

#ifdef _WRITE_ORV_MAP_
    dumpACORVMap(mapFileEnc, col, 0);
#endif /*_WRITE_ORV_MAP_*/

      /* quantize and mark zerotree structure for AC coefficients */
      if (err=encQuantizeAndMarkAC(col))
        errorHandler("encQuantizeAndMarkAC");
      
      /* output maximums used by different prob. models */
      noteDebug("max_root=%d, max_valz=%d, max_valnz=%d, max_resid=%d",
                 ROOT_MAX(col),VALZ_MAX(col),VALNZ_MAX(col),
                 RESID_MAX(col));

      /* losslessly encoding AC coefficients */
      wavelet_higher_bands_encode_SQ(col);
      
#ifdef _WRITE_COEFF_MAP_
      /* Write out coeffinfo stuff to file. */
      dumpACMap(mapFileEnc, col);
#endif /*_WRITE_COEFF_MAP_*/
    }


    /*------- AC: Close bitstream file -------*/
    if (singleBitFile)
      flush_buffer_file();
    else
      close_buffer_file(bitfile);

    /* store the max snr_lev and spa_lev so that the decoder can
       decode the bitstream up to the max level. */ 
    mzte_codec.target_spatial_levels = 1;
    mzte_codec.target_snr_levels = 1;
  }
 

  /*------- AC: MULTI-QUANT MODE -------*/
  else if (mzte_codec.quantization_type == MULTIPLE_Q)
  {
    /* Initialize AC coefficients for all colors based on spatial level 0 */
    for (col=0; col<mzte_codec.colors; col++) 
    {
      setSpatialLevelAndDimensions(0, col);
      if (err=ztqInitAC(0, col))
        errorHandler("ztqInitAC");
    }

    /* Loop through spatial layers */
    for (spa_lev=0; spa_lev<mzte_codec.spatial_scalability_levels; 
         spa_lev++)
    {
      /*----- AC: Set global spatial layer and SNR scalability level. -----*/
      /* Assumes all three color components have the same number of SNR 
         levels */
      mzte_codec.curSpatialLev = spa_lev;
      mzte_codec.SPlayer[0].SNR_scalability_levels = 
        mzte_codec.Qinfo[0][spa_lev].SNR_scalability_levels;
      snr_scalability_levels = mzte_codec.SPlayer[0].SNR_scalability_levels;

      /* Update spatial level coeff info if changing spatial levels.
         Do this for all color components */
      if (spa_lev != 0)
      {
        for (mzte_codec.curColor = 0; mzte_codec.curColor<mzte_codec.colors;
             mzte_codec.curColor++) 
        {
          setSpatialLevelAndDimensions(mzte_codec.curSpatialLev, 
                                       mzte_codec.curColor);
          spatialLayerChangeUpdate(mzte_codec.curColor);
        }
      }

      if (!snrStartCode)
      {
        /*------- AC: Open and initialize bitstream file -------*/
	if (singleBitFile==0)
	{
	  sprintf(fname,bitFileAC, mzte_codec.curSpatialLev);
	  if ((bitfile=fopen(fname,"wb"))==NULL)
	    errorHandler("Can't open file '%s' for writing.",fname);
	}

	/* initialize the buffer */
        init_bit_packing_fp(bitfile,1);


        /*------- AC: Write header info to bitstream file -------*/
        emit_bits(TEXTURE_SPATIAL_LAYER_START_CODE >> 16, 16);
        emit_bits(TEXTURE_SPATIAL_LAYER_START_CODE, 16);
        emit_bits(texture_spatial_layer_id, 5);
        emit_bits(snr_scalability_levels, 5);
      }


      /* Loop through SNR layers */      
      for (snr_lev=0; snr_lev<snr_scalability_levels; snr_lev++) 
      {
        /*----- AC: Set global SNR layer -----*/
        mzte_codec.curSNRLev = snr_lev;
        
        if (snrStartCode)
        {
          /*------- AC: Open and initialize bitstream file -------*/
	  if (singleBitFile==0)
	  {
	    sprintf(fname,bitFileAC,
		    mzte_codec.curSpatialLev, mzte_codec.curSNRLev);
	    if ((bitfile=fopen(fname,"wb"))==NULL)
	      errorHandler("Can't open file '%s' for writing.",fname);
	  }

          /* initialize the buffer */
          init_bit_packing_fp(bitfile,1);

          if (snr_lev==0) {
            /*------- AC: Write header info to bitstream file -------*/
            emit_bits(TEXTURE_SPATIAL_LAYER_START_CODE >> 16, 16);
            emit_bits(TEXTURE_SPATIAL_LAYER_START_CODE, 16);
            emit_bits(texture_spatial_layer_id, 5);
            emit_bits(snr_scalability_levels, 5);
          }

        }

        /*------- AC: Quantize and encode all color components -------*/
        TextureSNRLayerMQ_encode(spa_lev, snr_lev, bitfile);
        if (snrStartCode)
	{
	  if (singleBitFile)
	    flush_buffer_file();
	  else
	    close_buffer_file(bitfile);
	}


      } /* snr_lev */
      
      if (!snrStartCode)
      {
	if (singleBitFile)
	  flush_buffer_file();
	else
	  close_buffer_file(bitfile);
      }

    }  /* spa_lev */

    /* store the max snr_lev and spa_lev so that the decoder can
       decode the bitstream up to the max level. */ 
    mzte_codec.target_spatial_levels = spa_lev;
    mzte_codec.target_snr_levels = snr_lev;

  } /* multi-quant mode */

  /*------- AC: BILEVEL-QUANT MODE -------*/
  else if (mzte_codec.quantization_type == BILEVEL_Q)
    {
      PEZW_SPATIAL_LAYER *SPlayer[NCOLOR];
      int h,w;
      int Quant[NCOLOR];
      int levels;

      /*------- AC: Open and initialize bitstream file -------*/
      if (singleBitFile==0)
        {
          sprintf(fname,bitFileAC,
                  mzte_codec.curSNRLev,mzte_codec.curSpatialLev);
          if ((bitfile=fopen(fname,"wb"))==NULL)
            errorHandler("Can't open file '%s' for writing.",fname);
        }
      
      /* initialize the buffer */
      init_bit_packing_fp(bitfile,1);
    
      /* for now, start_code is always enabled */
      mzte_codec.SNR_start_code_enable = 1;

      /*------- AC: encode all color components -------*/
      for (col=0; col<mzte_codec.colors; col++)
        { 
          noteProgress("Bilevel-Quant Mode - Color %d",col);

          if (col ==0)  /* Lum */
            {
              h = mzte_codec.height;
              w = mzte_codec.width;
              levels = mzte_codec.wvtDecompLev;
            }
          else { /* should depend on mzte_codec.color_format;
                    not implemented yet, only support 4:2:0  */
            h = mzte_codec.height/2;
            w = mzte_codec.width/2;
            levels = mzte_codec.wvtDecompLev-1;
          }

          /* initialize data */
          SPlayer[col] = Init_PEZWdata (col,levels,w,h);

          /* get quantization information */
          Quant[col] = mzte_codec.Qinfo[col]->Quant[0];

          /* encode this color componenet */
          WvtPEZW_Spatial_SNR (levels, SPlayer[col], Quant[col],1);
            
        }        

      /* package the bitstream and write to bitfile */
      PEZW_bitpack (SPlayer);

      if (singleBitFile)
        flush_buffer_file();
      else
        close_buffer_file(bitfile);
      
    } /* bilevel-quant mode */


  if (singleBitFile)
    fclose(bitfile);

  /*------- AC: Free Qlist structure -------*/
  ztqQListExit();
}


/*************************************/
/* Decode SNR layer in separete file */
/*************************************/  
static void TextureSNRLayerMQ_decode(int spa_lev, int snr_lev,FILE *fp)
{
  SNR_IMAGE *snr_image;
  int col;
  int texture_snr_layer_id;

#ifdef _CHECK_HI_LO_
  acfp=fopen("decode_hl.dat","a");
#endif

  if(snrStartCode){
    noteProgress("Decoding Multi-Quant Mode Layer with SNR start code....");
    /* header info */
    if(get_X_bits(32) != texture_snr_layer_start_code)
      errorHandler("Error in decoding texture_snr_layer_start_code");
    texture_snr_layer_id=get_X_bits(5);
  }
  else 
    noteProgress("Decoding Multi-Quant Mode Layer without SNR start code....");


  for(col=0;col<mzte_codec.colors;col++){  
    
    noteProgress("Multi-Quant Mode - Spatial %d, SNR %d, Color %d",
                 spa_lev,snr_lev,col);
    
    /* Set global color variable */
    mzte_codec.curColor = col;
    
    /* initialization of spatial dimensions for each color component  */
    setSpatialLevelAndDimensions(mzte_codec.curSpatialLev, col);
    
    snr_image=&(mzte_codec.SPlayer[col].SNRlayer.snr_image);
    snr_image->quant=get_param(7);
    /* snr_image->quant=get_X_bits(8); */
    noteDebug("AC quant=%d", snr_image->quant);
    
    updateResidMaxAndAssignSkips(col);
    noteDebug("resid_max=%d\n",snr_image->residual_max);

    if((snr_image->allzero=get_X_bits(1))==0)
      wavelet_higher_bands_decode_MQ(snr_image,col);
    else
    {
      noteProgress("Note: Current AC band has all zero condition.");
      /* put zeros in all AC coefficients */
      zeroACBand(col);
      decMarkAC(col);
    }
    
#ifdef _WRITE_COEFF_MAP_
    /* Write out coeffinfo stuff to file. */
    dumpACMap(mapFileDec, col);
#endif _WRITE_COEFF_MAP_
    
    /* quantize and mark zerotree structure for AC coefficients */
    if (decIQuantizeAC(col))
      errorHandler("decQuantizeAndMarkAC");
    noteDebug("max_root=%d max_valz=%d max_valnz=%d max_resi=%d",
               ROOT_MAX(col),VALZ_MAX(col),VALNZ_MAX(col),
               RESID_MAX(col));
 
    /* Update states of ac coefficients */
    if (decUpdateStateAC(col))
      errorHandler("decUpdateStateAC");

#ifdef _WRITE_ORV_MAP_
    dumpACORVMap(mapFileDec, col, 0);
#endif _WRITE_ORV_MAP_
  } /* col */

#ifdef _CHECK_HI_LO_
  fclose(acfp);
#endif
}



static void TextureObjectLayer_dec(SOL_PARAMETERS_DEC *vm_param_dec,
				   int  target_spatial_levels,
				   int  target_snr_levels, FILTER *wvtfilter)
{
  FILE *bitfile;
  int  texture_spatial_layer_start_code;
  int  texture_object_id;
  int  marker_bit;
  int  wavelet_download;
  int  texture_object_layer_shape;
  int  wavelet_stuffing;
  int  texture_spatial_layer_id;
  int  spa_lev,snr_lev,col;
  char fname[100];
  int  err;
  int  snr_scalability_levels;
  int   still_texture_object_start_code;
  int  x,y,w,h,k;
  int  decoder_done=0;

  unsigned char *inmask[3], *outmask[3];
  int Width[3], Height[3];
  int nLevels[3], ret;
  int Nx[3], Ny[3];
  int usemask;
  int useint=1;
  PICTURE *Image;

  PEZW_SPATIAL_LAYER **SPlayer;
  WINT *output;
  int hsize, vsize;
  int Resol;
  int bitplane;
  int i,j;
  int adj;
  int all_zero;
  int *lens;
  unsigned char *data;
  int n;


  /*-------------------------------------------------*/
  /*--------- DC (and overall header info) ----------*/
  /*-------------------------------------------------*/

  /*------- DC: Open and initialize bitstream file -------*/
  if ((bitfile=fopen(bitFile,"rb"))==NULL)
    errorHandler("Can't open file '%s' for reading.",bitFile);

  /* initialize variables */
  init_bit_packing_fp(bitfile,1);


  /*------- DC: Read header info from bitstream file -------*/
  still_texture_object_start_code = get_X_bits(32);
  if (still_texture_object_start_code != STILL_TEXTURE_OBJECT_START_CODE)
    errorHandler("Wrong texture_object_layer_start_code.");

  texture_object_id = get_X_bits(16);
  marker_bit = get_X_bits(1);
  wvtfilter->Type = get_X_bits(1);
  wavelet_download = get_X_bits(1);
  if (wavelet_download == 1) {
    download_wavelet_filters(wvtfilter);
  }
  else if(wvtfilter->Type==INT_TYPE){
    mzte_codec.wvtType = 0;
    *wvtfilter=DefaultSynthesisFilterInt;
  }
  else{
    mzte_codec.wvtType = 1;
    *wvtfilter=DefaultSynthesisFilterDbl;
  }

  mzte_codec.wvtDecompLev = get_X_bits(8);
 
  texture_object_layer_shape = get_X_bits(2);
  wavelet_stuffing = get_X_bits(3);
  if (texture_object_layer_shape==0) {
    mzte_codec.sa_dwt       = 0; 
    mzte_codec.width = get_X_bits(15);
    marker_bit = get_X_bits(1);
    mzte_codec.height = get_X_bits(15);
    marker_bit = get_X_bits(1);
  }
  else { /* Arbitrary shape header */
    mzte_codec.sa_dwt       = 1; 
    mzte_codec.origin_x = get_X_bits(15);
    marker_bit = get_X_bits(1);
    mzte_codec.origin_y = get_X_bits(15);
    marker_bit = get_X_bits(1);
    mzte_codec.width = get_X_bits(15);
    marker_bit = get_X_bits(1);
    mzte_codec.height = get_X_bits(15);
    marker_bit = get_X_bits(1);
    mzte_codec.real_width = vm_param_dec->width;
    mzte_codec.real_height = vm_param_dec->height;
  }

  /* decode the shape info from bitstream */
  if(mzte_codec.sa_dwt)
    noteProgress("Decoding Shape Information...");
  Image = (PICTURE *)malloc(sizeof(PICTURE)*3);
  
  get_virtual_mask(Image, mzte_codec.wvtDecompLev,
		   mzte_codec.width, mzte_codec.height, 
		   mzte_codec.sa_dwt, vm_param_dec->colors);
    
  if (mzte_codec.sa_dwt)
    mzte_codec.SegImage  = vm_param_dec->SegImage;
  noteDetail("Creating and initializing data structures....");
  mzte_codec.colors=3;  /* hard code */
  for (col=0; col<mzte_codec.colors; col++) {

    h = mzte_codec.height >> (col>0);
    w  = mzte_codec.width  >> (col>0);

   /*--------- create coeffinfo structure ----------*/
    if ((mzte_codec.SPlayer[col].coeffinfo = 
	 (COEFFINFO **)calloc(h, sizeof(COEFFINFO *))) == NULL)
       errorHandler("Allocating memory for coefficient structure (I).");

    if ((mzte_codec.SPlayer[col].coeffinfo[0] = 
	 (COEFFINFO *)calloc(h*w, sizeof(COEFFINFO))) == NULL)
      errorHandler("Allocating memory for coefficient structure (II).");
   
    for (y = 1; y < h; ++y)
      mzte_codec.SPlayer[col].coeffinfo[y] = 
	mzte_codec.SPlayer[col].coeffinfo[y-1]+w;
  }
  noteDetail("Completed creating and initializing data structures.");

  mzte_codec.dcHeight  = mzte_codec.height >> mzte_codec.wvtDecompLev;
  mzte_codec.dcWidth   = mzte_codec.width >> mzte_codec.wvtDecompLev;

  mzte_codec.spatial_scalability_levels = get_X_bits(5);
  mzte_codec.quantization_type = get_X_bits(2);  
  mzte_codec.SNR_start_code_enable = get_X_bits(1); 
  mzte_codec.colors = 3;
  mzte_codec.bit_depth = 8;  

  usemask = mzte_codec.sa_dwt;
  useint =1;
  if (mzte_codec.sa_dwt==0) {
    for (col=0; col<mzte_codec.colors; col++) {
      int x,y;
      int width = mzte_codec.width >> (col>0);
      int height = mzte_codec.height >> (col>0);
      for (y=0; y<height; y++)
	for (x=0; x<width; x++) 
	  COEFF_MASK(x,y,col) = IN;
    }
  }
  else {
    /* copy over the inmask[0] */

    Width[0] = mzte_codec.width;
    Width[1] = Width[2] = (Width[0] >> 1);
    Height[0] = mzte_codec.height;
    Height[1] = Height[2] = (Height[0] >> 1);
    nLevels[0] = mzte_codec.wvtDecompLev ;
    nLevels[1] = nLevels[2] = nLevels[0]-1;

    Nx[0] = Ny[0]=2;
    for(col=1;col<3;col++) Nx[col]=Ny[col]=1;

#ifdef _DECODER_
    mzte_codec.Image = Image;
#endif
    for (col=0; col<mzte_codec.colors; col++) {
      inmask[col] = mzte_codec.Image[col].mask; 
      outmask[col] = (unsigned char *)malloc(sizeof(unsigned char)*
					     Width[col]*Height[col]);
      ret = do_DWTMask(inmask[col], outmask[col], 
		       Width[col], Height[col], 
		       nLevels[col], wvtfilter); 
      if (ret!= DWT_OK) 
	errorHandler("DWT Error Code %d\n", ret);
      
      for (k=0,y=0; y<Height[col]; y++)
	for (x=0; x<Width[col]; x++) 
	  COEFF_MASK(x,y,col) = outmask[col][k++];
      free(outmask[col]);

    }
  }
  
  if (target_spatial_levels<=0 || target_snr_levels<= 0)
    errorHandler("Neither target_spatial_levels nor target_snr_levels can be zero");

  if (target_spatial_levels > mzte_codec.spatial_scalability_levels) {
     target_spatial_levels = mzte_codec.spatial_scalability_levels;
     mzte_codec.target_spatial_levels = target_spatial_levels;
  }


  /*------- DC: Decode and inverse quantize all color components -------*/
  noteProgress("Decoding DC coefficients....");
  for (col=0; col<mzte_codec.colors; col++) 
  {
    /* initilize all wavelet coefficients */
    mzte_codec.curColor=col;
    err=ztqInitDC(1, col);

    /* losslessly decoding DC coefficients */
    wavelet_dc_decode(col);

#ifdef _WRITE_COEFF_MAP_
    /* Write out coeffinfo stuff to file. */
    dumpDCMap(mapFileDec, col);
#endif _WRITE_COEFF_MAP_

    /* dequantize DC coefficients */
    err=decIQuantizeDC(col);

#ifdef _WRITE_ORV_MAP_
    dumpDCORVMap(mapFileDec, col, 1);
#endif _WRITE_ORV_MAP_
  }  


  /*------- DC: Close bitstream file -------*/
  if (singleBitFile==0)
    fclose(bitfile);

  noteProgress("Completed decoding of DC coefficients.");

  /*-------------------------------------------------*/
  /*--------------------- AC ------------------------*/
  /*-------------------------------------------------*/

  /*------- AC: Initialize QList Structure -------*/
  if (err=ztqQListInit())
    errorHandler("Allocating memory for QList information.");
  
  
  /*------- AC: SINGLE-QUANT MODE -------*/
  if (mzte_codec.quantization_type == SINGLE_Q) 
  {
    /*------- AC: Set spatial and SNR levels to zero -------*/
    mzte_codec.curSpatialLev = 0;
    mzte_codec.curSNRLev = 0;


    /*------- AC: Open and initialize bitstream file -------*/
    if (singleBitFile==0)
    {
      sprintf(fname,bitFileAC,
	      mzte_codec.curSNRLev,mzte_codec.curSpatialLev);
      if ((bitfile=fopen(fname,"rb"))==NULL)
	errorHandler("Can't open file '%s' for reading.",fname);
      init_bit_packing_fp(bitfile,1);
    }
    else
      init_bit_packing_fp(bitfile,0);
 
   
    /*------- AC: Read header info from bitstream file -------*/
    texture_spatial_layer_start_code = get_X_bits(32);
    if (texture_spatial_layer_start_code != TEXTURE_SPATIAL_LAYER_START_CODE)
      errorHandler("Wrong texture_spatial_layer_start_code.");
    
    texture_spatial_layer_id = get_X_bits(5);
    

    /*------- AC: Decode and inverse quantize all color components -------*/
    for (col=0; col<mzte_codec.colors; col++) 
    {  
      noteProgress("Single-Quant Mode - Color %d",col);
      
      /* Set global color variable */
      mzte_codec.curColor = col;
       
      /* Read quant value from bitstream */
      /* mzte_codec.SPlayer[col].SNRlayer.snr_image.quant = 
        get_X_bits(8); */
      mzte_codec.SPlayer[col].SNRlayer.snr_image.quant = 
        get_param(7);
      noteDebug("Quant=%d", mzte_codec.SPlayer[col].SNRlayer.snr_image.quant);

      /* initialization of spatial dimensions for each color component */
      setSpatialLevelAndDimensions(mzte_codec.curSpatialLev, col);

      /* initialize AC coefficient info */
      if (err=ztqInitAC(1, col))
        errorHandler("ztqInitAC");
 
      /* losslessly decoding AC coefficients */
      wavelet_higher_bands_decode_SQ( col);

#ifdef _WRITE_COEFF_MAP_
      /* Write out coeffinfo stuff to file. */
      dumpACMap(mapFileDec, col);
#endif _WRITE_COEFF_MAP_
      
      /* Inverse quantize AC coefficients */
      if (err=decIQuantizeAC(col))
        errorHandler("decIQuantizeAC");
        
      noteDebug("max_root=%d max_valz=%d max_valnz=%d max_resi=%d",
                 ROOT_MAX(col),VALZ_MAX(col),VALNZ_MAX(col),RESID_MAX(col));


#ifdef _WRITE_ORV_MAP_
    dumpACORVMap(mapFileDec, col, 1);
#endif _WRITE_ORV_MAP_
    }

    /*------- AC: Close bitstream file -------*/
    if (singleBitFile==0)
      fclose(bitfile);
  }


  /*------- AC: MULTI-QUANT MODE -------*/
  else if (mzte_codec.quantization_type == MULTIPLE_Q)
  {
    /* Initialize AC coefficients for all colors based on spatial level 0 */
    for (col=0; col<mzte_codec.colors; col++) 
    {
      setSpatialLevelAndDimensions(0, col);
      if (err=ztqInitAC(1, col))
        errorHandler("ztqInitAC");
    }

    /* Loop through spatial layers */

    spa_lev = 0;
    do {

      /*----- AC: Set global spatial layer and SNR scalability level. -----*/
      /* Assumes all three color components have the same number of SNR 
         levels */
      mzte_codec.curSpatialLev = spa_lev;

      /* Update spatial level coeff info if changing spatial levels.
         Do this for all color components */
      if (spa_lev != 0)
      {
        for (mzte_codec.curColor = 0; mzte_codec.curColor<mzte_codec.colors;
             mzte_codec.curColor++) 
        {
          setSpatialLevelAndDimensions(mzte_codec.curSpatialLev, 
                                       mzte_codec.curColor);
          spatialLayerChangeUpdate(mzte_codec.curColor);
        }
      }

      if (!snrStartCode)
      {
	if (singleBitFile==0)
	{
	  /*------- AC: Open and initialize bitstream file -------*/
	  sprintf(fname,bitFileAC, mzte_codec.curSpatialLev);
	  if ((bitfile=fopen(fname,"rb"))==NULL)
	    errorHandler("Can't open file '%s' for reading.",fname);
 
	  /* initialize the buffer */
	  init_bit_packing_fp(bitfile,1);
	}
	else
	  init_bit_packing_fp(bitfile,0);

        /*------- AC: Read header info from bitstream file -------*/
        texture_spatial_layer_start_code = get_X_bits(32);
        if (texture_spatial_layer_start_code != 
	    TEXTURE_SPATIAL_LAYER_START_CODE)
          errorHandler("Wrong texture_spatial_layer_start_code.");
        texture_spatial_layer_id = get_X_bits(5);
        snr_scalability_levels = get_X_bits(5);  /* hjlee */
        mzte_codec.SPlayer[0].SNR_scalability_levels = snr_scalability_levels;
 
      }

     /* Loop through SNR layers */  
      snr_lev=0;
      do {

        /*----- AC: Set global SNR layer -----*/
        mzte_codec.curSNRLev = snr_lev;

        if (snrStartCode)
        {
          /*------- AC: Open and initialize bitstream file -------*/
	  if (singleBitFile==0)
	  {
	    sprintf(fname,bitFileAC,
		    mzte_codec.curSpatialLev,mzte_codec.curSNRLev);
	    if ((bitfile=fopen(fname,"rb"))==NULL)
	      errorHandler("Can't open file '%s' for reading.",fname);
	
	    /* initialize the buffer */
	    init_bit_packing_fp(bitfile,1);
	  }
	  else
	    init_bit_packing_fp(bitfile,0);

          /*------- AC: Read header info from bitstream file -------*/
          if (snr_lev==0) {
            texture_spatial_layer_start_code = get_X_bits(32);
            if (texture_spatial_layer_start_code != 
                TEXTURE_SPATIAL_LAYER_START_CODE)
              errorHandler("Wrong texture_spatial_layer_start_code.");
         
            texture_spatial_layer_id = get_X_bits(5);
            snr_scalability_levels = get_X_bits(5);  /* hjlee */
            mzte_codec.SPlayer[0].SNR_scalability_levels = 
              snr_scalability_levels;
          }

        }

        /*------- AC: Decode and inverse quantize all color components ------*/
        TextureSNRLayerMQ_decode(spa_lev, snr_lev, bitfile);
        if (snrStartCode)
	  if (singleBitFile==0)
	    fclose(bitfile);

	if (spa_lev>=target_spatial_levels-1 && 
	    snr_lev>=target_snr_levels-1)
	   decoder_done = 1;

	perform_layer_IDWT(spa_lev, snr_lev, mzte_codec.sa_dwt,
			   mzte_codec.colors, mzte_codec.wvtDecompLev,  
			   wvtfilter, FULLSIZE);



        snr_lev++;
      } while (snr_lev<snr_scalability_levels && (!decoder_done)); 

      if (!snrStartCode)
	if (singleBitFile==0)
	  fclose(bitfile);

      spa_lev++;
    } while (spa_lev<target_spatial_levels && (!decoder_done));  /* spa_lev */

  } /* multi-quant mode */

  /*------- AC: BILEVEL-QUANT MODE -------*/
  else if (mzte_codec.quantization_type == BILEVEL_Q)
    {

      SPlayer = (PEZW_SPATIAL_LAYER **)calloc(mzte_codec.colors, sizeof(void *))
;

      /*------- AC: Open and initialize bitstream file -------*/
      if (singleBitFile==0)
        {
          sprintf(fname,bitFileAC,
                  mzte_codec.curSNRLev,mzte_codec.curSpatialLev);
          if ((bitfile=fopen(fname,"r"))==NULL)
            errorHandler("Can't open file '%s' for reading.",fname);
          init_bit_packing_fp(bitfile,1);
        }
      else
        init_bit_packing_fp(bitfile,0);
      
      
      /* parse the bitstreams into spatial and snr layers */
      PEZW_bit_unpack (SPlayer);

      /* decode the bitstream */
      snr_scalability_levels = SPlayer[0][0].SNR_scalability_levels;
      lens = calloc(snr_scalability_levels*mzte_codec.wvtDecompLev,
                    sizeof(int));
      

      for(i=0;i<mzte_codec.wvtDecompLev;i++)
        for (j=0;j<snr_scalability_levels;j++)
          lens[i*snr_scalability_levels+j] = 
            SPlayer[0][i].SNRlayer[j].snr_bitstream.length;

      for (col=0;col<mzte_codec.colors;col++)
        {
          /* adjustment for decompositionlevels for different color */
          if(col==0) 
            adj=0;
          else 
            adj=1;

          /* target bitplane */
          bitplane = mzte_codec.target_snr_levels;
          if (bitplane>SPlayer[col][0].SNR_scalability_levels)
            bitplane = SPlayer[col][0].SNR_scalability_levels;

          /* target resolution */
          Resol = mzte_codec.wvtDecompLev-mzte_codec.target_spatial_levels;
          if (Resol<0) Resol=0;
          Resol=0;
          
          if(DEBUG_DEC_BS)
            fprintf(stdout, "\nResolution Off: %d \n", Resol);

          SPlayer[col][0].SNRlayer[0].Quant *= 1<<(SPlayer[col][0].SNR_scalability_levels-1);
          output = SPlayer[col][0].SNRlayer[0].snr_image.data;
          hsize = SPlayer[col][0].SNRlayer[0].snr_image.width;
          vsize = SPlayer[col][0].SNRlayer[0].snr_image.height;

          WvtDecodePEZW_Spatial_SNR (output,hsize,vsize,SPlayer[col], 
            mzte_codec.wvtDecompLev-adj,Resol,bitplane,mzte_codec.SNR_start_code_enable);

          /* parse the bitstream for the next color component */
          if (col<mzte_codec.colors-1){
            for(i=0;i<mzte_codec.wvtDecompLev-adj;i++)
              for (j=0;j<snr_scalability_levels;j++){

                lens[i*snr_scalability_levels+j] -= 
                  SPlayer[col][i].SNRlayer[j].snr_bitstream.length;

                (char *) SPlayer[col+1][i].SNRlayer[j].snr_bitstream.data =
                       (char *) SPlayer[col][i].SNRlayer[j].snr_bitstream.data +
 
                       SPlayer[col][i].SNRlayer[j].snr_bitstream.length;

                /* shift the data to the left by one bit */
                data = SPlayer[col+1][i].SNRlayer[j].snr_bitstream.data;
                all_zero = lshift_by_NBit(data,lens[i*snr_scalability_levels+j],
1);

                if(all_zero)
                  SPlayer[col+1][i].SNR_scalability_levels--;

                if(DEBUG_BILEVEL){
                  fprintf(stdout,"\ncurrent componenet lenght: %d",
                          SPlayer[col][i].SNRlayer[j].snr_bitstream.length);
                  fprintf(stdout,"\ncolor %d spalev %d snrlev %d\n",col+1,i,j);
                  for(n=0;n<lens[i*snr_scalability_levels+j];n++)
                         fprintf(stdout,"%d ", data[n]);
                }
              } /* end of snr levels */
          }
        } /* end of color loop */

      /* put the decoded wavelet coefficients back into the data structure */
      restore_PEZWdata (SPlayer);

    }  /* bilevel quantization */

  for(col=0; col< mzte_codec.colors; col++)
    free(Image[col].mask);
  free(Image);
  ztqQListExit();
} 



/***********************************************************************/
/***********************************************************************/
/*          ENCODING PARAMATERS                                        */
/***********************************************************************/
/***********************************************************************/

void VTC_Encode()
{
  int y;
  int col;
  int width,height;
  FILTER *anafilter,*synfilter;
  SOL_PARAMETERS *vm_param;

  noteProgress("\n----- MPEG-4 Visual Texture Coding: Encoding -----\n");

  /* read the control file */
  noteProgress("Reading input parameters from '%s'....", parmFile);
  if ((vm_param = (SOL_PARAMETERS *)malloc(sizeof(SOL_PARAMETERS)*1)) == NULL)
    errorHandler("Allocating memory for parameters.");

  read_control_file(parmFile, vm_param);
  noteProgress("Completed reading input parameters.");
  
  /* read the source image */
  noteProgress("Reading in image '%s'....", vm_param->ImagePath);
  read_image(vm_param->ImagePath, vm_param->Image[0].width, 
             vm_param->Image[0].height, vm_param->colors, 
             vm_param->bit_depth, vm_param->Image); 
  noteProgress("Completed reading in image.");

  /* read the segmentation map of the source image */
  noteProgress("Reading in seg map '%s'....", vm_param->SegPath);
  vm_param->sa_dwt = read_segimage(vm_param->SegPath, 
	               vm_param->Image[0].width, 
                       vm_param->Image[0].height, vm_param->colors, 
                       vm_param->bit_depth, vm_param->Image); 
  noteProgress("Completed reading in seg map.");


  get_virtual_image(vm_param->Image, vm_param->wvtDecompLev, 
		    vm_param->sa_dwt, vm_param->colors, 
		    vm_param->alphaTH, vm_param->change_CR_disable);

  /* Open statistics file. All subsequent calls to noteStat will write
     to this file. */
  if ((sfp=fopen(statsFile,"w"))==NULL)
     errorHandler("Couldn't open stat file '%s' for write.", statsFile);
 
  /* fill in encoding/decoding parameters */
  mzte_codec.Image                      = vm_param->Image;
  mzte_codec.wvtDecompLev               = vm_param->wvtDecompLev; 
  mzte_codec.wvtType                    = vm_param->wvtType; 
  mzte_codec.sa_dwt                     = vm_param->sa_dwt; 
  if (mzte_codec.sa_dwt)
     mzte_codec.SegImage                = vm_param->SegImage;
  mzte_codec.spatial_scalability_levels = vm_param->spatial_scalability_levels;
  /* mzte_codec.height                     = vm_param->Image[0].height;
  mzte_codec.width                      = vm_param->Image[0].width; */
  mzte_codec.spatial_scalability_levels = vm_param->spatial_scalability_levels;
  mzte_codec.quantization_type          = vm_param->quantization_type;
  mzte_codec.SNR_start_code_enable      = vm_param->SNR_start_code_enable;
  mzte_codec.colors                     = vm_param->colors;
  mzte_codec.bit_depth                  = vm_param->bit_depth;
  mzte_codec.deringWindowSize           = vm_param->deringWindowSize;
  mzte_codec.deringThreshold            = vm_param->deringThreshold;
  mzte_codec.target_spatial_levels      = vm_param->target_spatial_levels; 
  mzte_codec.target_snr_levels          = vm_param->target_snr_levels; 
  mzte_codec.Qdc[0]                     = vm_param->QdcY;
  if (vm_param->colors != MONO) {
    mzte_codec.Qdc[1]                   = vm_param->QdcUV;
    mzte_codec.Qdc[2]                   = vm_param->QdcUV;
  }
  for (col=0; col<vm_param->colors; col++)
    mzte_codec.Qinfo[col]      = vm_param->SPlayer[col];
  mzte_codec.dcHeight          = mzte_codec.height >> mzte_codec.wvtDecompLev;
  mzte_codec.dcWidth           = mzte_codec.width >> mzte_codec.wvtDecompLev;
  for (col=0; col<mzte_codec.colors; col++) {
    mzte_codec.Image[col].height = mzte_codec.height >> (col>0);
    mzte_codec.Image[col].width  = mzte_codec.width >> (col>0);
  }

  singleBitFile = vm_param->singleBitFile;
  bitFile = vm_param->bitFile;
  bitFileAC = vm_param->bitFileAC;

  noteDetail("Completed creating and initializing data structures.");

  for (col=0; col<mzte_codec.colors; col++) {
    height = mzte_codec.Image[col].height; 
    width  = mzte_codec.Image[col].width;
    
    /*--------- create coeffinfo structure ----------*/
    if ((mzte_codec.SPlayer[col].coeffinfo = 
	 (COEFFINFO **)calloc(height, sizeof(COEFFINFO *))) == NULL)
      errorHandler("Allocating memory for coefficient structure (I).");

    if ((mzte_codec.SPlayer[col].coeffinfo[0] = 
	 (COEFFINFO *)calloc(height*width, sizeof(COEFFINFO))) == NULL)
      errorHandler("Allocating memory for coefficient structure (II).");
    
    for (y = 1; y < height; ++y)
      mzte_codec.SPlayer[col].coeffinfo[y] = 
	mzte_codec.SPlayer[col].coeffinfo[y-1]+width;
  }

  /***********************************************************************/
  /***********************************************************************/
  /*            DISCRETE WAVELET TRANSFORM                               */
  /***********************************************************************/
  /***********************************************************************/
  noteProgress("Wavelet Transform....");  
  choose_wavelet_filter(&anafilter,&synfilter,vm_param->wvtType);
  perform_DWT(anafilter);
  noteProgress("Completed wavelet transform.");
  

#ifdef _CHECK_BITS_
  fpe=fopen("encode_bits","w");
#endif

  TextureObjectLayer_enc(synfilter,vm_param);  

#ifdef _CHECK_BITS_
  fclose(fpe);
#endif

  /*  close statistics file */
  fclose(sfp);

  /*----- free up coeff data structure -----*/
  noteDetail("Freeing up encoding data structures....");
  for (col=0; col<mzte_codec.colors; col++) {
    free(mzte_codec.SPlayer[col].coeffinfo[0]);
    free(mzte_codec.SPlayer[col].coeffinfo);
  }
  noteDetail("Completed freeing up encoding data structures.");

  noteProgress("\n----- Encoding Completed. -----\n");
  
  free_vm_param(vm_param);
}


/***********************************************************************/
/***********************************************************************/
/*                            DECODING                                 */
/***********************************************************************/
/***********************************************************************/

void VTC_Decode()
{
  int col;
  SOL_PARAMETERS_DEC *vm_param_dec;
  FILTER wvtfilter;

  noteProgress("\n----- MPEG-4 Visual Texture Coding: Decoding -----\n");

  /* Read in decoding parameters */
  noteProgress("Reading input parameters from '%s'....", parmFile);
  if ((vm_param_dec = 
       (SOL_PARAMETERS_DEC *)malloc(sizeof(SOL_PARAMETERS_DEC)*1)) == NULL)
    errorHandler("Allocating memory for parameters.");

  read_control_file_dec(parmFile, vm_param_dec);
  noteProgress("Completed reading input parameters.");


#ifdef _CHECK_BITS_
  fpe=fopen("decode_bits","w");
#endif

#ifdef _DECODER_
  mzte_codec.target_spatial_levels = vm_param_dec->target_spatial_levels;
  mzte_codec.target_snr_levels = vm_param_dec->target_snr_levels;
#endif

  singleBitFile = vm_param_dec->singleBitFile;
  bitFile = vm_param_dec->bitFile;
  bitFileAC = vm_param_dec->bitFileAC;

  TextureObjectLayer_dec(vm_param_dec, mzte_codec.target_spatial_levels,
			 mzte_codec.target_snr_levels,&wvtfilter); 

#ifdef  _CHECK_BITS_
  fclose(fpe);
#endif


  /***********************************************************************/
  /***********************************************************************/
  /*            DISCRETE INVERSE WAVELET TRANSFORM                       */
  /***********************************************************************/
  /***********************************************************************/
  noteProgress("\nInverse Wavelet Transform....");
  perform_IDWT(vm_param_dec,&wvtfilter);
  noteProgress("Completed inverse wavelet transform.");
  
  noteDetail("Freeing up decoding data structures....");
  for (col=0; col<mzte_codec.colors; col++) {
    free(mzte_codec.SPlayer[col].coeffinfo[0]);
    free(mzte_codec.SPlayer[col].coeffinfo);
  }
  noteDetail("Completed freeing up decoding data structures.");

  noteProgress("\n----- Decoding Completed. -----\n");
  free_vm_param_dec(vm_param_dec);
}

main(int argc, char *argv[])
{
  int i;
  
  /* Set defaults */
  parmFile      = parmFileDef;
  statsFile     = statsFileDef;
  encRecImgFile = encRecImgFileDef;
  decRecImgFile = decRecImgFileDef;

  /*----------------- Process arguments -----------------*/
  /* Parse command line */
  for (i=1; i<argc; ++i)
    if (argv[i][0]=='-' && argv[i][2]=='\0')
      {
	if (argv[i][1]=='p') /* parm file */
	  parmFile = argv[++i];
	else if (argv[i][1]=='s')
	  statsFile = argv[++i];
	else if (argv[i][1]=='d')
	  decRecImgFile= argv[++i];
	else if (argv[i][1]=='e')
	  encRecImgFile = argv[++i];
	else
	  goto USAGE;
      }
    else
      goto USAGE;
  /*----------------- End of process arguments -----------------*/
  
  
#ifndef _DECODER_
  VTC_Encode();
  recImgFile = encRecImgFile;
#else
  recImgFile = decRecImgFile;
#endif

  VTC_Decode();

  return 0;

USAGE:
  fprintf(stderr,"\n%s - compiled on %s", argv[0], __DATE__);
  fprintf(stderr,"\n\nUsage: %s"\
	  "\n\t[-p ParameterFile]"\
	  "\n\t[-s StatisticsFile]"\
	  "\n\t[-e EncoderReconstructedImageFile]"\
	  "\n\t[-d DecoderReconstructedImageFile]"\
	  "\n\t[-h (usage message)]",
	  argv[0]);
  fprintf(stderr,"\n\nDefaults: %s"\
	  "\n\t-p %s"\
	  "\n\t-s %s"\
	  "\n\t-e %s"\
	  "\n\t-d %s\n",
	  argv[0],parmFileDef,statsFileDef,encRecImgFileDef,decRecImgFileDef);
  
  return -1;
}
