/* $Id: read_control_file.c,v 1.9 1998/02/20 21:14:15 hjlee Exp $ */
/****************************************************************************/
/*   MPEG4 Visual Texture Coding (VTC) Mode Software                        */
/*                                                                          */
/*   This software was developed by                                         */
/*   Sarnoff Coporation                   and    Texas Instruments          */
/*   Iraj Sodagar   (iraj@sarnoff.com)           Jie Liang (liang@ti.com)   */
/*   Hung-Ju Lee    (hjlee@sarnoff.com)                                     */
/*   Paul Hatrack   (hatrack@sarnoff.com)                                   */
/*   Shipeng Li     (shipeng@sarnoff.com)                                   */
/*   Bing-Bing Chai (bchai@sarnoff.com)                                     */
/*                                                                          */
/* In the course of development of the MPEG-4 standard. This software       */
/* module is an implementation of a part of one or more MPEG-4 tools as     */
/* specified by the MPEG-4 standard.                                        */
/*                                                                          */
/* The copyright of this software belongs to ISO/IEC. ISO/IEC gives use     */
/* of the MPEG-4 standard free license to use this  software module or      */
/* modifications thereof for hardware or software products claiming         */
/* conformance to the MPEG-4 standard.                                      */
/*                                                                          */
/* Those intending to use this software module in hardware or software      */
/* products are advised that use may infringe existing  patents. The        */
/* original developers of this software module and their companies, the     */
/* subsequent editors and their companies, and ISO/IEC have no liability    */
/* and ISO/IEC have no liability for use of this software module or         */
/* modification thereof in an implementation.                               */
/*                                                                          */
/* Permission is granted to MPEG memebers to use, copy, modify,             */
/* and distribute the software modules ( or portions thereof )              */
/* for standardization activity within ISO/IEC JTC1/SC29/WG11.              */
/*                                                                          */
/* Copyright (C) 1998  Sarnoff Coporation and Texas Instruments             */ 
/****************************************************************************/

/************************************************************/
/*     Sarnoff Very Low Bit Rate Still Image Coder          */
/*     Copyright 1995, 1996, 1997, 1998 Sarnoff Corporation */
/************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "dataStruct.h"
#include "errorHandler.h"
extern int STO_const_alpha;
extern unsigned char STO_const_alpha_value;
static char *hasIntFmt(char *s)
{
  if ((s=strchr(s,'%'))!=NULL)
  {
    ++s;
    while (*s >= '0' && *s <= '9')
      s++;
    
    s = (*s == 'd') ? (s+1) : NULL;
  }

  return s;
}

static void checkParams(SOL_PARAMETERS *vm_param)
{
  char *s;

  /* Check parameters */
  
  /* check first bit file */
  if (strchr(vm_param->bitFile,'%') != NULL)
    errorHandler("'%s' is an invalid 'first' bitfile name. '%%' not allowed.",
		 vm_param->bitFile);

  if (vm_param->singleBitFile==0)
  {
    if (vm_param->quantization_type==1)
    {
      if (strchr(vm_param->bitFileAC,'%') != NULL)
	errorHandler("'%s' is an invalid AC bitfile name for single-quant.\n"\
		     "\t'%%' not allowed.",
		     vm_param->bitFileAC);
    }
    else if (vm_param->quantization_type==2)
    {
      if ((s=hasIntFmt(vm_param->bitFileAC))==NULL)
	errorHandler("'%s' is an invalid AC bitfile format for "\
		     "multi-quant.\n"\
		     "\tNeed an integer format.",
		     vm_param->bitFileAC);
      
      /* bitFileAC must exist and have two %d's */
      if (vm_param->SNR_start_code_enable)
      {
	if ((s=hasIntFmt(s))==NULL)
	  errorHandler("'%s' is an invalid AC bitfile format for "\
		       "multi-quant.\n"\
		       "\tSNR start code is enabled - need two "\
		       "integer formats.", vm_param->bitFileAC);
      }
      else
      {
	if ((s=strchr(s,'%')) != NULL)
	    errorHandler("'%s' is an invalid AC bitfile format for "\
		       "multi-quant.\n"\
		       "\tSNR start code is not enabled - only need one "\
		       "integer format.", vm_param->bitFileAC);
      }
    }
  }
}


/*--------------------------------------------------------------------
  read_control_file()
  
  Return Values
  -------------
  None.
  
  Comments
  --------
  Exits if error occurs.
  --------------------------------------------------------------------*/

void read_control_file(char *ctlfname, SOL_PARAMETERS *vm_param)
{
  FILE *vminfp;
  char *garbage=(char *)malloc(sizeof(char)*80);
  int  spa_lev,snr_lev,col;
  int  SNR_scalability_levels;
  
  if (garbage==NULL)
    errorHandler("Malloc error \n");
  
  if ( (vminfp = fopen(ctlfname,"r")) == NULL) 
    errorHandler("Can't open the control file '%s'\n",ctlfname);
  
  /* read the path to the source image */
  if ((vm_param->ImagePath = (char *)malloc(sizeof(char)*100)) == NULL)
    errorHandler("error allocating memory \n");
  fscanf(vminfp,"%s",vm_param->ImagePath);
  fgets(garbage,80,vminfp);

  /* read the path to the segmentation map of the source image */
  if ((vm_param->SegPath = (char *)malloc(sizeof(char)*100)) == NULL)
    errorHandler("error allocating memory \n");
  fscanf(vminfp,"%s",vm_param->SegPath);
  fgets(garbage,80,vminfp);

  /*read the output bitstream filename */
  if ((vm_param->bitFile = (char *)malloc(sizeof(char)*100)) == NULL)
    errorHandler("Allocating memory.");
  fscanf(vminfp,"%d %s",&vm_param->singleBitFile,vm_param->bitFile);
  if (vm_param->singleBitFile==0)
  {
    if ((vm_param->bitFileAC = (char *)malloc(sizeof(char)*100)) == NULL)
      errorHandler("Allocating memory.");

    fscanf(vminfp,"%s",vm_param->bitFileAC);
  }
  fgets(garbage,80,vminfp);


  /*read the number of color components of the source image */
  fscanf(vminfp,"%ld",&vm_param->colors);
  fgets(garbage,80,vminfp);

  if ((vm_param->Image = 
       (PICTURE *)malloc(sizeof(PICTURE)*vm_param->colors)) == NULL)
    errorHandler("error allocating memory \n");

  if ((vm_param->SegImage = 
       (PICTURE *)malloc(sizeof(PICTURE)*1)) == NULL)
    errorHandler("error allocating memory \n");

  /* read the bit-depth of the source image */
  fscanf(vminfp,"%ld",&vm_param->bit_depth);
  fgets(garbage,80,vminfp);
  
  /* read the width of the source image */
  fscanf(vminfp,"%ld",&vm_param->Image[0].width);
  fgets(garbage,80,vminfp);
  /* vm_param->Image[0].width = width; */
  
  /* read the height of the source image */
  fscanf(vminfp,"%ld",&vm_param->Image[0].height);
  fgets(garbage,80,vminfp);
  
  /* read filter type, currently, there are 10 filters available in
     default.h */
  fscanf(vminfp,"%ld",&vm_param->wvtType);
  fgets(garbage,80,vminfp);
  
  /* read wavelet decomposition levels */
  fscanf(vminfp,"%ld",&vm_param->wvtDecompLev);
  fgets(garbage,80,vminfp);
  
  /* read quantization type */
  fscanf(vminfp,"%ld",&vm_param->quantization_type);
  fgets(garbage,80,vminfp);
  
  /* read  SNR start code enable*/
  fscanf(vminfp,"%ld",&vm_param->SNR_start_code_enable);
  fgets(garbage,80,vminfp);
  
  
  fscanf(vminfp,"%ld",&vm_param->target_bitrate);
  fgets(garbage,80,vminfp);
  
  fscanf(vminfp,"%ld",&vm_param->target_spatial_levels);
  fgets(garbage,80,vminfp);
  
  fscanf(vminfp,"%ld",&vm_param->target_snr_levels);
  fgets(garbage,80,vminfp);
  
  /* Read in deringing parameters. If either is zero deringing is disabled */
  fscanf(vminfp,"%ld %ld",
         &vm_param->deringWindowSize,&vm_param->deringThreshold);
  fgets(garbage,80,vminfp);

  /* Read in DC Lum Quant */
  fscanf(vminfp,"%ld",&vm_param->QdcY);
  fgets(garbage,80,vminfp);
  if (vm_param->colors != MONO) { 
    /* Read in DC Chroma Quant */
    fscanf(vminfp,"%ld",&vm_param->QdcUV);
    fgets(garbage,80,vminfp);
  }
  
  /* Read in the spatial scalability levels */
  fscanf(vminfp,"%ld",&vm_param->spatial_scalability_levels);
  fgets(garbage,80,vminfp);
  for (col=0; col<vm_param->colors; col++) {
      if ((vm_param->SPlayer[col] = (SNR_PARAM *)malloc(sizeof(SNR_PARAM)*
			   vm_param->spatial_scalability_levels )) == NULL)
	errorHandler("error allocating memory \n");
  }

  for (spa_lev=0; spa_lev<vm_param->spatial_scalability_levels; spa_lev++) {
    fscanf(vminfp,"%ld",&SNR_scalability_levels);

    for (col=0; col<vm_param->colors; col++) {
         vm_param->SPlayer[col][spa_lev].SNR_scalability_levels = 
	                           SNR_scalability_levels;
         if ((vm_param->SPlayer[col][spa_lev].Quant = 
	      (int *)malloc(sizeof(int)*SNR_scalability_levels))==NULL)
	   errorHandler("error allocating memory \n");
    }

    for (snr_lev=0; snr_lev<SNR_scalability_levels; snr_lev++) {
	fscanf(vminfp,"%ld",&vm_param->SPlayer[0][spa_lev].Quant[snr_lev]);
	if (vm_param->colors != MONO) { 
	  fscanf(vminfp,"%ld",&vm_param->SPlayer[1][spa_lev].Quant[snr_lev]);
	  vm_param->SPlayer[2][spa_lev].Quant[snr_lev] =
	    vm_param->SPlayer[1][spa_lev].Quant[snr_lev];
	}
    }

    fgets(garbage,80,vminfp);
  }
  fclose(vminfp);
  free(garbage);
  /* need to read the following from the  parameter file */
  vm_param->alphaTH = 0;
  vm_param->change_CR_disable = 1;
  STO_const_alpha=vm_param->STO_const_alpha = 1;
  STO_const_alpha_value = vm_param->STO_const_alpha_value = 0xff;
  checkParams(vm_param);
}

/*--------------------------------------------------------------------
  read_control_file_dec()
  
  Return Values
  -------------
  None.
  
  Comments
  --------
  Exits if error occurs.
  --------------------------------------------------------------------*/

void read_control_file_dec(char *ctlfname, SOL_PARAMETERS_DEC *vm_param)
{
  FILE *vminfp;
  char *garbage=(char *)malloc(sizeof(char)*80);
  
  if (garbage==NULL)
    errorHandler("Malloc error \n");
  
  if ( (vminfp = fopen(ctlfname,"r")) == NULL) 
    errorHandler("can't find the control file %s\n",ctlfname);
  
  /* Throw away line: read the path to the source image */
  fgets(garbage,80,vminfp);

  /* read the path to the segmentation map of the source image */
  /*  if ((vm_param->SegPath = (char *)malloc(sizeof(char)*100)) == NULL)
      errorHandler("error allocating memory \n");
  fscanf(vminfp,"%s",vm_param->SegPath);*/
  fgets(garbage,80,vminfp);

  /*read the number of color components of the source image */
  if ((vm_param->bitFile = (char *)malloc(sizeof(char)*100)) == NULL)
    errorHandler("Allocating memory.");
  fscanf(vminfp,"%d %s",&vm_param->singleBitFile,vm_param->bitFile);
  if (vm_param->singleBitFile==0)
  {
    if ((vm_param->bitFileAC = (char *)malloc(sizeof(char)*100)) == NULL)
      errorHandler("Allocating memory.");

    fscanf(vminfp,"%s",vm_param->bitFileAC);
  }
  fgets(garbage,80,vminfp);


  /* Throw away line: read the number of color components of the source 
     image */
  fscanf(vminfp,"%ld",&vm_param->colors);
  fgets(garbage,80,vminfp);

  if ((vm_param->SegImage = 
       (PICTURE *)malloc(sizeof(PICTURE)*1)) == NULL)
    errorHandler("error allocating memory \n");

  /* read the bit-depth of the source image */
  fscanf(vminfp,"%ld",&vm_param->bit_depth);
  fgets(garbage,80,vminfp);
  
  /* Throw away line: read the width of the source image */
  fscanf(vminfp,"%ld",&vm_param->width);
  fgets(garbage,80,vminfp);
  
  /*  Throw away line: read the height of the source image */
  fscanf(vminfp,"%ld",&vm_param->height);
  fgets(garbage,80,vminfp);
  
  /* read filter type */
  fscanf(vminfp,"%ld",&vm_param->wvtType);
  fgets(garbage,80,vminfp);
  
  /*  Throw away line: read wavelet decomposition levels */
  fscanf(vminfp,"%ld",&vm_param->wvtDecompLev);
  fgets(garbage,80,vminfp);
  
  /*  Throw away line: read quantization type */
  fgets(garbage,80,vminfp);
  
  /*  Throw away line: read  SNR start code enable*/
  fgets(garbage,80,vminfp);
  
  
  fscanf(vminfp,"%ld",&vm_param->target_bitrate);
  fgets(garbage,80,vminfp);
  
  fscanf(vminfp,"%ld",&vm_param->target_spatial_levels);
  fgets(garbage,80,vminfp);
  
  fscanf(vminfp,"%ld",&vm_param->target_snr_levels);
  fgets(garbage,80,vminfp);
  
  /* Read in deringing parameters. If either is zero deringing is disabled */
  fscanf(vminfp,"%ld %ld",
         &vm_param->deringWindowSize,&vm_param->deringThreshold);
  fgets(garbage,80,vminfp);

  /*  Throw away line: Read in DC Lum Quant */
  fgets(garbage,80,vminfp);
  if (vm_param->colors != MONO) { 
    /*  Throw away line: Read in DC Chroma Quant */
    fgets(garbage,80,vminfp);
  }
  

  fclose(vminfp);
  free(garbage);

  /*checkParams(vm_param);*/
}

void free_vm_param(SOL_PARAMETERS* vm_param)
{
  free(vm_param->ImagePath);
  free(vm_param->SegPath);
  free(vm_param->bitFile);
  if (vm_param->singleBitFile==0)
    free(vm_param->bitFileAC);
  /* free(vm_param->Image); */
  free(vm_param->SegImage);
  /*for (col=0; col<vm_param->colors; col++) 
    free(vm_param->SPlayer[col]);
    for (spa_lev=0; spa_lev<vm_param->spatial_scalability_levels; spa_lev++) {
    for (col=0; col<vm_param->colors; col++) {
    free(vm_param->SPlayer[col][spa_lev].Quant);
    }
  }*/
  free(vm_param);
}

void free_vm_param_dec(SOL_PARAMETERS_DEC *vm_param_dec)
{
  free(vm_param_dec->bitFile);
  if (vm_param_dec->singleBitFile==0)
    free(vm_param_dec->bitFileAC);
  free(vm_param_dec->SegImage);
  free(vm_param_dec);
}
