/****************************************************************************/
/*   MPEG4 Visual Texture Coding (VTC) Mode Software                        */
/*                                                                          */
/*   This software was jointly developed by                                 */
/*   Sarnoff Coporation                   and    Texas Instruments          */
/*   Iraj Sodagar   (iraj@sarnoff.com)           Jie Liang (liang@ti.com)   */
/*   Hung-Ju Lee    (hjlee@sarnoff.com)                                     */
/*   Paul Hatrack   (hatrack@sarnoff.com)                                   */
/*   Shipeng Li     (shipeng@sarnoff.com)                                   */
/*   Bing-Bing Chai (bchai@sarnoff.com)                                     */
/*                                                                          */
/* In the course of development of the MPEG-4 standard. This software       */
/* module is an implementation of a part of one or more MPEG-4 tools as     */
/* specified by the MPEG-4 standard.                                        */
/*                                                                          */
/* The copyright of this software belongs to ISO/IEC. ISO/IEC gives use     */
/* of the MPEG-4 standard free license to use this  software module or      */
/* modifications thereof for hardware or software products claiming         */
/* conformance to the MPEG-4 standard.                                      */
/*                                                                          */
/* Those intending to use this software module in hardware or software      */
/* products are advised that use may infringe existing  patents. The        */
/* original developers of this software module and their companies, the     */
/* subsequent editors and their companies, and ISO/IEC have no liability    */
/* and ISO/IEC have no liability for use of this software module or         */
/* modification thereof in an implementation.                               */
/*                                                                          */
/* Permission is granted to MPEG memebers to use, copy, modify,             */
/* and distribute the software modules ( or portions thereof )              */
/* for standardization activity within ISO/IEC JTC1/SC29/WG11.              */
/*                                                                          */
/* Copyright (C) 1998  Sarnoff Coporation  and Texas Instruments            */ 
/****************************************************************************/

/****************************************************************************/
/*     Texas Instruments Predictive Embedded Zerotree (PEZW) Image Codec    */
/*     Copyright 1996, 1997, Texas Instruments	      		            */
/****************************************************************************/

/* utilities for belevel coding and decoding
   Jie Liang (Texas Instruments)  liang@ti.com                */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "globals.h"
#include "msg.h"
#include "bitpack.h"
#include "startcode.h"

#include "wvtPEZW.h"
#include "PEZW_zerotree.h"

static int shiftbits;
static long out_len;

/* this funciton is to create and initiate the data structure
   used for bilevel image coding */

PEZW_SPATIAL_LAYER *Init_PEZWdata (int color, int levels, int w, int h)
{
  PEZW_SPATIAL_LAYER *SPlayer;
  int x,y;
  int i;

  /* generate the Spatial Layer Structure */
  SPlayer = (PEZW_SPATIAL_LAYER *)calloc (levels, sizeof(PEZW_SPATIAL_LAYER));
  for (i=0; i<levels; i++)
     SPlayer[i].SNRlayer = calloc(MAXSNRLAYERS, sizeof(PEZW_SNR_LAYER));
     

  /* the original wavelet coefficients are in 
     SPlayer[0].SNRlayer[0].snr_image   */
  SPlayer[0].SNRlayer[0].snr_image.height = h;
  SPlayer[0].SNRlayer[0].snr_image.width = w;
  if ((SPlayer[0].SNRlayer[0].snr_image.data = calloc(h*w, sizeof(WINT)))
      == NULL){
    noteError ("Can not allocate memory in Init_PEZWdata()");
    exit(-1);
  }
  
  
  /* copy the wavelet coefficients into SPlayer structure */
  for (y=0;y<h;y++)
    for (x=0;x<w;x++)
      ((WINT *)(SPlayer[0].SNRlayer[0].snr_image.data))[y*w+x] = (WINT) 
	  (mzte_codec.SPlayer[color].coeffinfo[y][x].wvt_coeff);

  return (SPlayer);  

}

/* put the restored wavelet coefficients back to the data structure */
void restore_PEZWdata (PEZW_SPATIAL_LAYER **SPlayer)
{
  int x,y;
  int col;
  int h,w;
  int dch, dcw;
  int levels;

  for(col=0;col<mzte_codec.colors;col++){
    h = SPlayer[col][0].SNRlayer[0].snr_image.height;
    w = SPlayer[col][0].SNRlayer[0].snr_image.width;    
    
    if(col==0)
      levels=mzte_codec.wvtDecompLev;
    else
      levels=mzte_codec.wvtDecompLev-1;

    dch = h/(1<<levels);
    dcw = w/(1<<levels);

    /* copy the wavelet coefficients into data structure */
    for (y=0;y<h;y++)
      for (x=0;x<w;x++){
	if((x>=dcw) || (y>=dch))
	  mzte_codec.SPlayer[col].coeffinfo[y][x].rec_coeff = 	
	    ((WINT *)(SPlayer[col][0].SNRlayer[0].snr_image.data))[y*w+x]; 
      }
  }

  return;

}

/* this function pack the bitstream and write to file */
void PEZW_bitpack (PEZW_SPATIAL_LAYER **SPlayer)
{
  int levels = mzte_codec.wvtDecompLev;
  int levels_UV;
  int Quant[NCOLOR];
  int texture_spatial_layer_id;

  int lev;
  int col;
  int snr_scalability_levels=0;
  int snr_lev;

  int all_zero;
  int n;

  unsigned char *data;
  long len;

  levels_UV = levels-1;  /* this should depend on the color format 
			    mzte_codec.color_format */

  /* get quantization information */
  for (col=0;col<mzte_codec.colors; col++)
    Quant[col] = mzte_codec.Qinfo[col]->Quant[0];

  for (lev=0;lev<levels;lev++){   /* spatial scalability level */
    /*------- AC: Write header info to bitstream file -------*/
    flush_bits();
    emit_bits(TEXTURE_SPATIAL_LAYER_START_CODE >> 16, 16);
    emit_bits(TEXTURE_SPATIAL_LAYER_START_CODE, 16);
    texture_spatial_layer_id = lev;
    emit_bits(texture_spatial_layer_id, 5);

    /* output quant for each color component */
    for (col=0;col<mzte_codec.colors; col++)    
      emit_bits(Quant[col],8);
      
    /* output snr_scalability_levels: choose the maximum
       snr_scalability_levels to put here for all color components */
    if (lev<levels_UV){
      for (col=0;col<mzte_codec.colors; col++){
	if (snr_scalability_levels < SPlayer[col][lev].SNR_scalability_levels)
	  snr_scalability_levels = SPlayer[col][lev].SNR_scalability_levels;
      }
    }
    else  /* only luminance component at this level */
        snr_scalability_levels = SPlayer[0][lev].SNR_scalability_levels;

    emit_bits(snr_scalability_levels, 5);

    for (snr_lev=0;snr_lev<snr_scalability_levels;snr_lev++){
      
      /* output the SNR_START_CODE if enabled */
      if (mzte_codec.SNR_start_code_enable)
	{
	  flush_bits();
	  emit_bits(texture_snr_layer_start_code >> 16, 16);
	  emit_bits(texture_snr_layer_start_code, 16);
	  emit_bits(snr_lev,5);
	}

      emit_bits_checksc_init();

      for (col=0;col<mzte_codec.colors; col++){ /* for each color compoent */
	/* output SNR_ALL_ZERO flag */
	if ((snr_lev >= SPlayer[col][lev].SNR_scalability_levels) ||
	    ((lev>=levels_UV) && (col!=0)))    /* no color bits when lev>levels_UV */
	  all_zero = 1;
	else
	  all_zero = 0;

	emit_bits_checksc (all_zero,1); 
	  
	/* send the bitstream for this snr layer if not all_zero */
	if (!all_zero){
	  data = (char *)SPlayer[col][lev].SNRlayer[snr_lev].snr_bitstream.data;
	  len = SPlayer[col][lev].SNRlayer[snr_lev].snr_bitstream.length;
	  
	  /* output the bitstream */
	  for (n=0;n<len;n++){
	    emit_bits_checksc (data[n],8);
	    if (DEBUG_BS_DETAIL)
	      fprintf(stdout,"%d ", data[n]);
	  }

	  if(DEBUG_BS)
	    fprintf(stdout, "color %d spa_lev %d snr_lev %d: %d\n",
		    col, lev, snr_lev, len);


	} /* end of !all_zero */
      } /* end of color plane */      
    }	/* end of snr_scalability */
  }  /* end of spatial scalability */
}

/* unpacke the bitstream into Spatial and SNR layer bitstreams 
   SPlayer data structure will be created here  */

void PEZW_bit_unpack (PEZW_SPATIAL_LAYER **SPlayer)
{
   int splev, snrlev;
   int spatial_id, snr_id;
   int color;
   int Quant[3];
   int snr_scalability_levels;
   int texture_spatial_layer_start_code;
   int snr_layer_start_code;
   unsigned char buffer[MAXSIZE];
   long len;
   int h,w;
   int n, all_zero;
   int Snrlevels;

   /* initialize data structure */
   h=mzte_codec.height;
   w=mzte_codec.width;
   for (color=0;color<mzte_codec.colors;color++)
       SPlayer[color] = (PEZW_SPATIAL_LAYER *)calloc (mzte_codec.wvtDecompLev,
					     sizeof(PEZW_SPATIAL_LAYER));
   


   for (splev=0;splev<mzte_codec.wvtDecompLev;splev++)
     {
       /* check start code */
       
       align_byte();
       texture_spatial_layer_start_code = get_X_bits(32);
       if (texture_spatial_layer_start_code != 
	   TEXTURE_SPATIAL_LAYER_START_CODE)
		 errorHandler("Wrong texture_spatial_layer_start_code.");
       spatial_id = get_X_bits(5);

       /* get the quantization step sizes */
       for (color=0;color<mzte_codec.colors;color++)
	 Quant[color]= get_X_bits(8);
	 
       /* get the snr scalability levels 
	  this is the maximum level for all three color components,
	  it will be adjusted later on according to all_zero flag */
       
       snr_scalability_levels = get_X_bits(5);

       for (color=0;color<mzte_codec.colors;color++){
	 SPlayer[color][splev].SNR_scalability_levels = snr_scalability_levels;
       
	 /* allocate SNR data structure */
	 SPlayer[color][splev].SNRlayer = (PEZW_SNR_LAYER *)
	   calloc(snr_scalability_levels, sizeof(PEZW_SNR_LAYER));
	 SPlayer[color][splev].SNRlayer[0].Quant = Quant[color];

	 if (color==0){
	   SPlayer[color][0].SNRlayer[0].snr_image.height = h;
	   SPlayer[color][0].SNRlayer[0].snr_image.width =  w;
	 }
	 else{
	   SPlayer[color][0].SNRlayer[0].snr_image.height = h/2;
	   SPlayer[color][0].SNRlayer[0].snr_image.width =  w/2;
	 }
	 
	 if ((SPlayer[color][0].SNRlayer[0].snr_image.data = calloc(h*w, sizeof(WINT)))
	     == NULL){
	   noteError ("Can not allocate memory in Init_PEZWdata()");
	   exit(-1);
	 }

       }

       /* get the snr layer */
       if (mzte_codec.SNR_start_code_enable)
	 Snrlevels=snr_scalability_levels;
       else
	 Snrlevels=1;

       for (snrlev=0;snrlev<Snrlevels;snrlev++)
	 {
	   if (mzte_codec.SNR_start_code_enable)
	     {
	       align_byte();
	       snr_layer_start_code = get_X_bits(32);
	       if (snr_layer_start_code != texture_snr_layer_start_code)
		 errorHandler("Wrong texture_snr_layer_start_code.");
	       snr_id = get_X_bits(5);
	     }

	   len=0;
	   
	   get_X_bits_checksc_init();

	   /* the all_zero flag for luminance */
	   all_zero=get_X_bits_checksc(1);
	   
	   /* first put all the data in SPlayer[0] */
	   /* the separation of the color components bitstreams will
	      be done by the decoding process */

	   if ((splev==mzte_codec.wvtDecompLev-1)&&
		(snrlev == Snrlevels-1))
	     len=get_allbits_checksc(buffer);

	   else{
	     if ((!mzte_codec.SNR_start_code_enable) /* only one bitstream */
		 || (snrlev==Snrlevels-1))
	       {
		 /* look for the next spatial layer start code */
		 while(1){
		   /* chekc for the next start code*/
		   if(!Is_startcode(TEXTURE_SPATIAL_LAYER_START_CODE)){
		     buffer[len] = get_X_bits_checksc(8);
		     len++;
		   }
		   else{
		     buffer[len++]=align_byte_checksc();
		     break; /* start code found */
		   }
		 } /* end of while */
	       }
	     else{
	       while(1){
		 /* chekc for the next start code*/
		 if(!Is_startcode(texture_snr_layer_start_code)){
		   buffer[len] = get_X_bits_checksc(8);
		   len++;
		 }
		 else{
		   buffer[len++]=align_byte_checksc();
		   break; /* start code found */
		 }
	       } /* end of while(1) */
	     } /* end of if(!mzte_...) */
	   } /* end of if(splev== ...) */

	   SPlayer[0][splev].SNRlayer[snrlev].snr_bitstream.length = len;
	   SPlayer[0][splev].SNRlayer[snrlev].snr_bitstream.data =
	     (char *)calloc(len,sizeof(char));
	   memcpy(SPlayer[0][splev].SNRlayer[snrlev].snr_bitstream.data,
		  buffer,len);
	   
	   if(DEBUG_BS_DETAIL)
	     for(n=0;n<len;n++)
	       fprintf(stdout,"%d ", 
		  ((unsigned char *)SPlayer[0][splev].SNRlayer[snrlev].snr_bitstream.data)[n]);

	   if (DEBUG_BS)
	     fprintf(stdout,"spa_lev %d snr_lev %d: %d\n", splev, snrlev, len);

	 }  /* snr levels */       
     }  /* spatial levels */
   
}

/* left shift a stream by N<8 bits */

int lshift_by_NBit (unsigned char *data, int len, int N)
{
  int n;
  char mask=0;
  int output;

  if(len==0)
    return 1;

  output = data[0]>>(8-N);

  for(n=0;n<N;n++)
    mask = (mask<<1) | 0x01;

  for (n=0;n<len-1;n++)
    data[n] = (data[n]<<N) | ((data[n+1]>>(8-N)) & mask);

  data[len-1] <<= N;
			    
  return output;
}


