/*
 *	  Binary Arithmetic Coder & Decoder Function  (BinArCodec.c)
 */
/* This code was obtained from Lehigh University/Vector Vision, Inc.. 
It was originally developed by OKI. 
Dr. Hongqiao Sun/Weiping integrated it to the Lehigh/VVI software. 

The copyrights or credits should belong to the following people and companies:

Zhixiong Wu and Yoshihiro Ueda of OKI,
Hongqiao Sun of Vector Vision,
Weiping Li of Lehigh University,

Minor modifications was made by Shipeng Li to integrate it to Sarnoff's SOL
software.
*/
#include <stdio.h>
#include <math.h>
#include "bitpack.h"
#include "ShapeCommon.h"
#include "BinArCodec.h"

/***************************/
/*      ArCodec            */
/***************************/


void
StartArCoder_Still(ArCoder *coder) 
{
    coder -> L = 0;
    coder -> R = HALF-1;
    coder -> bits_to_follow = 0;
    coder -> first_bit = 1;
    coder -> nzeros = MAXHEADING;
    coder -> nonzero = 0;
}

void
StopArCoder_Still(ArCoder *coder,BSS *bitstream)
{
    int a = ( coder -> L )              >> ( CODE_BITS - 3 );
    int b = ( coder -> R + coder -> L ) >> ( CODE_BITS - 3 );

    int nbits, bits, i;

    if ( b == 0 ) b = 8;

    if ( b-a >= 4 || (b-a == 3 && (a & 1) == 1 ) ) {
        nbits = 2;
        bits  = (a>>1) + 1;
    } else {
        nbits = 3;
        bits  = a + 1;
    }
  
    for ( i=1; i<=nbits; i++ ) {
        BitPlusFollow_Still ( ((bits >> (nbits-i)) & 1), coder, bitstream );
    }
	
    if ( coder -> nzeros < MAXMIDDLE-MAXTRAILING || coder -> nonzero == 0 ) {
        BitPlusFollow_Still( 1, coder, bitstream );
    }

    return;
}

/* ENCODE A BINARY SYMBOL */

void
ArCodeSymbol_Still(ArCoder *coder,BSS *bitstream,unsigned char bit,unsigned int c0)
{
	int		bits = 0;
	unsigned int	c1   = (1<<16) - c0;
  
	unsigned char	LPS  = (c0 > c1);
	unsigned int	cLPS = (LPS) ? c1 : c0;

	unsigned long rLPS;
	rLPS = ((coder -> R) >> 16) * cLPS;
  
	if (bit == LPS) {
		coder -> L += coder->R - rLPS;
		coder -> R  = rLPS;
	} else {
		coder -> R -= rLPS;
	}

	EncRenormalize( coder, bitstream );
}

void
EncRenormalize(ArCoder *coder,BSS *bitstream)
{
    while ( coder -> R < QUARTER ) {
        if( coder -> L >= HALF ) {				 
            BitPlusFollow_Still( 1, coder, bitstream );
            coder -> L -= HALF;
        } else {
            if ( coder -> L + coder -> R <= HALF ) {
                BitPlusFollow_Still( 0, coder, bitstream );
            } else {				 
                coder -> bits_to_follow++;		 
                coder -> L -= QUARTER;		 
            }
        }
        coder -> L += coder -> L;				 
        coder -> R += coder -> R;				 
    }
}

void
BitByItself_Still(int bit,ArCoder *coder,BSS *bitstream) 
{
    BitstreamPutBit_Still( bit, bitstream );	/* Output the bit */
    if ( bit == 0 ) {
        coder -> nzeros --;
        if ( coder -> nzeros == 0 ) {
            BitstreamPutBit_Still( 1, bitstream );
            coder -> nonzero = 1;
            coder -> nzeros  = MAXMIDDLE;
        }
    } else {
        coder -> nonzero = 1;
        coder -> nzeros  = MAXMIDDLE;
    }
}

/* OUTPUT BITS PLUS FOLLOWING OPPOSITE BITS */

void
BitPlusFollow_Still(int bit,ArCoder *coder,BSS *bitstream)
{
	if ( ! coder -> first_bit )
		BitByItself_Still( bit, coder, bitstream );
	else
		coder -> first_bit = 0;
	while ( (coder -> bits_to_follow) > 0 ) {
		BitByItself_Still( !bit, coder, bitstream );
		coder -> bits_to_follow -= 1;
	}
}


/*  Bitstream operation function in CAE */

void
BitstreamPutBit_Still(int bit,BSS *bitstream)
{
	int		res = bitstream -> res;
	unsigned char	*p  = bitstream -> ptr;

	*p |= (unsigned char)( ( bit & 1 ) << ( 7 - res ) );
	res ++;
	if ( res == 8 ) {
		*(++p) = (unsigned char)0;
		res    = 0;
	}
	bitstream -> ptr = p;
	bitstream -> res = res;
	bitstream -> cnt ++;
}

/***************************/
/*      Decoder            */
/***************************/

void
StartArDecoder_Still(ArDecoder *decoder) 
{
	int		i, j;
  
	decoder -> V			= 0;
	decoder -> extrabits	= 0;
	decoder -> nzerosf		= MAXHEADING;

	for ( i=1; i<CODE_BITS; i++ ) {
		j = BitstreamLookBit( i + decoder -> extrabits);
		decoder -> V += decoder -> V + j;
		if ( j == 0 ) {
			decoder -> nzerosf--;
			if (decoder -> nzerosf == 0) {
				decoder -> extrabits ++;
				decoder -> nzerosf   = MAXMIDDLE;
			}
		} else {
			decoder -> nzerosf = MAXMIDDLE;
		}
	}

	decoder -> L				= 0;
	decoder -> R				= HALF - 1;
	decoder -> bits_to_follow	= 0;
	decoder -> arpipe			= decoder -> V;
	decoder -> nzeros			= MAXHEADING;
	decoder -> nonzero			= 0;

	return;
}

  
void
StopArDecoder_Still(ArDecoder *decoder) 
{
	int		a = ( decoder -> L >> ( CODE_BITS - 3 ) );
	int		b = ( decoder -> R + decoder -> L ) >> ( CODE_BITS - 3 );

	int		nbits, i;

	if (b == 0) {
		b =  8;
	}

	if ( (b-a) >= 4 || ( (b-a) == 3 && (a & 1) == 1 ) )	nbits = 2;
	else												nbits = 3;

	for ( i=1; i<nbits; i++ )
		AddNextInputBit_Still( decoder);	
  
	if ( decoder->nzeros<MAXMIDDLE-MAXTRAILING || decoder->nonzero==0 ) {
		BitstreamFlushBits_Still( 1);
		nbits ++;
	}

	return;
}


unsigned char
ArDecodeSymbol_Still(ArDecoder *decoder,unsigned int c0)
{
	int		bit;
	unsigned int	c1		= (1<<16) - c0;
	unsigned char	LPS		= (c0 > c1);
	unsigned int	cLPS	= LPS ? c1 : c0;
	unsigned int	rLPS;

	rLPS = (decoder -> R >> 16) * cLPS;
	if ( (decoder -> V - decoder -> L) >= (decoder -> R - rLPS) ) {
		bit			  = LPS;
		decoder -> L += decoder -> R - rLPS;
		decoder -> R  = rLPS;
	} else {
		bit			  = (1-LPS);
		decoder -> R -= rLPS;
	}
	DecRenormalize( decoder);

	return (bit);
}


void
AddNextInputBit_Still(ArDecoder *decoder)
{
	int		i;

	if ( ((decoder -> arpipe >> (CODE_BITS - 2)) & 1) == 0 ) {
		decoder -> nzeros--;
		if ( decoder -> nzeros == 0 ) {
			BitstreamFlushBits_Still( 1);
			decoder -> extrabits--;
			decoder -> nzeros	= MAXMIDDLE;
			decoder -> nonzero	= 1;
		}
	} else {
		decoder -> nzeros	= MAXMIDDLE;
		decoder -> nonzero	= 1;
	}

	BitstreamFlushBits_Still( 1);

	i = BitstreamLookBit( (CODE_BITS-1+decoder->extrabits));

	decoder -> V		+= decoder -> V + i;
	decoder -> arpipe	+= decoder -> arpipe + i;

	if ( i == 0 ) {
		decoder -> nzerosf--;
		if (decoder -> nzerosf == 0) {
			decoder -> nzerosf = MAXMIDDLE;
			decoder -> extrabits++;
		}
	} else {
		decoder -> nzerosf = MAXMIDDLE;
	}
	return;

}

void
DecRenormalize(ArDecoder *decoder)
{
	while ( decoder -> R < QUARTER ) {
		if (decoder -> L >= HALF) {
			decoder -> V -= HALF;
			decoder -> L -= HALF;
			decoder -> bits_to_follow = 0;
		} else {
			if (decoder -> L + decoder -> R <= HALF) {
				decoder -> bits_to_follow = 0;
			} else {
				decoder -> V -= QUARTER;
				decoder -> L -= QUARTER;
				(decoder -> bits_to_follow)++;
			}
		}
		decoder -> L += decoder -> L;	
		decoder -> R += decoder -> R;
		AddNextInputBit_Still( decoder );
	}
	return;
}

/*  Bitstream Operation Functions */


void
InitBitstream(int flag,BSS *bitstream)
{
	bitstream -> ptr = bitstream->bs;
	bitstream -> cnt = 0;
	bitstream -> res = 0;
	if ( flag == 1 )
	*bitstream -> ptr = (unsigned char)0;
}

int
ByteAlignmentEnc()
{
  flush_bits_zeros();
  return(0);
}

int
ByteAlignmentDec()
{
  align_byte();
  return(0);
}

void
PutBitstoStream(int bits,unsigned int code,BSS *bitstream)
{
	unsigned char	*p	= bitstream -> ptr;
	int		res	= bitstream -> res;
	bitstream -> cnt += bits;
	while ( bits > 0 ) {
		*p |= (unsigned char) ( ((code >> (bits-1)) & 1) << (7-res) );
		res ++;
		bits--;
		if ( res == 8 ) {
			*(++p) = (unsigned char)0;
			res    = 0;
		}
	}
	bitstream -> ptr  = p;
	bitstream -> res  = res;

	return;
}
void
PutBitstoStreamMerge(int bits,unsigned int code)
{
  while(bits>16) {
    emit_bits(code>>(bits-16), 16);
    bits-=16;
  }
  
  emit_bits(code, bits);
  return;
}

void
BitstreamFlushBits_Still(int i)
{
  get_X_bits(i);
  return;
}


int
BitstreamLookBit(int pos)
{
  return((int )(LookBitsFromStream(pos) &1 ) );
}

unsigned int
GetBitsFromStream(int bits)
{
  unsigned int code;
  code = get_X_bits(bits);
  return ( code );
}

void
BitstreamFlushBitsCopy(int i,BSS *bitstream)
{
	int		mdl = 0;
	int		res = bitstream -> res + i;

	while ( res >= 8 ) {
		mdl ++;
		res -= 8;
	}
	bitstream -> ptr	+= mdl;
	bitstream -> cnt	+= i;
	bitstream -> res	 = res;

	return;
}

int
BitstreamLookBitCopy(int pos,BSS *bitstream)
{
	unsigned char	*ptr = bitstream -> ptr;
	int		res  = bitstream -> res + pos - 1;

	while ( res >= 8 ) {
		ptr ++;
		res -= 8;
	}

	return( (int)( ( (*ptr) >> (7 - res) ) & 1 ) );
}

unsigned int
LookBitsFromStreamCopy(int bits,BSS *bitstream)
{
    int	i;
    unsigned int code=0;

    for ( i=1; i<=bits; i++ ) {
        code = (code << 1) + ( BitstreamLookBitCopy( i, bitstream ) & 1 );
    }
    return ( code );
}

unsigned int
GetBitsFromStreamCopy(int bits,BSS *bitstream)
{
    unsigned int code;
    code = LookBitsFromStreamCopy(bits,bitstream);
    BitstreamFlushBitsCopy(bits,bitstream);
    return ( code );
}

void
BitStreamCopy(int cnt,BSS *bitstream1,BSS *bitstream2)
{
    unsigned int code;
    int	bits;

    while ( cnt>=32 ) {
        bits  = 32;
        code  = GetBitsFromStreamCopy ( bits, bitstream1 );
                PutBitstoStream   ( bits, code, bitstream2 );
        cnt  -= 32;
    }
    if ( cnt>0 ) {
        code  = GetBitsFromStreamCopy ( cnt, bitstream1 );
                PutBitstoStream   ( cnt, code, bitstream2 );
    }

    return;
}

void
BitStreamMerge(int cnt,BSS *bitstream)
{
  unsigned int code;
  int	bits;

  while ( cnt>=32 ) {
    bits  = 32;
    code  = GetBitsFromStreamCopy ( bits, bitstream );
    PutBitstoStreamMerge  ( bits, code);
    cnt  -= 32;
  }
  if ( cnt>0 ) {
        code  = GetBitsFromStreamCopy ( cnt, bitstream );
	PutBitstoStreamMerge   ( cnt, code );
  }

  return;
}



