/*
 *	  ShapeCommon.c
 */
/* This code was obtained from Lehigh University/Vector Vision, Inc.. 
It was originally developed by OKI. 
Dr. Hongqiao Sun/Weiping integrated it to the Lehigh/VVI software. 

The copyrights or credits should belong to the following people and companies:

Zhixiong Wu and Yoshihiro Ueda of OKI,
Hongqiao Sun of Vector Vision,
Weiping Li of Lehigh University,

Minor modifications was made by Shipeng Li to integrate it to Sarnoff's SOL
software.
*/
#include <stdio.h>
#include "ShapeUtil.h"
#include "ShapeCodec.h"
#include "ShapeCommon.h"
/*
extern void
AddBorderToBAB (int x,int y,int blkn,int cr,int MAX_blkx,unsigned char **BABin,unsigned char **BABout,char **shape);
extern void
AdaptiveUpSampling(unsigned char **BABdown,unsigned char **BABresult,int sub_size);
*/
/* global variables for checking the boundary of
object */
int ObjectWidth, ObjectHeight; 

void
UpSampling_Still(int x,int y,int blkn,int cr,int blkx,unsigned char **buff,unsigned char**data,unsigned char **shape)
{
    int	i, j;

    int b_size=blkn/cr;

    if ( b_size==16 ) {
        for ( j=0; j<b_size; j++ ) {
        for ( i=0; i<b_size; i++ ) {
            data[j][i] = buff[j][i];
        }
        }
    } else if ( b_size==8 ) {
        unsigned char **down_tmp = (unsigned char**) malloc_2d_char( blkn + 4, blkn + 4);
        AddBorderToBAB(x,y,blkn,cr,blkx,buff,down_tmp,shape);
        AdaptiveUpSampling(data,down_tmp,8);
        free(down_tmp);
    } else if ( b_size==4 ) {
        unsigned char **down_tmp = (unsigned char **)malloc_2d_char( blkn + 4, blkn + 4);
        unsigned char **up_tmp   = (unsigned char **)malloc_2d_char( blkn    , blkn    );
        AddBorderToBAB(x,y,blkn,cr,blkx,buff,down_tmp,shape);
        AdaptiveUpSampling(down_tmp,up_tmp,4);
        AddBorderToBAB(x,y,blkn,cr/2,blkx,up_tmp,down_tmp,shape);
        AdaptiveUpSampling(down_tmp,data,8);
        free(down_tmp);
        free(up_tmp);
    } else {
        fprintf(stderr,"UpSampling():Wrong UpSampling Size (%d->16)\n",b_size);
        exit( 0 );
    }
}

#define	GetContextUS(a,b,c,d,e,f,g,h) ((int)((a)+((b)<<1)+((c)<<2)+((d)<<3)+((e)<<4)+((f)<<5)+((g)<<6)+((h)<<7)))
#define WEIGHT(a,b,c,d,e,f,g,h,i,j,k,l,t) ( (( 4*(a) + 2*((b)+(c)+(d)) + (e)+(f)+(g)+(h)+(i)+(j)+(k)+(l) )>(t))?1:0 )

static void
AdaptiveUpSampling(unsigned char **BABdown,unsigned char **BABresult,int sub_size)
{
    static int Th[256]={ /* look up table for the filter Threshold. */
			5, 6, 6, 8,  6, 7, 7, 8,  6, 5, 7, 6,  8, 8, 8, 8,
			6, 5, 5, 9,  8, 8, 8, 8,  7, 6, 8, 8,  8, 8, 8, 9,
			6, 7, 5, 8,  7, 9, 8, 8,  7, 8, 8, 8,  8, 8, 8, 9,
			7, 8, 8, 8,  8,10, 8, 9,  8,10, 8,11,  8,11, 9,10,
			6, 7, 7, 8,  7, 8, 8, 8,  7, 8, 8, 8,  9, 8,10, 9,
			7, 8, 8, 8,  8, 7, 8, 9,  8,10, 8, 9,  8,11, 9,10,
			7, 8, 8,10,  8,10,10,11,  8, 8,10, 9, 10,11,11,10,
			8, 8, 8,11,  8, 9,11,12,  8, 9,11,12, 11,12,10,11,
			6, 7, 7, 9,  7, 6, 8, 8,  5, 8, 8, 8,  6, 8,10, 9,
			8, 8, 8,10,  8,10, 8, 9,  8,10,10,11, 10,11, 9,10,
			8, 8, 6,10,  8, 8,10, 9,  8, 8,10, 9,  8, 9, 9,10,
			9, 8,10,11,  8,13, 9,10, 10,11,11,14, 11,12,10,11,
			7, 8, 8, 8,  8, 8, 8, 9,  8, 8, 8, 9,  8, 9,11,10,
			9, 8, 8, 9,  8,11, 9,10, 10,11, 9,10, 11,12,10,11,
			8, 8, 8,11, 10,11,11,10,  8,11, 9,12, 11,12,14,11,
		       10, 9,11,12, 11,12,12,11,  9,12,12,13, 12,13,11,14};

	/* Sampling position difference indicator */
	/* [upsampling grid (up to 16)][original position: m (12tap)] */
	/* m : original position for each filtering. see Fig. 0.10. */
	/* --  4  5 -- */
	/*  6  0  1  7 */
	/*  8  2  3  9 */
	/* -- 10 11 -- */
    static int xx[12] = { -1, +0, -1, +0, -1, +0, -2, +1, -2, +1, -1, +0, };
    static int yy[12] = { -1, -1, +0, +0, -2, -2, -1, -1, +0, +0, +1, +1, };
    static int d[16] = {0,+1,+1,+2,+2,+3,+3,+4,+4,+5,+5,+6,+6,+7,+7,+8};

    unsigned char val[12];
    int	context, m, iup, jup, up_size;

    up_size = sub_size<<1;

    for	( jup=0; jup<up_size; jup++ ) {
        for ( iup=0; iup<up_size; iup++ ) {

            for ( m=0; m<12; m++ )
                val[m] = BABdown[yy[m]+d[jup]+2][xx[m]+d[iup]+2];

            if ( iup%2==1 && jup%2==1 ) { /* P1 case : */

                context=GetContextUS(val[5],val[4],val[6],val[8],
                                     val[10],val[11],val[9],val[7]);
                BABresult[jup][iup] = WEIGHT(val[0],val[1],val[2],val[3],
                                             val[4],val[5],val[6],val[7],
                                             val[8],val[9],val[10],val[11],
                                             Th[context]);

            } else if ( iup%2==0 && jup%2==1 ) { /* P2 case : */

                context=GetContextUS(val[9],val[7],val[5],val[4],
                                     val[6],val[8],val[10],val[11]);
                BABresult[jup][iup] = WEIGHT(val[1],val[0],val[2],val[3],
                                             val[4],val[5],val[6],val[7],
                                             val[8],val[9],val[10],val[11],
                                             Th[context]);

            } else if ( iup%2==1 && jup%2==0 ) { /* P4 case : */

                context=GetContextUS(val[6],val[8],val[10],val[11],
                                     val[9],val[7],val[5],val[4]);
                BABresult[jup][iup] = WEIGHT(val[2],val[1],val[0],val[3],
                                             val[4],val[5],val[6],val[7],
                                             val[8],val[9],val[10],val[11],
                                             Th[context]);

            } else { /* P3 case : */

                context=GetContextUS(val[10],val[11],val[9],val[7],
                                     val[5],val[4],val[6],val[8]);
                BABresult[jup][iup] = WEIGHT(val[3],val[1],val[2],val[0],
                                             val[4],val[5],val[6],val[7],
                                             val[8],val[9],val[10],val[11],
                                             Th[context]);
            }
        }
    }
    return;
}

void
DownSampling_Still(unsigned char **buff,unsigned char **data,int b_size,int s_size)
{
    int	i, j, k, l;
    int	dat;
 
    if ( s_size == 1 ) {;
        for ( j=0; j<b_size; j++ ) {
        for ( i=0; i<b_size; i++ ) {
            data[j][i] = buff[j][i];
        }
        }
    } else {
        for ( j=0; j<b_size; j++ ) {
        for ( i=0; i<b_size; i++ ) {
            dat = 0;
            for ( l=0; l<s_size; l++ ) {
            for ( k=0; k<s_size; k++ ) {
                dat += buff[j*s_size+l][i*s_size+k];
            }
            }
            data[j][i] = ( (dat*2)<(s_size*s_size) ) ? 0 : 1;
        }
        }
    }
    return;
}

void
AddBorderToBAB(int blkx,int blky,int blksize,int cr_size,int MAX_blkx,unsigned char **BABinput,unsigned char **BABresult,unsigned char **shape)
{
    int	blkn = blksize/cr_size;
    int	i, j, k, tmp;

    for ( j=0; j<blkn+4; j++ )
    for ( i=0; i<blkn+4; i++ )
        BABresult[j][i] = 0;

    for ( j=0; j<blkn; j++ )
    for ( i=0; i<blkn; i++ )
        BABresult[j+2][i+2] = BABinput[j][i];

	/*  Up-left  */
    if ( blkx!=0 && blky!=0 ) {
        int	X = blkx * blksize - 2;
        int	Y = blky * blksize - 2;
        for ( j=0; j<2; j++ ) {
        for ( i=0; i<2; i++ ) {
	  if(Y+j < ObjectHeight && X+i < ObjectWidth)
            BABresult[j][i] = ( shape[Y+j][X+i]!=0 );
	  else
	    BABresult[j][i] = 0;
        }
        }
    }

	/*  up  */
    if ( blky!=0 ) {
        int	X = blkx * blksize;
        int	Y = blky * blksize - 2;
        for ( j=0; j<2   ; j++ ) {
        for ( i=0; i<blkn; i++ ) {
            tmp = 0;
            for ( k=0; k<cr_size; k++ ) 
	      if(Y+j < ObjectHeight && X+i*cr_size+k < ObjectWidth)
                tmp += ( shape[Y+j][X+i*cr_size+k]!=0 );
	    
            BABresult[j][i+2] = ( (tmp*2 < cr_size) ? 0 : 1 );
        }
        }
    }

	/*  left  */
    if ( blkx!=0 ) {
        int	X = blkx * blksize - 2;
        int	Y = blky * blksize;
        for ( j=0; j<blkn; j++ ) {
        for ( i=0; i<2   ; i++ ) {
            tmp = 0;
            for ( k=0; k<cr_size; k++ ) 
	      if(Y+j*cr_size+k < ObjectHeight && X+i < ObjectWidth)
                tmp += ( shape[Y+j*cr_size+k][X+i]!=0 );
            BABresult[j+2][i] = ( (tmp*2 < cr_size) ? 0 : 1 );
        }
        }
    }

	/*  up-right  */
    if ( blky!=0 && blkx<MAX_blkx-1 ) {
        int	X = (blkx + 1) * blksize;
        int	Y = blky * blksize - 2;
        for ( j=0; j<2; j++ ) {
        for ( i=0; i<2; i++ ) {
	  if(Y+j < ObjectHeight && X+i < ObjectWidth)
            BABresult[j][blkn+2+i] = ( shape[Y+j][X+i]!=0 );
	  else
	    BABresult[j][blkn+2+i] = 0;
        }
        }
    }

	/* padding right & bottom border pixels unknown at decoding time. */

    for	( i=0; i<blkn; i++ ) {
        BABresult[i+2][blkn+3] =
        BABresult[i+2][blkn+2] = BABresult[i+2][blkn+1];
        BABresult[blkn+3][i+2] =
        BABresult[blkn+2][i+2] = BABresult[blkn+1][i+2];
    }
    for	( i=-2; i<0; i++ ) {
        BABresult[blkn+3][i+2] =
        BABresult[blkn+2][i+2] = BABresult[blkn+1][i+2];
    }
    return;
}
