/*
 *	  ShapeDeCoding.c
 */
/* This code was obtained from Lehigh University/Vector Vision, Inc.. 
It was originally developed by OKI. 
Dr. Hongqiao Sun/Weiping integrated it to the Lehigh/VVI software. 

The copyrights or credits should belong to the following people and companies:

Zhixiong Wu and Yoshihiro Ueda of OKI,
Hongqiao Sun of Vector Vision,
Weiping Li of Lehigh University,

Minor modifications was made by Shipeng Li to integrate it to Sarnoff's SOL
software.
*/
#include <stdio.h>
#include "ShapeCodec.h"
#include "ShapeCommon.h"

#include "shape_def.h"
#include "ShapeUtil.h"

#define Error -1
extern int ObjectWidth, ObjectHeight;
extern int STO_const_alpha;
extern unsigned char STO_const_alpha_value;

static int ShapeContentDecode(int i,int j,int bsize,char **BAB,SBI *infor);
static int ShapeHeaderDecode(int i,int j,int blkx,SBI *infor);

int
ShapeDeCoding(unsigned char *OutShape, int object_width, int object_height)
{
    int	change_CR_disable; 
   /*  int	alphaTH		  = alphaTH;  */
    int	cols		  = object_width;
    int	rows		  = object_height;
    int	blkn		= 16;
    int	blkx		= (cols+15)/16;
    int	blky		= (rows+15)/16;
    SBI	ShapeInf, *infor;

    unsigned char **BAB		= malloc_2d_char ( blkn, blkn );
    unsigned char **BABdown	= malloc_2d_char ( blkn, blkn );
    unsigned char **BAB_ext	= malloc_2d_char ( blkn+4, blkn+4 );
    unsigned char **ip, **shape;

    int	i, j, k, l;
    int	smode;
    int	bsize;
    int	CR;
    int	bits = 0;

    ObjectWidth = object_width;
    ObjectHeight = object_height;
    
    shape = (unsigned char **)malloc(sizeof(unsigned char *)*object_height);
    if(shape == NULL) {
      fprintf(stderr,"Memory allocation failed\n");
      return Error;
    }
    for(i=0;i<object_height;i++) shape[i] = OutShape + i*object_width;

    infor = &ShapeInf;
    infor -> shape_mode = malloc_2d_int(blky, blkx);
    infor -> CR = malloc_2d_int(blky, blkx);
    infor -> ST = malloc_2d_int(blky, blkx);

    change_CR_disable = GetBitsFromStream(1);
    STO_const_alpha = GetBitsFromStream(1);
    if(STO_const_alpha) 
      STO_const_alpha_value = GetBitsFromStream(8);

    infor -> change_CR_disable	= change_CR_disable; 
   /*  infor -> alphaTH			= alphaTH;  */
    infor -> block_size			= blkn;
    infor -> block_num_hor		= blkx;
    infor -> block_num_ver		= blky;

    for ( j=0; j<blky; j++ ) {
    for ( i=0; i<blkx; i++ ) {
        if ( ShapeHeaderDecode ( i, j, blkx, infor ) == Error ) {
            fprintf(stderr,"\n CAE arithmetic decoding Error !\n");
            return  Error;
        }

        smode = infor -> shape_mode[j][i];

        if ( smode == ALL0 || smode == ALL255 ) {

            for ( l=0; l<blkn; l++ ) {
            for ( k=0; k<blkn; k++ ) {
	      if(j*blkn+l < ObjectHeight && i*blkn+k < ObjectWidth)
                shape[j*blkn+l][i*blkn+k] = (char)( smode == ALL0 ? 0 : 1 );
            }
            }

        } else {

            CR    = infor -> CR[j][i];
            bsize = blkn / CR;

            for ( l=0; l<bsize; l++ ) {
            for ( k=0; k<bsize; k++ ) {
                BABdown [l][k] = (char)0;
            }
            }

			/* Add surrounding pixels for CAE. */
            AddBorderToBAB ( i, j, blkn, CR, blkx, BABdown, BAB_ext, shape );

            if ( ShapeContentDecode ( i, j, bsize, BAB_ext, infor ) == Error ) {
                fprintf ( stderr, "\n CAE arithmetic decoding Error !\n" );
                return  Error;
            }

			/* make reconstructed 16x16 shape by BAB */
            for ( l=0; l<bsize; l++ ) {
            for ( k=0; k<bsize; k++ ) {
                BABdown [l][k] = BAB_ext [l+2][k+2];
            }
            }

            if ( CR != 1 ) {
				/* Reconstruct shape from BAB */
                UpSampling_Still  ( i, j, blkn, CR, blkx, BABdown, BAB, shape );
            } else {
                ip = BAB;
                BAB = BABdown;
                BABdown = ip;
                ip = (unsigned char **)NULL;
            }

			/* store result BAB.  */
            for ( l=0; l<blkn; l++ ) {
            for ( k=0; k<blkn; k++ ) {
	      if(j*blkn+l < ObjectHeight && i*blkn+k < ObjectWidth)
                shape[j*blkn+l][i*blkn+k] = ( BAB[l][k] == 0 ? 0 : 1 );
            }
            }

        }
    }
    }


    free_2d_char ( BAB, blkn );
    free_2d_char ( BABdown, blkn );
    free_2d_char ( BAB_ext, blkn+4 );
    free_2d_int  (infor->shape_mode, blky); 
    free_2d_int  (infor->CR, blky); 
    free_2d_int  (infor->ST, blky); 
    ByteAlignmentDec();
    free(shape);
    return ( 0 );
}

static int
ShapeContentDecode(int i,int j,int bsize,char **BAB,SBI *infor)
{
    int	k, l;
    int	dir = infor -> ST[j][i];
    int	context;
    int	ctx_max = 1023;
    ArDecoder decoder;

    StartArDecoder_Still( &decoder);

/*    if ( dir == HOR ) { */
    if ( dir == 1 ) {

        for ( l=0; l<bsize; l++ ) {
            for ( k=0; k<bsize; k++ ) {
                context = ( BAB [l+2+(+0)][k+2+(-1)] << 0 ) +
                          ( BAB [l+2+(+0)][k+2+(-2)] << 1 ) +
                          ( BAB [l+2+(-1)][k+2+(+2)] << 2 ) +
                          ( BAB [l+2+(-1)][k+2+(+1)] << 3 ) +
                          ( BAB [l+2+(-1)][k+2+(+0)] << 4 ) +
                          ( BAB [l+2+(-1)][k+2+(-1)] << 5 ) +
                          ( BAB [l+2+(-1)][k+2+(-2)] << 6 ) +
                          ( BAB [l+2+(-2)][k+2+(+1)] << 7 ) +
                          ( BAB [l+2+(-2)][k+2+(+0)] << 8 ) +
                          ( BAB [l+2+(-2)][k+2+(-1)] << 9 );

                if ( context > ctx_max ) {
                    fprintf ( stderr, "\n Shape context Error !\n" );
                    return  Error;
                }
                BAB [l+2][k+2] = ArDecodeSymbol_Still( &decoder, intra_prob[context] );
            }
            BAB [l+2][bsize+2] = BAB [l+2][bsize+3] = BAB [l+2][bsize+1];
        }

    } else {

        for ( k=0; k<bsize; k++ ) {
            for ( l=0; l<bsize; l++ ) {
                context = ( BAB [l+2+(-1)][k+2+(+0)] << 0 ) +
                          ( BAB [l+2+(-2)][k+2+(+0)] << 1 ) +
                          ( BAB [l+2+(+2)][k+2+(-1)] << 2 ) +
                          ( BAB [l+2+(+1)][k+2+(-1)] << 3 ) +
                          ( BAB [l+2+(+0)][k+2+(-1)] << 4 ) +
                          ( BAB [l+2+(-1)][k+2+(-1)] << 5 ) +
                          ( BAB [l+2+(-2)][k+2+(-1)] << 6 ) +
                          ( BAB [l+2+(+1)][k+2+(-2)] << 7 ) +
                          ( BAB [l+2+(+0)][k+2+(-2)] << 8 ) +
                          ( BAB [l+2+(-1)][k+2+(-2)] << 9 );

                if ( context > ctx_max ) {
                    fprintf ( stderr, "\n Shape context Error !\n" );
                    return  Error;
                }
                BAB [l+2][k+2] = ArDecodeSymbol_Still( &decoder, intra_prob[context] );
            }
            BAB [bsize+2][k+2] = BAB [bsize+3][k+2] = BAB [bsize+1][k+2];
        }

    }

    StopArDecoder_Still ( &decoder);

    return ( 0 );
}

static int ShapeHeaderDecode(int i,int j,int blkx,SBI *infor)
{
    int	dis_CR = infor -> change_CR_disable;
    int	UL= ( (i == 0      || j == 0) ? 0 : infor -> shape_mode[j-1][i-1] );
    int	UR= ( (i == blkx-1 || j == 0) ? 0 : infor -> shape_mode[j-1][i+1] );
    int	U = ( (j == 0) ? 0 : infor -> shape_mode[j-1][i] );
    int	L = ( (i == 0) ? 0 : infor -> shape_mode[j][i-1] );
    int	index = ( 27 * UL + 9 * U + 3 * UR + L );

    int	mode = 0;
    int	CR = 1;
    unsigned int code = 0;
    int	bits;

    do {
        bits = LMMR_first_shape_code_I [index][mode];
        code = LookBitsFromStream ( bits);

    } while ( code != CMMR_first_shape_code_I [index][mode] && (++mode) <= 2 );
	
    if ( code == CMMR_first_shape_code_I [index][mode] ) {

        BitstreamFlushBits_Still( bits);

        if ( mode == BORDER ) {

            if ( dis_CR == 0 ) {

                do {
                    bits = LCR [CR];
                    code = LookBitsFromStream ( bits );

                } while ( code != CCR[CR] && (CR=(CR<<1)) <= 4 );
	
                if ( code == CCR [CR] ) {
                    BitstreamFlushBits_Still( bits);
                } else {
                    fprintf(stderr,"\n Decode shape information CR Error at [%d, %d] \n", i, j );
                    return  Error;
                }

            }

            bits = LST;
            code = GetBitsFromStream ( bits);

        }
        infor -> shape_mode[j][i] = mode;
        infor -> CR        [j][i] = CR;
        infor -> ST        [j][i] = (int)code;

    } else {
        fprintf ( stderr, "\n Decode shape mode Error at [%d, %d] \n", i, j );
        return  Error;
    }

    return ( 0 );
}
