/* $Id: ac.c,v 1.20 1998/02/20 21:14:34 hjlee Exp $ */

/****************************************************************************/
/*   MPEG4 Visual Texture Coding (VTC) Mode Software                        */
/*                                                                          */
/*   This software was developed by                                         */
/*   Sarnoff Coporation                   and    Texas Instruments          */
/*   Iraj Sodagar   (iraj@sarnoff.com)           Jie Liang (liang@ti.com)   */
/*   Hung-Ju Lee    (hjlee@sarnoff.com)                                     */
/*   Paul Hatrack   (hatrack@sarnoff.com)                                   */
/*   Shipeng Li     (shipeng@sarnoff.com)                                   */
/*   Bing-Bing Chai (bchai@sarnoff.com)                                     */
/*                                                                          */
/* In the course of development of the MPEG-4 standard. This software       */
/* module is an implementation of a part of one or more MPEG-4 tools as     */
/* specified by the MPEG-4 standard.                                        */
/*                                                                          */
/* The copyright of this software belongs to ISO/IEC. ISO/IEC gives use     */
/* of the MPEG-4 standard free license to use this  software module or      */
/* modifications thereof for hardware or software products claiming         */
/* conformance to the MPEG-4 standard.                                      */
/*                                                                          */
/* Those intending to use this software module in hardware or software      */
/* products are advised that use may infringe existing  patents. The        */
/* original developers of this software module and their companies, the     */
/* subsequent editors and their companies, and ISO/IEC have no liability    */
/* and ISO/IEC have no liability for use of this software module or         */
/* modification thereof in an implementation.                               */
/*                                                                          */
/* Permission is granted to MPEG memebers to use, copy, modify,             */
/* and distribute the software modules ( or portions thereof )              */
/* for standardization activity within ISO/IEC JTC1/SC29/WG11.              */
/*                                                                          */
/* Copyright (C) 1998  Sarnoff Coporation and Texas Instruments             */ 
/****************************************************************************/

/************************************************************/
/*     Sarnoff Very Low Bit Rate Still Image Coder          */
/*     Copyright 1995, 1996, 1997, 1998 Sarnoff Corporation */
/************************************************************/

/************************************************************/
/*  Filename: ac.c                                          */
/*  Author: Bing-Bing Chai                                  */
/*  Date Modified: January 6, 1998                          */
/*                                                          */
/*  Descriptions:                                           */
/*    This file contains routines for integer arithmetic    */
/*    coding, which is based on the ac.c file from the SOL  */
/*    package. The original ac.c was obtained from public   */
/*    domain.                                               */
/*                                                          */
/*    The following routines are modified or created for    */
/*    the latest VTC package:                               */
/*      static void output_bit ()                           */
/*      void ac_encoder_init ()                             */
/*      int ac_encoder_done ()                              */
/*      static int input_bit ()                             */
/*      void ac_decoder_done ()                             */
/*                                                          */
/************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include "ac.h"
#include "bitpack.h"
#include "msg.h"


#define Code_value_bits 16
#define Malloc(size) (/* fprintf(stderr, "Malloc size =%d at line %d in %s\n", (size), __LINE__, __FILE__)?malloc(size): */malloc(size))
#define Calloc(nelm, elsz) (/* fprintf(stderr, "Calloc size =%d x %d = %d at line %d in %s\n", (nelm), (elsz), (nelm)*(elsz), __LINE__, __FILE__)?calloc(nelm, elsz): */calloc(nelm, elsz))

#define Top_value (((long)1<<Code_value_bits)-1)
#define First_qtr (Top_value/4+1)
#define Half	  (2*First_qtr)
#define Third_qtr (3*First_qtr)
#define Max_frequency 16383

/*----- For first order and mixed order probability model modes -----*/
#define LAST_SYM_SLOT  (acm->lastSym / acm->freq1Q)

#define IPDF(_sym, context)   ((_sym)+(acm->nsym*(context)))
#define ICDF(_sym, context)   ((_sym)+((acm->nsym+1)*(context)))

/*----- For zeroth order probability model mode -----*/
#define CDFZERO(_sym) ((int)acm->cfreq_0[(_sym)])

/*----- For first order probability model mode -----*/
#define CDFFIRST(_sym, context) ((int)acm->cfreq_1[ICDF((_sym),(context))])

/*----- For mixed order probability model mode -----*/
#define CDFMIX(_sym, context) ((int)acm->cfreq_1[ICDF((_sym),(context))] + \
			(int)acm->cfreq_0[(_sym)])


#define MIN(a,b) (( (a)<(b) ) ? (a) : (b))


/* static function prototypes */
static void output_bit (ac_encoder *ace, int bit);
static void bit_plus_follow (ac_encoder *ace, int bit);
static void update_model (ac_model *acm, int sym, int context);
static int input_bit (ac_decoder *acd);

static zero_accum=0;

/************************************************************************/
/*              Error Checking and Handling Macros                      */
/************************************************************************/

#define error(m)                                           \
do  {                                                      \
  fflush (stdout);                                         \
  fprintf (stderr, "%s:%d: error: ", __FILE__, __LINE__);  \
  fprintf (stderr, m);                                     \
  fprintf (stderr, "\n");                                  \
  exit (1);                                                \
}  while (0)

#define check(b,m)                                         \
do  {                                                      \
  if (b)                                                   \
    error (m);                                             \
}  while (0)

/************************************************************************/
/*                           Bit Output                                 */
/************************************************************************/

/**************************************************/
/*  Added bit stuffing to prevent start code      */
/*  emulation, i.e., add a "1" bit after every 22 */
/*  consecutive "0" bits in the bit stream        */
/**************************************************/

static void output_bit (ac_encoder *ace, int bit)
{
  ace->buffer <<= 1;
  if (bit)
    ace->buffer |= 0x01;
  
  ace->bits_to_go -= 1;
  ace->total_bits += 1;
  if (ace->bits_to_go==0)  {
    
    if (ace->bitstream) {
      if (ace->bitstream_len >= MAX_BUFFER) 
	if ((ace->bitstream = realloc(ace->bitstream, sizeof(unsigned char)*
		   (ace->bitstream_len/MAX_BUFFER+1)*MAX_BUFFER+2))==NULL)
	{
	  fprintf(stderr,
	    "Couldn't reallocate memory for ace->bitstream in output_bit.\n");
	  exit(-1);
	}
      
      ace->bitstream[ace->bitstream_len++] = ace->buffer;
    }
    
    if (ace->fp) {
      putc (ace->buffer, ace->fp);
    }
    ace->bits_to_go = 8;
  }
  
#if STUFF
  if(bit==0){
    zero_accum++;
    if(zero_accum==22){
      output_bit(ace,1);
      zero_accum=0;
    }
  }
  else
    zero_accum=0;
#endif

  return;
}

static void bit_plus_follow (ac_encoder *ace, int bit)
{
  output_bit (ace, bit);
  while (ace->fbits > 0)  {
    output_bit (ace, !bit);
    ace->fbits -= 1;
  }

  return;
}

/* void mput_bistream(unsigned char value, ac_encoder *ace) */
void mput_bistream(int value, int nbits, ac_encoder *ace)
{

  int extension;
  int module = 1 << nbits;
  int mask = (1 << nbits) - 1;
  
  if (ace->bitstream) {
    if (ace->bitstream_len >= MAX_BUFFER) 
      if (realloc(ace->bitstream, sizeof(unsigned char)*
		  (ace->bitstream_len/MAX_BUFFER+1)*MAX_BUFFER+2)==NULL)
      {
	fprintf(stderr,"Couldn't reallocate memory in mput_bistream.\n");
	exit(-1);
      }


    while (value/module > 0) {
      extension = 1;
      ace->bitstream[ace->bitstream_len++] =
          (unsigned char)( (value%module) | (extension << nbits));
      value = value >> nbits;
    }
    extension = 0;
    ace->bitstream[ace->bitstream_len++] =
          (unsigned char)( (value%module) | (extension << nbits));
    
    /* ace->bitstream[ace->bitstream_len++] = (unsigned char)value; */
  }
}

/************************************************************************/
/*                             Encoder                                  */
/************************************************************************/

/*****************************************************/
/*  Added zero_accum initlization for stuffing bits  */
/*****************************************************/
void ac_encoder_init (ac_encoder *ace, const char *fn, 
		      unsigned char *bitstream)
{

  if (fn)  {
    ace->fp = fopen (fn, "w");
    check (!ace->fp, "arithmetic encoder could not open file");
  }  else  {
    ace->fp = NULL;
  }

  ace->bits_to_go = 8;

  ace->low = 0;
  ace->high = Top_value;
  ace->fbits = 0;
  ace->buffer = 0;

  ace->total_bits = 0;

  /* can't release the memory ! */
  ace->bitstream_len = 0;
  if(bitstream!=NULL) {
    if ((ace->bitstream = bitstream 
	 = (unsigned char *)realloc(bitstream, 
				    MAX_BUFFER*sizeof(unsigned char)))==NULL)
      errorHandler("Couldn't reallocate memory for ace->bitstream in " \
		   "output_bit.\n");
  }
  else {
    if ((ace->bitstream = bitstream 
	 = (unsigned char *)malloc(MAX_BUFFER*sizeof(unsigned char)))==NULL)
      errorHandler("Couldn't reallocate memory for ace->bitstream in " \
		   "output_bit.\n");
  }

  zero_accum=0;

#if STUFF
  /* always start arithmetic bitstream with a 1 bit. */
  emit_bits(1,1);
#endif

  return;
}



/***************************************************************/
/* Added stuffing bits to prevent start code emulation.        */
/* Reassigned bitbuffer pointer to prevent potential problem.  */
/***************************************************************/
int ac_encoder_done (ac_encoder *ace,unsigned char **bitbuffer)
{
  ace->fbits += 1;
  if (ace->low < First_qtr)
    bit_plus_follow (ace, 0);
  else
    bit_plus_follow (ace, 1);

  if (ace->fp) putc (ace->buffer >> ace->bits_to_go, ace->fp);
  if (ace->fp) fclose (ace->fp);

  if (ace->bits_to_go != 8){
     ace->bitstream[ace->bitstream_len++] = (ace->buffer << ace->bits_to_go);
#if STUFF
    if(ace->bitstream[ace->bitstream_len-1]&(1<<ace->bits_to_go) == 0){
      ace->bitstream[ace->bitstream_len-1] += (1<<ace->bits_to_go)-1;
      ace->total_bits +=1;
    }
#endif
  }
#if STUFF
  else if(ace->bitstream[ace->bitstream_len-1]&1 ==0){
    /* stuffing bits to prevent start code emulation */
    ace->bitstream[ace->bitstream_len++] = 255;
    ace->total_bits +=1;
  }	
#endif


  *bitbuffer=ace->bitstream;


  return ace->total_bits;
}

int ac_encode_symbol (ac_encoder *ace, ac_model *acm, int sym, ...)
{
  long range;
  int tmp;
  va_list ap;
  int context=0; /* set to avoid UMRs */

  if (sym<0 || sym>=acm->nsym) 
     fprintf(stderr,"sym=%d  acm->nsym=%d\n",sym,acm->nsym);
  check (sym<0||sym>=acm->nsym, "symbol out of range");

  tmp = ace->total_bits;
  range = (long)(ace->high-ace->low)+1;

  switch (acm->order)
  {
    case AC_ORDER_ZERO:
      ace->high = ace->low + (range*CDFZERO(sym))/CDFZERO(0)-1;
      ace->low  = ace->low + (range*CDFZERO(sym+1))/CDFZERO(0);
      break;
    case AC_ORDER_FIRST:
      if (acm->context < 0)
      {
	context = LAST_SYM_SLOT;
	ace->high = ace->low + 
	  (range*CDFFIRST(sym,context))/CDFFIRST(0,context)-1;
	ace->low  = ace->low + 
	  (range*CDFFIRST(sym+1,context))/CDFFIRST(0,context);
      }
      else
      {
	va_start(ap,sym);
	context = acm->context = va_arg(ap, int);
	ace->high = ace->low + 
	  (range*CDFFIRST(sym,context))/CDFFIRST(0,context)-1;
	ace->low  = ace->low + 
	  (range*CDFFIRST(sym+1,context))/CDFFIRST(0,context);
	va_end(ap);
      }
      break;
    case AC_ORDER_MIX:
      if (acm->context < 0)
      {
	context = LAST_SYM_SLOT;
	ace->high = ace->low + 
	  (range*CDFMIX(sym,context))/CDFMIX(0,context)-1;
	ace->low  = ace->low + 
	  (range*CDFMIX(sym+1,context))/CDFMIX(0,context);
      }
      else
      {
	va_start(ap,sym);
	context = acm->context = va_arg(ap, int);
	ace->high = ace->low + 
	  (range*CDFMIX(sym,context))/CDFMIX(0,context)-1;
	ace->low  = ace->low + 
	  (range*CDFMIX(sym+1,context))/CDFMIX(0,context);
	va_end(ap);
      }
      break;
    default:
      error("Invalid probability model order in ac_encode_symbol.\n");
  }

  for (;;)  {
    if (ace->high<Half)  {
      bit_plus_follow (ace, 0);
    }  
    else if (ace->low>=Half)  {
      bit_plus_follow (ace, 1);
      ace->low -= Half;
      ace->high -= Half;
    }  
    else if (ace->low>=First_qtr && ace->high<Third_qtr)  {
      ace->fbits += 1;
      ace->low -= First_qtr;
      ace->high -= First_qtr;
    }  
    else
      break;
    ace->low = 2*ace->low;
    ace->high = 2*ace->high+1;
  }

  if (acm->adapt)
    update_model (acm, sym, context);

  return (ace->total_bits - tmp);
}

#if 0
long ac_encoder_bits (ac_encoder *ace)
{
  return ace->total_bits;
}
#endif

/************************************************************************/
/*                            Bit Input                                 */
/************************************************************************/

/*********************************************************/
/* Modified to be consistant with the functions in       */
/* bitpack.c, i.e., using nextinputbit() to get the new  */
/* bits from the bit stream.                             */
/*                                                       */
/* Added remove stuffing bits, refer to output_bit() for */
/* more details.                                         */
/*********************************************************/
static int input_bit (ac_decoder *acd)
{
  int t;

  if (acd->bits_to_go==0)  {
     acd->bits_to_go = 8;
  }

  acd->bits_to_go -= 1;
  acd->total_bits ++;

  t=nextinputbit();

#if STUFF
  /* remove stuffing bits */
  if(t==0){
    zero_accum++;
    if(zero_accum==22){
      if((input_bit(acd)) !=1)
        errorHandler("Error in decoding stuffing bits");
      zero_accum=0;
    }
  }
  else
    zero_accum=0;    
#endif


  return(t);
}

int mget_bitstream(int nbits, ac_decoder *acd)
{
   int count=0;
   int word=0;
   int value=0;
   int module=1<<(nbits);
   
   do {
     word=acd->bitstream[acd->bitstream_ptr++];
     value += (word & (module-1))<<(count*nbits);
     count++;
   } while (word>>nbits);
   return (value);


   /* return (acd->bitstream[acd->bitstream_ptr++]); */
}


/************************************************************************/
/*                             Decoder                                  */
/************************************************************************/

void ac_decoder_open (ac_decoder *acd, const char *fn)
{
  if (fn != NULL) {
     acd->fp = fopen (fn, "r");
     check (!acd->fp, "arithmetic decoder could not open file");
  }

  return;
}

void ac_decoder_init (ac_decoder *acd, const char *fn, 
		      unsigned char *bitstream)
{
  int i;
 
#if STUFF 
  /* remove first stuffing bit */
  if(!get_X_bits(1))
    errorHandler("Error in extracting the stuffing bit at the beginning of\n"\
		 "arithmetic decoding.");
#endif

  acd->bits_to_go = 0;
  /*acd->bitstream = bitstream;*/

  acd->value = 0;
  for (i=1; i<=Code_value_bits; i++)  {
    acd->value = 2*acd->value + input_bit(acd); 
  }
  acd->low = 0;
  acd->high = Top_value;

  acd->total_bits = 0;

  zero_accum=0;

  return;
}


/*******************************************************/
/* Added restore_arithmetic_offset() called to recover */
/* the extra bits read in by decoder. This routine is  */
/* defined in bitpack.c                                */
/*******************************************************/
void ac_decoder_done (ac_decoder *acd)
{
  if (acd->fp) fclose (acd->fp);

  restore_arithmetic_offset(acd->bits_to_go);
  acd->total_bits +=acd->bits_to_go;

  if((acd->total_bits) %8 !=0)
    errorHandler("Did not get alignment in arithmetic decoding");
}

int ac_decode_symbol (ac_decoder *acd, ac_model *acm, ...)
{
  register high,low,value;
  long range;
  int cum;
  int sym;
  va_list ap;
  int context=0; /* set to avoid UMRs */

  high=acd->high; low=acd->low; value=acd->value;
  range = (long)(high-low)+1;

  switch (acm->order)
  {
    case AC_ORDER_ZERO:
      cum = (((long)(value-low)+1)*CDFZERO(0)-1)/range;
      
      for (sym = 0; CDFZERO(sym+1)>cum; sym++)
	/* do nothing */ ;
      
      /* check (sym<0||sym>=acm->nsym, "symbol out of range"); */
      
      high = low + (range*CDFZERO(sym))/CDFZERO(0)-1;
      low  = low +  (range*CDFZERO(sym+1))/CDFZERO(0);
      break;
    case AC_ORDER_FIRST:
      if (acm->context < 0)
      {
	context = LAST_SYM_SLOT;

	cum = 
	  (((long)(value-low)+1)*CDFFIRST(0,context)-1)/range;
      
	for (sym = 0; CDFFIRST(sym+1,context)>cum; sym++)
	  /* do nothing */ ;
	
	check (sym<0||sym>=acm->nsym, "symbol out of range");
	
	high = low + 
	  (range*CDFFIRST(sym,context))/CDFFIRST(0,context)-1;
	low  = low +  
	  (range*CDFFIRST(sym+1,context))/CDFFIRST(0,context);
      }
      else
      {
	va_start(ap,sym);
	context = acm->context = va_arg(ap, int);

	cum = 
	  (((long)(value-low)+1)*CDFFIRST(0,acm->context)-1)/range;
      
	for (sym = 0; CDFFIRST(sym+1,acm->context)>cum; sym++)
	  /* do nothing */ ;
	
	check (sym<0||sym>=acm->nsym, "symbol out of range");
	
	high = low + 
	  (range*CDFFIRST(sym,context))/CDFFIRST(0,context)-1;
	low  = low +  
	  (range*CDFFIRST(sym+1,context))/CDFFIRST(0,context);

	va_end(ap);
      }
      break;
    case AC_ORDER_MIX:
      if (acm->context < 0)
      {
	context = LAST_SYM_SLOT;

	cum = 
	  (((long)(value-low)+1)*CDFMIX(0,context)-1)/range;
	
	for (sym = 0; CDFMIX(sym+1,context)>cum; sym++)
	  /* do nothing */ ;
	
	check (sym<0||sym>=acm->nsym, "symbol out of range");
	
	high = low + 
	  (range*CDFMIX(sym,context))/CDFMIX(0,context)-1;
	low  = low +  
	  (range*CDFMIX(sym+1,context))/CDFMIX(0,context);
      }
      else
      {
	va_start(ap,sym);
	context = acm->context = va_arg(ap, int);
	cum = 
	  (((long)(value-low)+1)*CDFMIX(0,acm->context)-1)/range;
	
	for (sym = 0; CDFMIX(sym+1,acm->context)>cum; sym++)
	  /* do nothing */ ;
	
	check (sym<0||sym>=acm->nsym, "symbol out of range");
	
	high = low + 
	  (range*CDFMIX(sym,context))/CDFMIX(0,context)-1;
	low  = low +  
	  (range*CDFMIX(sym+1,context))/CDFMIX(0,context);

	va_end(ap);
      }
      break;
    default:
      error("Invalid probability model order in ac_decode_symbol.\n");
  }

  for (;;)  {
    if (high<Half)  {
      /* do nothing */
    }  else if (low>=Half)  {
      value -= Half;
      low -= Half;
      high -= Half;
    }  else if (low>=First_qtr && high<Third_qtr)  {
      value -= First_qtr;
      low -= First_qtr;
      high -= First_qtr;
    }  else
      break;
    low = low<<1;
    high = (high<<1)+1;
    value = (value<<1) + input_bit(acd);
  }
  acd->high=high; acd->low=low; acd->value=value;

  if (acm->adapt)
    update_model (acm, sym, context);

  return sym;
}


/************************************************************************/
/*                       Probability Model                              */
/************************************************************************/

static void acModelInitZero (ac_model *acm, int nsym, int *ifreq, int adapt)
{
  int i;
  
  acm->nsym = nsym;


  acm->freq_0 = (unsigned short *) Malloc (nsym*sizeof (unsigned short));
  check (!acm->freq_0, "arithmetic coder model allocation failure");
  acm->cfreq_0 = (unsigned short *) Calloc (nsym+1, sizeof (unsigned short));
  check (!acm->cfreq_0, "arithmetic coder model allocation failure");
  acm->adapt = adapt;
  
  if (ifreq)  {
    acm->cfreq_0[acm->nsym] = 0;
    for (i=acm->nsym-1; i>=0; i--)  {
      acm->freq_0[i] = ifreq[i];
      acm->cfreq_0[i] = acm->cfreq_0[i+1] + acm->freq_0[i];
    }
    
    
    if (acm->cfreq_0[0] > Max_frequency)  {
      int cum = 0;
      acm->cfreq_0[acm->nsym] = 0;
      for (i = acm->nsym-1; i>=0; i--)  {
	acm->freq_0[i] = ((int)acm->freq_0[i] + 1) / 2;
	cum += acm->freq_0[i];
	acm->cfreq_0[i] = cum;
      }
    }
    
    if (acm->cfreq_0[0] > Max_frequency)
      error ("arithmetic coder model max frequency exceeded");
  }  
  else  {
    for (i=0; i<acm->nsym; i++) {
      acm->freq_0[i] = 1;
      acm->cfreq_0[i] = acm->nsym - i;
    }
    acm->cfreq_0[acm->nsym] = 0;
  }
  
  return;
}

static void acModelInitFirst (ac_model *acm, int nsym, int *ifreq, int adapt,
			      int context, int firstOrderSlots)
{
  int i,j;

  acm->nsym          = nsym;
  acm->adapt         = adapt;
  acm->lastSym       = 0;        /* initialize to zeroth symbol */
  acm->context       = context; 
  acm->numFreq1Slots = firstOrderSlots;
  if (context < 0)
  {
    acm->freq1Q        = nsym/acm->numFreq1Slots; 
    acm->numFreq1Slots = nsym/acm->freq1Q + (nsym%acm->freq1Q != 0);
  }

  acm->freq_1 = (unsigned short *) Calloc (acm->numFreq1Slots*nsym, 
					 sizeof (unsigned short));
  check (!acm->freq_1, "arithmetic coder model allocation failure");

  acm->cfreq_1 = (unsigned short *) Calloc (acm->numFreq1Slots*(nsym+1), 
					  sizeof (unsigned short));
  check (!acm->cfreq_1, "arithmetic coder model allocation failure");
  
  if (ifreq)  
  {
    for (j=0; j<acm->numFreq1Slots; ++j)
    {
      acm->cfreq_1[(j*(acm->nsym+1))+acm->nsym] = 0;
      for (i=acm->nsym-1; i>=0; i--) 
      {
	acm->freq_1[(j*acm->nsym)+i] = ifreq[(j*acm->nsym)+i];
	acm->cfreq_1[(j*(acm->nsym+1))+i] = 
	  acm->cfreq_1[(j*(acm->nsym+1))+i+1] + acm->freq_1[(j*acm->nsym)+i];
      }
      
      if (acm->cfreq_1[(j*(acm->nsym+1))+0] > Max_frequency)
      {
	int cum = 0;
	acm->cfreq_1[(j*(acm->nsym+1))+acm->nsym] = 0;
	for (i = acm->nsym-1; i>=0; i--) 
	{
	  acm->freq_1[(j*acm->nsym)+i] =
	    ((int)acm->freq_1[(j*acm->nsym)+i] + 1) / 2;
	  cum += acm->freq_1[(j*acm->nsym)+i];
	  acm->cfreq_1[(j*(acm->nsym+1))+i] = cum;
	}
      }
      
      if (acm->cfreq_1[(j*(acm->nsym+1))+0] > Max_frequency)
	error ("arithmetic coder model max frequency exceeded");
    }
  }  
  else 
  {
    for (j=0; j<acm->numFreq1Slots; ++j)
    {
      for (i=0; i<acm->nsym; i++) 
      {
	acm->freq_1[(j*acm->nsym)+i] = 1;
	acm->cfreq_1[(j*(acm->nsym+1))+i] = acm->nsym - i;
      }
      acm->cfreq_1[(j*(acm->nsym+1))+acm->nsym] = 0;
    }
  }

  return;
}

static void acModelInitMix (ac_model *acm, int nsym, int *ifreq, int adapt,
			    int context, 
			    int firstOrderSlots, int firstOrderInc)
{
  int i;

  acm->nsym          = nsym;
  acm->adapt         = adapt;
  acm->lastSym       = 0; /* initialize to zeroth symbol */
  acm->context       = context;
  acm->numFreq1Slots = firstOrderSlots;
  if (context < 0)
  {
    acm->freq1Q        = nsym/acm->numFreq1Slots;
    acm->numFreq1Slots = nsym/acm->freq1Q + (nsym%acm->freq1Q != 0);
  }
  acm->firstOrderFreqInc = firstOrderInc;
  

  acm->freq_0 = (unsigned short *) Calloc (nsym, sizeof (unsigned short));
  check (!acm->freq_0, "arithmetic coder model allocation failure");

  acm->cfreq_0 = (unsigned short *) Calloc (nsym+1,sizeof (unsigned short));
  check (!acm->cfreq_0, "arithmetic coder model allocation failure");

  acm->freq_1 = (unsigned short *) Calloc (acm->numFreq1Slots*nsym, 
					 sizeof (unsigned short));
  check (!acm->freq_1, "arithmetic coder model allocation failure");

  acm->cfreq_1 = (unsigned short *) Calloc (acm->numFreq1Slots*(nsym+1), 
					  sizeof (unsigned short));
  check (!acm->cfreq_1, "arithmetic coder model allocation failure");
  
  /* we initialize the zeroth order part only */
  if (ifreq)  
  {
    acm->cfreq_0[acm->nsym] = 0;
    for (i=acm->nsym-1; i>=0; i--) 
    {
      acm->freq_0[i] = ifreq[i];
      acm->cfreq_0[i] = acm->cfreq_0[i+1] + acm->freq_0[i];
    }
    
    
    if (acm->cfreq_0[0] > Max_frequency)
    {
      int cum = 0;
      acm->cfreq_0[acm->nsym] = 0;
      for (i = acm->nsym-1; i>=0; i--) 
      {
	acm->freq_0[i] = ((int)acm->freq_0[i] + 1) / 2;
	cum += acm->freq_0[i];
	acm->cfreq_0[i] = cum;
      }
    }
    
    if (acm->cfreq_0[0] > Max_frequency)
      error ("arithmetic coder model max frequency exceeded");
  }  
  else 
  {
    int j;

    for (i=0; i<acm->nsym; i++) 
    {
      acm->freq_0[i] = 1;
      acm->cfreq_0[i] = acm->nsym - i;
    }
    acm->cfreq_0[acm->nsym] = 0;

    /* initialize first order part with an increment for last symbol
       in same slot as current symbol */
    if (acm->context < 0)
    {
      for (j = 0; j < acm->numFreq1Slots; ++j)
      {
	int k, upSym;
	
	upSym = MIN(acm->nsym,(j+1)*acm->freq1Q);
	for (i = j*acm->freq1Q; i < upSym; ++i)
	{
	  acm->freq_1[j*acm->nsym + i] += acm->firstOrderFreqInc;
	  for (k=i; k>=0; k--)
	    acm->cfreq_1[j*(acm->nsym+1) + k] += acm->firstOrderFreqInc;
	}
	acm->cfreq_1[(j*(acm->nsym+1))+acm->nsym] = 0;
      }
    }
  }

  return;
}

void ac_model_init (ac_model *acm, int nsym, int *ifreq, int adapt, int order,
		    int context, int firstOrderSlots, int firstOrderInc)
{
  acm->order = order;

  switch (acm->order)
  {
    case AC_ORDER_ZERO:
      acModelInitZero(acm,nsym,ifreq,adapt);
      break;
    case AC_ORDER_FIRST:
      acModelInitFirst(acm,nsym,ifreq,adapt,context,firstOrderSlots);
      break;      
    case AC_ORDER_MIX:
      acModelInitMix(acm,nsym,ifreq,adapt,context,
		     firstOrderSlots,firstOrderInc);
      break;
    default:
      error("Invalid probability model order in ac_model_init.\n");
  }
}

static void acModelDoneZero (ac_model *acm)
{
  acm->nsym = 0;
  
  free (acm->freq_0);
  acm->freq_0 = NULL;
  
  free (acm->cfreq_0);
  acm->cfreq_0 = NULL;

  return;
}

static void acModelDoneFirst (ac_model *acm)
{
  acm->nsym = 0;
  
  free (acm->freq_1);
  acm->freq_1 = NULL;
  
  free (acm->cfreq_1);
  acm->cfreq_1 = NULL;

  return;
}

static void acModelDoneMix (ac_model *acm)
{
  acm->nsym = 0;
  
  free (acm->freq_0);
  acm->freq_0 = NULL;
  
  free (acm->cfreq_0);
  acm->cfreq_0 = NULL;
  
  free (acm->freq_1);
  acm->freq_1 = NULL;
  
  free (acm->cfreq_1);
  acm->cfreq_1 = NULL;

  return;
}

void ac_model_done (ac_model *acm)
{
  switch (acm->order)
  {
    case AC_ORDER_ZERO:
      acModelDoneZero(acm);
      break;
    case AC_ORDER_FIRST:
      acModelDoneFirst(acm);
      break;      
    case AC_ORDER_MIX:
      acModelDoneMix(acm);
      break;
    default:
      error("Invalid probability model order in ac_model_done.\n");
  }
}

static void updateModelZero (ac_model *acm, int sym)
{
  register unsigned short *freq_0,*cfreq_0;
  register i;

  freq_0=acm->freq_0; cfreq_0=acm->cfreq_0;
  if (cfreq_0[0]==Max_frequency)  {
    register cum=0;
    cfreq_0[acm->nsym] = 0;
    for (i = acm->nsym-1; i>=0; i--)  {
      freq_0[i] = ((int)freq_0[i] + 1) /2;
      cum += freq_0[i];
      cfreq_0[i] = cum;
    }
  }

  freq_0[sym] += 1;
  for (i=sym; i>=0; i--)
    cfreq_0[i] += 1;

  return;
}



static void updateModelFirst (ac_model *acm, int sym, int context)
{
  int i;
  
 /* Check  1st orders for going over maximum */
  if (CDFFIRST(0,context) >= Max_frequency-1)
  {
    int j;
    unsigned short *fp, *cfp;
    int cfidx;

    for (j=0; j<acm->numFreq1Slots; ++j)
    {
      cfidx = j*(acm->nsym+1)+acm->nsym;
      acm->cfreq_1[cfidx] = 0;
      
      --cfidx;
      cfp = acm->cfreq_1+cfidx;
      fp  = acm->freq_1+cfidx-j;
      for (i = acm->nsym-1; i>=0; i--)  
      {
	++(*fp);
	*fp >>= 1;
	*cfp = *(cfp+1) + *fp;

	--cfp;
	--fp;
      }
    } 
  }
    
  /* Update 1st order df */
  ++acm->freq_1[IPDF(sym,context)];
  for (i=sym; i>=0; i--)
    ++acm->cfreq_1[ICDF(i,context)];

  /* Update state (last symbol encoded/decoded) */
  acm->lastSym = sym;

  return; 
}

static void updateModelMix (ac_model *acm, int sym, int context)
{
  int i;
  
  /* scale down *all* dfs if neccesary */
  /* Check zero-th and 1st orders for going over maximum */
  if (CDFMIX(0,context) >= Max_frequency-acm->firstOrderFreqInc)
  {
    int j;
    unsigned short *fp, *cfp;
    int cfidx;
    int cum;

    for (j=0; j<acm->numFreq1Slots; ++j)
    {
      cfidx = j*(acm->nsym+1)+acm->nsym;
      acm->cfreq_1[cfidx] = 0;
      
      --cfidx;
      cfp = acm->cfreq_1+cfidx;
      fp  = acm->freq_1+cfidx-j;
      for (i = acm->nsym-1; i>=0; i--)  
      {
	++(*fp);
	*fp >>= 1;
	*cfp = *(cfp+1) + *fp;

	--cfp;
	--fp;
      }
    } 

    cum=0;
    acm->cfreq_0[acm->nsym] = 0;
    for (i = acm->nsym-1; i>=0; i--)  
    {
      acm->freq_0[i] = ((int)acm->freq_0[i] + 1) / 2;
      cum += acm->freq_0[i];
      acm->cfreq_0[i] = cum;
    }
  }
    
  /* Update zero-th order df */  
  ++acm->freq_0[sym];
  for (i=sym; i>=0; i--)
    ++acm->cfreq_0[i];

  /* Update 1st order df */
  acm->freq_1[IPDF(sym,context)] += acm->firstOrderFreqInc;
  for (i=sym; i>=0; i--)
    acm->cfreq_1[ICDF(i,context)] += acm->firstOrderFreqInc;

  /* Update state (last symbol encoded/decoded) */
  acm->lastSym = sym;

  return; 
}

static void update_model (ac_model *acm, int sym, int context)
{
  switch (acm->order)
  {
    case AC_ORDER_ZERO:
      updateModelZero(acm, sym);
      break;
    case AC_ORDER_FIRST:
      updateModelFirst(acm, sym, context);
      break;      
    case AC_ORDER_MIX:
      updateModelMix(acm, sym, context);
      break;
    default:
      error("Invalid probability model order in update_model.\n");
  }
}

#if 0
/* Not currently used - also update ac.h if used */
/* new */
void ac_model_save (ac_model *acm, int *ifreq)
{
  int i;

  for (i=acm->nsym-1; i>=0; i--)  {
    ifreq[i] = acm->freq[i];
  }

  return;
}
#endif
